/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/* Generic read and write operations to a block device. */

#include <linux/bio.h>
#include <linux/delay.h>

#include "blkrw.h"
#include "debug.h"
#include "portability.h"

#ifdef WITH_EXT2BITMAP

#define SECTOR_SIZE 512

struct blk_priv {
	atomic_t *numbios;
	void *buffer;
	struct page *page;
	unsigned int length;
	int *bio_error;
	sector_t sector;
	int type;
};

/*
 * Set the status of the bio to completed
 */
static BIO_END_IO_PROTO(blk_read_endio, bio)
{
	struct blk_priv *bp = bio->bi_private;

	stcvsm_debug(LOG_BLKRW, "sector %llu, buffer %p, length %u, buff_type %d, err %d\n",
		(unsigned long long)bp->sector, bp->buffer, bp->length, bp->type, BIO_ERROR(bio));

	if (BIO_ERROR(bio) == 0) {
		if (bp->type == BLK_BUFF_MEM) {
			char *page_buffer;

			stcvsm_debug(LOG_BLKRW, "Finished Memory I/O number %d (%p)\n",
					atomic_read(bp->numbios), bp->buffer);
			page_buffer = stc_kmap_atomic(bp->page);
			memcpy(bp->buffer, page_buffer, bp->length);
			stcvsm_debug(LOG_BLKRW, "(unsigned long *)bp->buffer[1]: 0x%lx\n",
					((unsigned long *)bp->buffer)[1]);
			stc_kunmap_atomic(page_buffer);
		} else if (bp->type == BLK_BUFF_PAGE) {
			stcvsm_debug(LOG_BLKRW, "Finished Page I/O number %d (%p)\n",
					atomic_read(bp->numbios), bp->page);
		} else if (bp->type == BLK_BUFF_BIO) {
			stcvsm_debug(LOG_BLKRW, "Finished Bio I/O number %d (%p)\n",
					atomic_read(bp->numbios), bp->page);
		} else {
			stcvsm_debug(LOG_BLKRW, "Finished Unknown I/O number %d (%p)\n",
					atomic_read(bp->numbios), bp->page);
		}
#ifdef DEBUG
		{
			char *bv_buf;

			bv_buf = stc_kmap_atomic(bio->bi_io_vec[0].bv_page);
			stcvsm_debug(LOG_BLKRW, "(unsigned long*)bv_buf[1]: 0x%lx\n",
					((unsigned long *)bv_buf)[1]);
			stc_bio_kunmap_atomic(bv_buf);
		}
#endif
	} else {
		*bp->bio_error = BIO_ERROR(bio);
	}
	atomic_dec(bp->numbios);
	if (bp->buffer)
		__free_page(bp->page);
	kfree(bp);
	bio_put(bio);
}

/*
 * Create and submit a read bio from the parameters provided
 *
 * Bios are limited to being less than PAGE_SIZE in length unless buffer is a struct bio *.
 *
 * @param[in] bdev the block device to read from
 * @param[in] sector the sector to start reading from
 * @param[out] buffer the buffer to copy the result into
 * @param[in] length the size of the buffer
 * @param[in,out] numbios pointer to the number of bios that will be waited for
 * @param[in] bio_error
 * @param[in] buff_type Is buffer a void *, or a struct page *, or a struct bio *?
 * @returns  Zero if submitted; nonzero if failed to submit and numbios is decremented.
 */
int blk_read(struct block_device *bdev, sector_t sector, void *buffer, unsigned int length,
		atomic_t *numbios, int *bio_error, int buff_type)
{
	int retval = 0;
	struct page *page = NULL;
	struct bio *bio;
	struct blk_priv *bp;

	if (buff_type < BLK_BUFF_PAGE || buff_type > BLK_BUFF_BIO ||
			(buff_type != BLK_BUFF_BIO && length > PAGE_SIZE)) {
		retval = -EINVAL;
		goto error;
	}
	stcvsm_debug(LOG_BLKRW, "sector %llu, buffer %p, length %u, buff_type %d\n",
		(unsigned long long)sector, buffer, length, buff_type);
	/* Allocate the Bio */
	if (buff_type == BLK_BUFF_BIO)
		bio = (struct bio *)buffer;
	else
		bio = bio_alloc(GFP_NOIO, 1);
	if (!bio) {
		retval = -ENOMEM;
		goto error;
	}
	bio->bi_bdev = bdev;
	stc_bio_set_op(bio, READ);
	/* Allocate the Page */
	if (buff_type == BLK_BUFF_BIO) {
		if (bio->bi_vcnt == 0) {
			retval = -EINVAL;
			goto error_bio;
		}
	} else {
		if (buff_type == BLK_BUFF_MEM)
			page = alloc_page(GFP_KERNEL);
		else if (buff_type == BLK_BUFF_PAGE)
			page = (struct page *)buffer;
		if (!page) {
			retval = -ENOMEM;
			goto error;
		}
	}
	/* Add the new page to the bio */
	if (buff_type != BLK_BUFF_BIO) {
		/* bio_length needs to be the smallest multiple of SECTOR_SIZE that is >= length */
		unsigned int bio_length = DIV_ROUND_UP(length, SECTOR_SIZE) * SECTOR_SIZE;

		if (!bio_add_page(bio, page, bio_length, 0)) {
			retval = -EIO;
			stcvsm_err(LOG_BLKRW, "bio_add_page failed [%u|%u]\n", length, bio_length);
			goto error_page;
		}
	}
	/* Setup private data for the bio to be used in the endio function */
	bp = kmalloc(sizeof(*bp), GFP_KERNEL);
	if (!bp) {
		retval = -ENOMEM;
		goto error_page;
	}
	if (buff_type != BLK_BUFF_BIO) {  /* Should be set before coming in */
		bio_sector(bio) = sector;
	}
	bp->length = length;
	bio->bi_end_io = blk_read_endio;
	bio->bi_private = bp;
	bp->numbios = numbios;
	bp->page = page;
	bp->sector = sector;
	bp->type = buff_type;
	if (buff_type == BLK_BUFF_MEM)
		bp->buffer = buffer;
	else
		bp->buffer = NULL;
	bp->bio_error = bio_error;
	stcvsm_debug(LOG_BLKRW, "Submiting bio for sector %llu sectors %u\n",
			(unsigned long long)bio_sector(bio), bio_sectors(bio));
#ifdef DEBUG
	BUG_ON(bio_data_dir(bio) == WRITE);
#endif
	stc_submit_bio(bio);
	goto out;
error_page:
	if (buff_type == BLK_BUFF_MEM)
		__free_page(page);
error_bio:
	bio_put(bio);
error:
	atomic_dec(numbios);
	*bio_error = retval;
out:
	return retval;
}

/*
 * Wait for the read to complete
 *
 * @param[in] numbios
 * @returns
 */
void blk_read_wait(atomic_t *numbios)
{
	while (atomic_read(numbios) > 0) {
		/*
		 * stcvsm_debug(LOG_BLKRW, "%d\n", atomic_read(numbios));
		 */
		msleep(1);
	}
}

#endif  /* WITH_EXT2BITMAP */
