/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/* Block Device Based storage for COW pages */

#include <linux/slab.h>
#include <linux/bio.h>
#include <linux/delay.h>
#include <linux/backing-dev.h>

#include "cow.h"
#include "cow_range.h"
#include "debug.h"
#include "uassert.h"
#include "portability.h"
#include "range_table.h"

#define COW_BDEV_INDEX_MASK	0x00000000ffffffffULL
/* has a home on disk (COW_BDEV_INDEX_MASK is meaningful) */
#define COW_BDEV_ONDISK_MASK	0x0000000100000000ULL
/* is cow->pages and cow->lru usable (does not imply paged in!) */
#define COW_BDEV_OWNER_MASK	0x0000000200000000ULL
#define COW_BDEV_INFLIGHT_MASK	0x0000000400000000ULL

static void cow_bdev_flush(struct wq_work *work);

struct cow_ops cow_bdev_ops;
struct cow_bdev_tier {
	struct cow_tier;

	struct block_device *bdev;

	int sameq;
	struct list_head queued;	/* waiting to be flushed */
	int num_queued;
	atomic_t num_flushing;
	struct wq_work w;

	int is_freespace;

	/* COW_BDEV_DEDICATED */
	unsigned long next_index; /* index of next cow-sized chunk to be allocated */
	sector_t max_index;

	/* COW_BDEV_FREESPACE */
	struct rb_root owner;
	struct range_page_table *freespc;
	struct lookaside_list *freespc_lookaside;
	unsigned long relocations;
};

struct cow_bdev_read_job {
	atomic_t num_bios;
	struct cow_bdev_tier *btier;
	struct cow_buffer *cow;
	int error;
	cow_callback cb;
};

struct cow_bdev_write_job;
typedef void (*cow_bdev_write_cb)(struct cow_bdev_write_job *job);
struct cow_bdev_write_job {
	atomic_t num_bios;
	struct cow_bdev_tier *btier;
	struct cow_buffer *cow;
	int error;
	cow_bdev_write_cb cb;
};

#if 0
static struct cow_buffer *get_owner(struct cow_bdev_tier *btier, unsigned long index)
{
	return cow_search_range(&btier->owner, index);
}

{
	/* If cow->index is known to be completely freespace, short circuit the COW */
	if (btier->freespc &&
			range_check_for_dense_range(btier->freespc, cow->index, cow->index+1)) {
		struct cow_buffer *owner_cow;

		/* Since it's freespace, any cow already remapped here must move. */
		owner_cow = get_owner(btier, cow->index);
		if (owner_cow)
			;
	}
}
#endif

static void set_owner(struct cow_bdev_tier *btier, unsigned long index, struct cow_buffer *cow)
{
	cow_add_range(&btier->owner, index, cow);
}

static int cow_bdev_remap(struct cow_bdev_tier *btier, unsigned long *remapped)
{
	if (btier->is_freespace) {
		/* For now, the allocator is very simple; just scans backwards from the end, under
		 * the unsubstantiated belief that filesystems will typically allocate going
		 * forwards.  Hopefully a trivial yet non-pathological algorithm.
		 */
		struct range_ref range;

		range_last(btier->freespc, &range);
		if (range.Start == range.End)
			return -ENOSPC;
		if (!range_delete(btier->freespc, range.End-1, range.End))
			return -ENOMEM;
		*remapped = range.End-1;
		/* set_owner(btier, *remapped, cow); */
	} else {
		if (btier->next_index >= btier->max_index)
			return -ENOSPC;
		*remapped = btier->next_index++;
	}
	return 0;
}

static void cow_bdev_unmap(struct cow_bdev_tier *btier, unsigned long remapped)
{
	/* If possible, don't waste space on error */
	if (btier->is_freespace) {
		set_owner(btier, remapped, NULL);
		range_insert(btier->freespc, remapped, remapped + 1);
	} else {
		if (remapped + 1 == btier->next_index)
			btier->next_index--;
	}
}

static void cow_bdev_write_callback(struct cow_bdev_write_job *job)
{
	struct cow_bdev_tier *btier = job->btier;
	struct cow_buffer *cow = job->cow;
	struct page *pages[PAGES_PER_COW];
	int owned;

	cstor_lock_irqsave(btier->cstor);

	/* Clear in-flight flag:  you can no longer steal my pages, and if pages get re-pushed,
	 * they must be freed elsewhere.
	 */
	cow->tier_private[btier->depth] &= ~COW_BDEV_INFLIGHT_MASK;

	if (!job->error) {
		/* Did commit to disk, regardless whether still owned */
		cow->tier_private[btier->depth] |= COW_BDEV_ONDISK_MASK;
	}
	owned = !!(cow->tier_private[btier->depth] & COW_BDEV_OWNER_MASK);
	if (owned) {
		unsigned int i;
		/* Still own, so pages are my responsibility to free.  Move them aside so a
		 * subsequent pull can re-alloc in the cow.
		 */
		for (i = 0; i < PAGES_PER_COW; ++i) {
			pages[i] = cow->pages[i];
			cow->pages[i] = NULL;
		}

		/* Freeing the pages, so any error is fatal to this cow */
		if (job->error) {
			cow_state_trans(cow, COW_ERROR);
			cow->error = job->error;
		}
	}
	cstor_unlock_irqrestore(btier->cstor);

	stcvsm_debug(LOG_COWS, "stored cow %u: %d\n", (unsigned int)cow->index, job->error);
	if (owned)
		cow_free_pages(btier->cstor, pages);
	atomic_dec(&btier->num_flushing);

	/* Wake anyone waiting for freed memory.  Even if this cow was reclaimed, a waiter may
	 * need to wake up and re-evaluate what can be flushed.
	 */
	wake_up_all(&btier->cstor->waiters);
}

static BIO_END_IO_PROTO(cow_bdev_read_endio, bio)
{
	struct cow_bdev_read_job *job = bio->bi_private;

	if (ERROR_INJECT(ERRINJ_COW_IO))
		BIO_ERROR(bio) = -EIO;
	if (BIO_ERROR(bio))
		job->error = BIO_ERROR(bio);
	if (atomic_dec_and_test(&job->num_bios)) {
		if (job->error) {
			atomic_inc(&job->btier->errors);

			/* Until the read IO finishes or fails, the cow is PAGINGIN and not being
			 * used.  Safe to free pages.
			 */
			UASSERT(job->cow->state == COW_PAGINGIN, "COW state not PAGINGIN");
			cow_free_pages(job->btier->cstor, job->cow->pages);
		}
		job->cb((struct cow_tier *)job->btier, job->cow, job->error);
		kfree(job);
	}
	bio_put(bio);
}

BIO_END_IO_PROTO(cow_bdev_write_endio, bio)
{
	struct cow_bdev_write_job *job = bio->bi_private;

	if (ERROR_INJECT(ERRINJ_COW_IO))
		BIO_ERROR(bio) = -EIO;
	if (BIO_ERROR(bio))
		job->error = BIO_ERROR(bio);
	if (atomic_dec_and_test(&job->num_bios)) {
		if (job->error)
			atomic_inc(&job->btier->errors);
		job->cb(job);
		kfree(job);
	}
	bio_put(bio);
}

static int cow_bdev_read(struct cow_bdev_tier *btier, struct cow_buffer *cow, cow_callback cb)
{
	int retval;
	struct bio *bios;
	struct cow_bdev_read_job *job;
	struct blk_plug plug;

	if (ERROR_INJECT(ERRINJ_ALLOC))
		job = NULL;
	else
		job = kmalloc(sizeof(*job), GFP_NOFS);
	if (!job)
		return -ENOMEM;
	job->btier = btier;
	job->cow = cow;
	job->cb = cb;
	job->error = 0;

	blk_start_plug(&plug);
	retval = cow_pack_bios(btier->cstor, cow, cow->tier_private[btier->depth] & COW_BDEV_INDEX_MASK,
			btier->bdev, &bios);
	if (retval > 0) {
		struct bio *bio;

		atomic_set(&job->num_bios, retval);
		retval = 0;
		for (bio = bios; bio; bio = bios) {
			bios = bio->bi_next;
			bio->bi_next = NULL;
			bio->bi_private = job;
			stc_bio_assign_disk(bio, btier->bdev);
			bio->bi_end_io = cow_bdev_read_endio;
			stc_submit_bio(bio);
		}
	}
	blk_finish_plug(&plug);
	if (retval)
		kfree(job);
	return retval;
}

static struct cow_tier *cow_bdev_init(struct cow_storage *cstor, va_list args)
{
	struct cow_bdev_tier *btier;
	unsigned int cow_size = cstor->sectors_per_cow << 9;
	unsigned int type = va_arg(args, unsigned int);
	dev_t dev;
	int err;

	if (type != COW_BDEV_DEDICATED && type != COW_BDEV_FREESPACE) {
		err = -EINVAL;
		goto out1;
	}

	btier = kzalloc(sizeof(struct cow_bdev_tier), GFP_KERNEL);
	if (!btier) {
		err = -ENOMEM;
		goto out1;
	}
	dev = va_arg(args, dev_t);
	/* A single holder can exclusively claim multiple times, so be sure to pass distinct
	 * holders (for example: tier, not cstor) at least until the cow_bdev knows how to be
	 * shared.  Would be more ideal to pass snapshot->sc_snap_bdev as the holder (since it's
	 * a disk) but it does not exist yet.
	 */
	btier->bdev = stcvsm_open_blkdev(dev, FMODE_READ|FMODE_WRITE|FMODE_EXCL, btier);
	if (IS_ERR(btier->bdev)) {
		err = PTR_ERR(btier->bdev);
		goto out2;
	}
	if (type == COW_BDEV_FREESPACE) {
		btier->is_freespace = -1;

		/* Should be ignored for freespace, but further insure there can be no "next" */
		btier->max_index = 0;
		btier->next_index = 0xffffffff;

		btier->owner = RB_ROOT;
		/* TODO:  who creates btier->freespc?  more efficient if caller passes it in
		 * correctly formed (with cow-sized blocks) but that seems ugly.
		 * ranges indicate free space, mapping to cow->index.
		 */
	} else {
		btier->sameq = va_arg(args, int);
		btier->max_index = i_size_read(btier->bdev->bd_inode);
		do_div(btier->max_index, cow_size);
	}

	INIT_LIST_HEAD(&btier->queued);
	btier->w.fn = cow_bdev_flush;
	btier->ops = &cow_bdev_ops;
	btier->cstor = cstor;
	btier->max_kb = btier->max_index * cow_size / 1024;
	stcvsm_info(LOG_COWS, "(%pg) %lu kb\n", btier->bdev, btier->max_kb);

	return (struct cow_tier *)btier;

out2:
	kfree(btier);
out1:
	return ERR_PTR(err);
}

static void cow_bdev_flush(struct wq_work *work)
{
	struct cow_bdev_tier *btier = container_of(work, struct cow_bdev_tier, w);
	struct cow_storage *cstor = btier->cstor;
	struct cow_buffer *cow;
	unsigned long remapped = (unsigned long)-1;
	int i, error;
	struct cow_bdev_write_job *job = NULL;
	int free_pages;
	struct page *pages[PAGES_PER_COW];
	struct blk_plug plug;

	blk_start_plug(&plug);

	while (1) {
		struct bio *bios;

		if (!job && !ERROR_INJECT(ERRINJ_ALLOC))
			job = kmalloc(sizeof(*job), GFP_KERNEL);
		if (!job)
			break;

		/* Pull a cow off the list.  While writing to disk, leave the pages as-is in case
		 * the cow is pulled.
		 *
		 * NOTE  This does not support another tier pulling and then freeing the pages, but
		 * none does that yet.  If necessary, obtain another reference on the pages while
		 * they are being flushed to disk, and hold the array of pages on the job.
		 */
		cstor_lock_irq(cstor);
		if (list_empty(&btier->queued)) {
			cstor_unlock_irq(cstor);
			kfree(job);
			break;
		}
		cow = list_entry(btier->queued.next, struct cow_buffer, lru);
		list_del_init(&cow->lru);
		btier->num_queued--;
		UASSERT(cow->tier_private[btier->depth] & COW_BDEV_OWNER_MASK, "OWNER mask not set");

		if (cstor->shutdown || cow->tier_private[btier->depth] & COW_BDEV_ONDISK_MASK) {
			free_pages = 1;
			for (i = 0; i < PAGES_PER_COW; ++i) {
				pages[i] = cow->pages[i];
				cow->pages[i] = NULL;
			}
		} else {
			free_pages = 0;
			cow->tier_private[btier->depth] |= COW_BDEV_INFLIGHT_MASK;
		}
		cstor_unlock_irq(cstor);

		if (free_pages) {
			cow_free_pages(cstor, pages);
			continue;
		}

		/* Past this point, cow_bdev_write_callback is guaranteed to be called, which will
		 * do the wake.
		 */
		atomic_inc(&btier->num_flushing);

		job->btier = btier;
		job->cow = cow;
		job->cb = cow_bdev_write_callback;
		job->error = cow_bdev_remap(btier, &remapped);
		if (job->error) {
			stcvsm_debug(LOG_COWS, "failed to remap %u: %d\n", (unsigned int)cow->index,
					job->error);
			atomic_inc(&btier->errors);
			cow_bdev_write_callback(job);
			continue;
		}
		stcvsm_debug(LOG_COWS, "storing cow %u -> %lu\n", (unsigned int)cow->index, remapped);

		cow->tier_private[btier->depth] |= remapped;
		error = cow_pack_bios(btier->cstor, cow, remapped, btier->bdev, &bios);
		if (error > 0) {
			struct bio *bio;

			atomic_set(&job->num_bios, error);
			error = 0;
			for (bio = bios; bio; bio = bios) {
				bios = bio->bi_next;
				bio->bi_next = NULL;

				bio_set_prio(bio, IOPRIO_PRIO_VALUE(IOPRIO_CLASS_RT, 0));
				bio->bi_private = job;
				stc_bio_assign_disk(bio, btier->bdev);
				bio->bi_end_io = cow_bdev_write_endio;
				stc_bio_set_op(bio, WRITE);

				/* When COW-ing to the snapshotted disk, it is critical to not
				 * submit IO back to the top of the stack with submit_bio().  Doing
				 * so will put the COW behind other IO that may be waiting on it.
				 * Send the COW down, ahead of the triggering IO, by emulating a
				 * minimal amount of what occurs between submit_bio() and
				 * submit_bio_fn(), then calling the original submit_bio_fn()
				 * directly.  Memory spikes are not a concern, since the request
				 * queue does throttling.  Even though this submission is done from
				 * another context, the request queue will fill and throttle back
				 * the original writer.
				 */

				if (btier->sameq) {
					stcvsm_direct_submit_bio(bio, btier->bdev);
				} else {
					stc_submit_bio(bio);
				}
			}
		}

		if (error == 0) {
			btier->kb = (remapped + 1) << (cstor->sector_shift - 1);
			job = NULL;
		} else {
			job->error = error;
			cow_bdev_unmap(btier, remapped);
			atomic_inc(&btier->errors);
			cow_bdev_write_callback(job);
		}
	}

	blk_finish_plug(&plug);

	cstor_lock_irq(cstor);
	/* Either note that we're done flushing, or put back on end of workqueue */
	if (list_empty(&btier->queued))
		btier->w.data = 0;
	else
		wq_insert_work(&btier->cstor->flush_wq, &btier->w);
	cstor_unlock_irq(cstor);

	wake_up_all(&btier->cstor->waiters);
}

static void cow_bdev_push(struct cow_tier *tier, struct cow_buffer *cow)
{
	struct cow_bdev_tier *btier = (struct cow_bdev_tier *)tier;

	btier->stores++;

	/* Already have a copy? */
	if (cow->tier_private[btier->depth] & COW_BDEV_INFLIGHT_MASK) {
		stcvsm_trace(LOG_COWS, "already in flight %u\n", (unsigned int)cow->index);
		cow->tier_private[btier->depth] |= COW_BDEV_OWNER_MASK;
		/* pages are freed during endio */
		return;
	} else if (cow->tier_private[btier->depth] & COW_BDEV_ONDISK_MASK) {
		stcvsm_trace(LOG_COWS, "already on disk %u\n", (unsigned int)cow->index);
		cow->tier_private[btier->depth] |= COW_BDEV_OWNER_MASK;
		/* pages are freed in cow_bdev_flush */
	} else {
		UASSERT(!(cow->tier_private[btier->depth] & COW_BDEV_OWNER_MASK), "OWNER mask is set");
		cow->tier_private[btier->depth] = COW_BDEV_OWNER_MASK;
	}

	list_add_tail(&cow->lru, &btier->queued);
	btier->num_queued++;

	if (!btier->w.data) {
		btier->w.data = 1;  /* busy */
		wq_insert_work(&btier->cstor->flush_wq, &btier->w);
	}
}

static void cow_bdev_pull(struct cow_tier *tier, struct cow_buffer *cow, cow_callback cb)
{
	struct cow_bdev_tier *btier = (struct cow_bdev_tier *)tier;
	int owned, start_read = 0;

	cstor_lock_irq(btier->cstor);
	btier->loads++;
	owned = !!(cow->tier_private[btier->depth] & COW_BDEV_OWNER_MASK);
	if (owned) {
		cow->tier_private[btier->depth] &= ~COW_BDEV_OWNER_MASK;
		btier->hits++;
		if (list_empty(&cow->lru)) {
			if (cow->tier_private[btier->depth] & COW_BDEV_ONDISK_MASK)
				start_read = 1;
			/* else cow may be in-flight to disk, or errored. */
		} else {
			/* Still in memory!  Pull off list to steal and return. */
			list_del_init(&cow->lru);
		}
	}
	cstor_unlock_irq(btier->cstor);

	if (owned) {
		if (start_read) {
			int error;

			stcvsm_debug(LOG_COWS, "reading cow %u <- %lu\n", (unsigned int)cow->index,
					(unsigned long)(cow->tier_private[btier->depth] & COW_BDEV_INDEX_MASK));
			error = cow_alloc_pages(btier->cstor, cow->pages);
			if (error == 0) {
				error = cow_bdev_read(btier, cow, cb);
				if (error)
					cow_free_pages(btier->cstor, cow->pages);
			}
			if (error) {
				atomic_inc(&btier->errors);
				cb(tier, cow, error);
			}
		} else {
			stcvsm_debug(LOG_COWS, "found cow %u\n", (unsigned int)cow->index);
			cb(tier, cow, 0);
		}
	} else if (btier->next) {
		btier->next->ops->cow_pull(btier->next, cow, cb);
	} else {
		stcvsm_debug(LOG_COWS, "no cow %u\n", (unsigned int)cow->index);
		cb(tier, cow, -ENOENT);
	}
}

static void cow_bdev_wait(struct cow_tier *tier)
{
	struct cow_bdev_tier *btier = (struct cow_bdev_tier *)tier;

	/* Do throttling if the COW device does not share a request queue with
	 * the snapshotted device. */
	if (!btier->sameq) {
#ifdef HAVE_RQ_NR_CONGESTION
		struct request_queue *q = bdev_get_queue(btier->bdev);
#endif

		/* The limit of num_queued is rather arbitrary.
		 *
		 * "num_flushing" is in COWs, "nr_congestion_on" is in requests.  Roughly maps: a
		 * cow might be represented by a single request.
		 */
		wait_event(btier->cstor->waiters, btier->num_queued < 32
#ifdef HAVE_RQ_NR_CONGESTION
				&& atomic_read(&btier->num_flushing) <= q->nr_congestion_on
#else
				&& atomic_read(&btier->num_flushing) <= 32
#endif
		);
	}
}

static void cow_bdev_shutdown(struct cow_tier *tier)
{
	struct cow_bdev_tier *btier = (struct cow_bdev_tier *)tier;

	while (1) {
		int busy;
		cstor_lock_irq(tier->cstor);
		// flusher is busy, or flushed IO is still outstanding
		busy = btier->w.data || atomic_read(&btier->num_flushing);
		cstor_unlock_irq(tier->cstor);
		if (!busy)
			break;
		msleep(20);
	}

	if (btier->bdev) {
		stcvsm_close_blkdev(btier->bdev, FMODE_READ|FMODE_WRITE|FMODE_EXCL);
		btier->bdev = NULL;
	}
}

static void cow_bdev_log(struct cow_tier *tier, struct seq_file *s)
{
	struct cow_bdev_tier *btier = (struct cow_bdev_tier *)tier;
	dev_t dev = btier->bdev->bd_dev;

	seq_printf(s, "%d queued %u flushing to %u:%u ", btier->num_queued,
			atomic_read(&btier->num_flushing), MAJOR(dev), MINOR(dev));
	if (btier->is_freespace)
		seq_printf(s, "freespace  %lu relocations", btier->relocations);
	else
		seq_puts(s, "dedicated");
}

static void cow_bdev_deinit(struct cow_tier *tier)
{
	struct cow_bdev_tier *btier = (struct cow_bdev_tier *)tier;
	struct list_head *pos;

	list_for_each(pos, &btier->queued) {
		struct cow_buffer *cow = list_entry(pos, struct cow_buffer, lru);

		cow_free_pages(tier->cstor, cow->pages);
	}

	if (btier->is_freespace) {
		range_cleanup(btier->freespc);
		kfree(btier->freespc);
		kfree(btier->freespc_lookaside);
	}

	kfree(btier);
}

struct cow_ops cow_bdev_ops = {
	.cow_init     = cow_bdev_init,
	.cow_push     = cow_bdev_push,
	.cow_pull     = cow_bdev_pull,
	.cow_log      = cow_bdev_log,
	.cow_wait     = cow_bdev_wait,
	.cow_shutdown = cow_bdev_shutdown,
	.cow_deinit   = cow_bdev_deinit,
	.cow_name     = "bdev"
};

