/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * The purpose of these functions is to construct the bitmap for the ext2 file system
 * We need to:
 * 1) Read superblock
 * 2) Determine number of Block Groups
 * 3) Read bitmap from each Block Group
 */
#include <linux/bio.h>
#include <linux/ext2_fs.h>

#include "blkrw.h"
#include "debug.h"
#include "ext2_bitmap.h"
#include "portability.h"

#ifdef WITH_EXT2BITMAP
#define SECTOR_SIZE 512

int ext2_create_bitmap(struct block_device *bdev, unsigned long *bitmap, unsigned long bits_in_map,
		unsigned long _block_size)
{
	/* Read super block from sector 2 (byte 1024) */
	int retval = -1;
	struct ext2_super_block *sb;
	struct ext2_group_desc *bgd;
	off_t block_group_count;
	off_t sectors_per_block;
	off_t block_size;
	int i;
	int error;
	int bio_error;
	atomic_t numbios;

	sb = kmalloc(sizeof(struct ext2_super_block), GFP_KERNEL);
	if (sb == NULL)
		return -1;

	bio_error = 0;
	atomic_set(&numbios, 1);
	error = blk_read(bdev, 2, sb, sizeof(*sb), &numbios, &bio_error, BLK_BUFF_MEM);
	if (error) {
		stcvsm_debug(LOG_FS, "blk_read for superblock return %d\n", error);
		goto out;
	}
	blk_read_wait(&numbios);
	if (bio_error) {
		stcvsm_debug(LOG_FS, "Error reading superblock: %d\n", bio_error);
		goto out;
	}
	e2sb_debug(sb);
	BUG_ON(DIV_ROUND_UP(sb->s_blocks_count, sb->s_blocks_per_group) !=
		DIV_ROUND_UP(sb->s_inodes_count, sb->s_inodes_per_group));
	block_size = 1024 << sb->s_log_block_size;
	BUG_ON(_block_size != block_size);
	sectors_per_block = block_size / SECTOR_SIZE;
	block_group_count = DIV_ROUND_UP(sb->s_blocks_count, sb->s_blocks_per_group);
	/* TODO: I'm not ready to deal with reads > PAGE_SIZE yet */
	if (block_group_count * sizeof(*bgd) > PAGE_SIZE) {
		stcvsm_debug(LOG_FS, "blk_read not ready to handle reads > PAGE_SIZE yet\n");
		goto out;
	}
	stcvsm_debug(LOG_FS, "Bits_in_map: %lu block_count: %u\n", bits_in_map, sb->s_blocks_count);
	stcvsm_debug(LOG_FS, "Block group count: %lu, Block Group size %lu\n",
			block_group_count, (long) sizeof(*bgd));

	bgd = kmalloc_array(block_group_count, sizeof(*bgd), GFP_KERNEL);
	if (bgd == NULL)
		goto out;
	bio_error = 0;
	atomic_set(&numbios, 1);
	error = blk_read(bdev, (sb->s_first_data_block + 1) * sectors_per_block, bgd,
			sizeof(*bgd) * block_group_count, &numbios, &bio_error, BLK_BUFF_MEM);
	if (error) {
		stcvsm_debug(LOG_FS, "blk_read for group descriptor table return %d\n", error);
		goto out_bgd;
	}
	blk_read_wait(&numbios);
	bio_error = 0;
	atomic_set(&numbios, block_group_count);
	for (i = 0; i < block_group_count; i++) {
		sector_t sector = bgd[i].bg_block_bitmap * sectors_per_block;
		unsigned long *start_bit = bitmap + i * block_size / 8;
		unsigned long length = sb->s_blocks_per_group / 8;
		/*
		 * if (i == block_group_count - 1)
		 *	length = (bits_in_map % sb->s_blocks_per_group) / 8;
		 */
		e2bgd_debug(i, &bgd[i]);
		error = blk_read(bdev, sector, start_bit, length, &numbios, &bio_error,
				BLK_BUFF_MEM);
		if (error) {
			stcvsm_debug(LOG_FS, "blk_read for sector %lu failed with error %d\n",
					bgd[i].bg_block_bitmap * sectors_per_block, error);
		}
	}
	blk_read_wait(&numbios);
	retval = 0;

out_bgd:
	kfree(bgd);

out:
	kfree(sb);

	return retval;
}

/**Print debug info from superblock
 * TODO: See if we can match the output of dumpe2fs
 */
void e2sb_debug(struct ext2_super_block *sb)
{
	stcvsm_debug(LOG_FS, "Output should match dumpe2fs\n");
	stcvsm_debug(LOG_FS, "Filesystem volume name:   %s\n", sb->s_volume_name);
	stcvsm_debug(LOG_FS, "Last mounted on:          %s\n", sb->s_last_mounted);
	/* Needs conversion to UUID: */
	stcvsm_debug(LOG_FS, "Filesystem UUID:          %s\n", sb->s_uuid);
	stcvsm_debug(LOG_FS, "Filesystem magic number:  0x%X\n", sb->s_magic);
	stcvsm_debug(LOG_FS, "Filesystem revision #:    %u\n", sb->s_rev_level);
	stcvsm_debug(LOG_FS, "Filesystem features:      %u\n", sb->s_feature_compat);
	stcvsm_debug(LOG_FS, "Filesystem flags:         %u\n", sb->s_feature_compat);
	stcvsm_debug(LOG_FS, "Default mount options:    %s\n", "??");
	stcvsm_debug(LOG_FS, "Filesystem state:         %u\n", sb->s_state);
	stcvsm_debug(LOG_FS, "Errors behavior:          %u\n", sb->s_errors);
	stcvsm_debug(LOG_FS, "Filesystem OS type:       %u\n", sb->s_creator_os);
	stcvsm_debug(LOG_FS, "Inode count:              %u\n", sb->s_inodes_count);
	stcvsm_debug(LOG_FS, "Block count:              %u\n", sb->s_blocks_count);
	stcvsm_debug(LOG_FS, "Reserved block count:     %u\n", sb->s_r_blocks_count);
	stcvsm_debug(LOG_FS, "Free blocks:              %u\n", sb->s_free_blocks_count);
	stcvsm_debug(LOG_FS, "Free inodes:              %u\n", sb->s_free_inodes_count);
	stcvsm_debug(LOG_FS, "First block:              %u\n", sb->s_first_data_block);
	stcvsm_debug(LOG_FS, "Block size:               %u\n", 1024 << sb->s_log_block_size);
	stcvsm_debug(LOG_FS, "Fragment size:            %u\n", 1024 << sb->s_log_frag_size);
	stcvsm_debug(LOG_FS, "Reserved GDT blocks:      %s\n", "??");
	stcvsm_debug(LOG_FS, "Blocks per group:         %u\n", sb->s_blocks_per_group);
	stcvsm_debug(LOG_FS, "Fragments per group:      %u\n", sb->s_frags_per_group);
	stcvsm_debug(LOG_FS, "Inodes per group:         %u\n", sb->s_inodes_per_group);
	stcvsm_debug(LOG_FS, "Inode blocks per group:   %s\n", "needs calculation");
	stcvsm_debug(LOG_FS, "Flex block group size:    %s\n", "??");
	stcvsm_debug(LOG_FS, "Filesystem created:       %s\n", "??");
	stcvsm_debug(LOG_FS, "Last mount time:          %u\n", sb->s_mtime);
	stcvsm_debug(LOG_FS, "Last write time:          %u\n", sb->s_wtime);
	stcvsm_debug(LOG_FS, "Mount count:              %u\n", sb->s_mtime);
	stcvsm_debug(LOG_FS, "Maximum mount count:      %u\n", sb->s_max_mnt_count);
	stcvsm_debug(LOG_FS, "Last checked:             %u\n", sb->s_lastcheck);
	stcvsm_debug(LOG_FS, "Check interval:           %u\n", sb->s_checkinterval);
	stcvsm_debug(LOG_FS, "Next check after:         %s\n", "needs calculation");
	stcvsm_debug(LOG_FS, "Lifetime writes:          %s\n", "??");
	stcvsm_debug(LOG_FS, "Reserved blocks uid:      %u\n", sb->s_def_resuid);
	stcvsm_debug(LOG_FS, "Reserved blocks gid:      %u\n", sb->s_def_resgid);
	stcvsm_debug(LOG_FS, "First inode:              %u\n", sb->s_first_ino);
	stcvsm_debug(LOG_FS, "Inode size:               %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal inode:            %u\n", sb->s_journal_inum);
	stcvsm_debug(LOG_FS, "Default directory hash:   %s\n", "??");
	stcvsm_debug(LOG_FS, "Directory Hash Seed:      %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal backup:           %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal features:         %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal size:             %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal length:           %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal sequence:         %s\n", "??");
	stcvsm_debug(LOG_FS, "Journal start:            %s\n", "??");
}

/*
 * Print debug info about a Block Group Descriptor
 */
void e2bgd_debug(unsigned int index, struct ext2_group_desc *bgd)
{
	/*
	 * Group 0: (Blocks 1-8192) [ITABLE_ZEROED]
	 * Checksum 0x3c1d, unused inodes 1846
	 * Primary superblock at 1, Group descriptors at 2-3
	 * Reserved GDT blocks at 4-259
	 * Block bitmap at 260 (+259), Inode bitmap at 276 (+275)
	 * Inode table at 292-547 (+291)
	 * 1047 free blocks, 1846 free inodes, 2 directories, 1846 unused inodes
	 * Free blocks: 7146-8192
	 * Free inodes: 203-2048
	 * Group 1: (Blocks 8193-16384) [INODE_UNINIT, ITABLE_ZEROED]
	 * Checksum 0x2b50, unused inodes 2048
	 * Backup superblock at 8193, Group descriptors at 8194-8195
	 * Reserved GDT blocks at 8196-8451
	 * Block bitmap at 261 (+4294959364), Inode bitmap at 277 (+4294959380)
	 * Inode table at 548-803 (+4294959651)
	 * 0 free blocks, 2048 free inodes, 0 directories, 2048 unused inodes
	 * Free blocks:
	 * Free inodes: 2049-4096
	 * Group 2: (Blocks 16385-24576) [INODE_UNINIT, ITABLE_ZEROED]
	 * Checksum 0x855d, unused inodes 2048
	 * Block bitmap at 262 (+4294951173), Inode bitmap at 278 (+4294951189)
	 * Inode table at 804-1059 (+4294951715)
	 * 321 free blocks, 2048 free inodes, 0 directories, 2048 unused inodes
	 * Free blocks: 24256-24576
	 * Free inodes: 4097-6144
	 */
	stcvsm_debug(LOG_FS, "%3u %7u %7u %7u %4u %4u %4u\n",
			index,
			bgd->bg_block_bitmap,
			bgd->bg_inode_bitmap,
			bgd->bg_inode_table,
			bgd->bg_free_blocks_count,
			bgd->bg_free_inodes_count,
			bgd->bg_used_dirs_count);
}

#endif
