/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCVSM_FILTER_H
#define STCVSM_FILTER_H

#include <linux/blkdev.h>
#include <linux/wait.h>
#include <linux/spinlock.h>

#include "portability.h"

/* NOTE: 16 is an artificial number usually used by scsi drivers as the maximum number of
 * partitions the driver supports */
#define MAX_SNAPS_PER_FILTER  15
#define MAX_VDIFFS_PER_FILTER 16

struct snapshot_context;
struct vdiff_context;

/* All IO must touch this, so probably worth minimizing size and keeping the hottest
 * data in the first cache line.
 */
struct block_device_filter {
	atomic_t bdf_refs;
	atomic_t writes;
	wait_queue_head_t flush_waiters;

	short num_snapshots;
	unsigned short snapshot_mask; /* allocated snapshot minor numbers */
	short num_vdiffs;
	short idx;

#ifdef HAVE_BDOPS_SUBMIT_BIO
	const struct block_device_operations *bdf_orig_fops;
	struct block_device_operations bdf_fops;
#else
	submit_bio_fn *bdf_orig_reqfn;
#endif
	submit_bio_fn *bdf_reqfn;

	/* Array of context pointers */
	struct vdiff_context *(vdiffs[MAX_VDIFFS_PER_FILTER]);
	struct snapshot_context *(snapshots[MAX_SNAPS_PER_FILTER]);

	struct block_device *bdf_bdev;
	struct gendisk *bdf_disk;

	int quiesced;
};

static inline spinlock_t *bdf_lock(struct block_device_filter *bdfilter)
{
	/* share a single spinlock for the entire filter */
	return &bdfilter->flush_waiters.lock;
}

void init_filters(void);
int quiesce_filters(int quiet);
void shutdown_filters(void);
void activate_filter(struct block_device_filter *bdfilter);
void deactivate_filter(struct block_device_filter *bdfilter);
int attach_filter_vdiff(struct block_device_filter *bdfilter, struct vdiff_context *vdiff);
void detach_filter_vdiff(struct block_device_filter *bdfilter, struct vdiff_context *vdiff);
struct vdiff_context *get_filter_vdiff(struct block_device_filter *bdfilter, dev_t dev);
MAKE_REQUEST_RETTYPE block_device_filter_reqfn(struct request_queue *q, struct bio *bio,
		struct block_device_filter *bdfilter);
struct block_device *get_parent_bdev(struct block_device *bdev);
struct block_device_filter *get_existing_filter(dev_t dev);
/** Gets the filter structure, optionally creating it.
 * @return Filter pointer, or NULL if not creating and none exists, or ERR_PTR.
 */
struct block_device_filter *get_filter(struct block_device *parent_bdev, bool get_new);
void dec_filtered_writes(struct block_device_filter *bdfilter);
void put_filter(struct block_device_filter *bdfilter);
submit_bio_fn *bdf_get_orig_submit_bio(struct block_device_filter *bdfilter);
bool bdf_is_active(struct block_device_filter *bdfilter);

#endif
