/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * Snapshot Driver for Linux systems
 *
 * Purpose: Implements a Linux kernel block device snapshot interface.
 * The request function for an existing block device is replaced with
 * one provided here that holds all writes until this driver reads the
 * existing data from the block device.  The data is stored
 * *somewhere* so that an associated "snapshot" block device can use
 * the copied data instead of reading directly from disk.
 *
 */

#include <linux/init.h>
#include <linux/module.h>

#include "debug.h"
#include "filter.h"
#include "import.h"
#include "portability.h"
#include "proc.h"
#include "stcvsm_version.h"
#include "stcvsmctl.h"
#include "vdiff.h"
#include "workqueue.h"

MODULE_LICENSE("GPL");
MODULE_VERSION(STCVSM_VERSION);

int stcvsm_major;
module_param(stcvsm_major, int, 0);
MODULE_PARM_DESC(stcvsm_major, "Major number of snapshots");

static unsigned long export_addr;
module_param(export_addr, ulong, 0);
MODULE_PARM_DESC(export_addr, "Address of stcvsm_export_vdiff, for dynamic driver update");

extern int deny_export;
module_param(deny_export, int, 0);
MODULE_PARM_DESC(deny_export, "Disables dynamic driver update");

static int __init stcvsm_init(void)
{
	int err;

	stcvsm_debug(LOG_MODULE, "----- ----- ----- stcvsm_init ----- ----- -----\n");

	init_filters();

	/* setup snapshot block device */
	mutex_init(&stcvsm_freeze_lock);
	stcvsm_major = register_blkdev(stcvsm_major, "stcvsm");
	if (stcvsm_major < 0) {
		stcvsm_err(LOG_MODULE, "unable to get major number\n");
		err = stcvsm_major;
		goto err1;
	}

	/* setup /proc directory entries */
	err = stcvsm_init_proc();
	if (err) {
		stcvsm_err(LOG_MODULE, "init_proc failed with %d\n", err);
		goto err2;
	}

	/* setup character control device */
	err = stcvsmctl_init();
	if (err) {
		stcvsm_err(LOG_MODULE, "chrdev failed with %d\n", err);
		goto err3;
	}

	if (export_addr != 0) {
		void *(*stcvsm_export_vdiff)(unsigned int cmd);

		stcvsm_export_vdiff = (void *(*)(unsigned int cmd))export_addr;
		err = stcvsm_import_vdiff(stcvsm_export_vdiff);
		if (err) {
			stcvsm_err(LOG_MODULE, "Failed to import prior tracking state.  Retry or 'rmmod stcvsm_upgrade' to abort.\n");
			goto err3;
		}
	}

	wq_init(global_wq(), "stcvsm-wq");
	err = wq_start(global_wq());
	if (err) {
		stcvsm_err(LOG_MODULE, "Failed to start global work queue.");
		goto err4;
	}
	return 0;

err4:
	wq_deinit(global_wq());
err3:
	stcvsm_remove_proc();
err2:
	unregister_blkdev(stcvsm_major, "stcvsm");
err1:
	if (err)
		stcvsm_err(LOG_MODULE, "Unable to initialize module: %d\n", err);
	return err;
}

static void __exit stcvsm_exit(void)
{
	stcvsm_debug(LOG_MODULE, "----- ----- ----- stcvsm_exit ----- ----- -----\n");

	wq_stop(global_wq(), NULL, NULL);
	wq_deinit(global_wq());

	quiesce_filters(1);
	shutdown_filters();

	stcvsm_remove_proc();
	unregister_blkdev(stcvsm_major, "stcvsm");
	stcvsmctl_destroy();
}

module_init(stcvsm_init);
module_exit(stcvsm_exit);
