/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/* Page-based range table (compressed bitmap) implementation - 64bit ready */

#include "stcvsm_os.h"

#include "uassert.h"
#include "range_table.h"


/*
 *                                      Notes
 *
 *  - each page is an array of "walls", their count is always even (couples of "walls")
 *  - even "walls" (Walls[0], Walls[2] etc) are starts (inclusive) of in-bitmap ranges
 *  - odd "walls" (Walls[1], Walls[3] etc) are ends (non-inclusive) of in-bitmap ranges, and
 *      starts (inclusive) of holes
 *  - also, u64 Page->Base is added to URPVALUE "walls" to get the actual ranges
 *  - so, the bitmap ranges are:
 *      [Page->Base + Walls[0]..Page->Base + Walls[1])
 *      [Page->Base + Walls[2]..Page->Base + Walls[3])
 *  and so on
 */


/*  Inline functions */

/*
 * true if (Second - First) fits URPVALUE, false otherwise
 * The caller must guarantee Second >= First
 */
static inline bool range_diff_fits_URPValue(u64 First, u64 Second)
{
	if (Second <= (URPVALUE)(-1))
		/* First is <= Second, the difference surely fits URPVALUE */
		return true;
	/* We can subtract (URPVALUE)(-1) from Second without overflow */
	if ((Second - (URPVALUE)(-1)) <= First)
		/* First */
		return true;
	/* Does not fit */
	return false;
}

/*
 * Adds (URPVALUE)(-1) to Value
 * Safe, never wraps around (u64)(s64)(-1)
 */
static inline u64 range_plus_max_URPValue(u64 Value)
{
	/* Check for wraparound */
	if (Value >= ((u64)(s64)(-1) - (URPVALUE)(-1)))
		/* Maximum possible value */
		return (u64)(s64)(-1);
	/* Value is small enough, we can add */
	return Value + (URPVALUE)(-1);
}

#ifdef UNITTEST_CODE
u64 test_range_plus_max_URPValue(u64 Value)
{
	return range_plus_max_URPValue(Value);
}
#endif

/*
 * Functions internal for this file
 *
 * Checks whether it is possible to insert a part of the [Start..End) range to the head
 *  of Page
 * Returns End if not possible, otherwise, returns the minimum start value (>= Start)
 *  of the range to be inserted
 * The caller must guarantee that End < any value in the page
 * Does not update anything
 */
static u64 range_insert_to_page_head_check(struct page_of_ranges *Page,
					  u64 Start, u64 End);
/*
 * Actually inserts the range to the head of the page
 * The table is fully set up on return
 * Must only be called after range_insert_to_page_head_check, with the same End value
 * Start must be the return value of range_insert_to_page_head_check,
 *  and must never be == End
 */
static void range_insert_to_page_head_do(struct range_page_table *Table,
					struct page_of_ranges *Page,
					u64 Start, u64 End);
/*
 * Checks whether it is possible to insert a part of the [Start..End) range to the tail
 *  of Page
 * Returns Start if not possible, otherwise, returns the maximum end value (<= End)
 *  of the range to be inserted
 * Page can be NULL, this is properly handled
 * The caller must guarantee that Start > any value in the page
 * Does not update anything
 */
static u64 range_insert_to_page_tail_check(struct page_of_ranges *Page,
					  u64 Start, u64 End);
/*
 * Actually inserts the range to the tail of the page
 * The table is fully set up on return
 * Must only be called after range_insert_to_page_tail_check, with the same Start value
 * End must be the return value of range_insert_to_page_tail_check,
 *  and must never be == Start
 */
static void range_insert_to_page_tail_do(struct range_page_table *Table,
					struct page_of_ranges *Page,
					u64 Start, u64 End);
/*
 * Creates the chain of the preallocated pages long enough to cover [Start..End)
 * Returns NULL on allocation error or the pointer to the first page on success
 * The pages are single-linked via the first word in them, interpreted as next page pointer
 */
static void *range_page_chain_prealloc(struct range_page_table *Table,
					u64 Start,
					u64 End);
/*
 * Undo for range_page_chain_prealloc
 * ListHead is allowed to be NULL, no-op in this case
 */
static void range_page_chain_prefree(struct range_page_table *Table, void *ListHead);
/*
 * Fills the pages allocated by range_page_chain_prealloc to cover [Start..End)
 *  and attaches them to the tree
 * [Start..End) must be the same as for range_page_chain_prealloc
 */
static void range_page_chain_attach(struct range_page_table *Table,
				     void *ListHead,
				     u64 Start,
				     u64 End);
/*
 * Updates the page so it still describes the same ranges, but ->Base it set to be
 *  NewBase
 * On call, NewBase must be < ->Base
 * The caller must check that the last "wall" in the page will still fit URPVALUE
 *  after rebasing, this function has no such checks
 * Does not update the page's CRC
 */
static void page_of_ranges_rebase_down(struct page_of_ranges *Page, u64 NewBase);
/*
 * Splits Page it two, moving half of its items to the newly allocated page
 *  which is inserted next to Page in the tree
 * Page is still valid after this
 * Returns false on allocation error, no updates are done in this case
 */
static bool page_of_ranges_split(struct tree_table *Tree, struct page_of_ranges *Page);
/*
 * Tries to merge Page with neighbours
 * Does nothing and returns false if there is no need to merge
 * Returns true otherwise
 */
static bool page_of_ranges_try_to_merge(struct range_page_table *Table,
				     struct page_of_ranges *Page);
/*
 * Checks whether the Base value in Page is good enough for a merge, i.e.
 *  all ranges in NextPage can be represented with this base
 */
static bool page_of_ranges_is_base_ok_for_merge(struct page_of_ranges *Page,
					   struct page_of_ranges *NextPage);
/*
 * Merges NextPage to Page, deletes NextPage from Table after this
 * NextPage must be the next page after Page in Table
 */
static void page_of_ranges_merge(struct range_page_table *Table, struct page_of_ranges *Page,
				struct page_of_ranges *NextPage);
/*
 * Searches the tree of pages for given value
 * Flags is a bitmask of RANGE_SEARCH_xxx above
 * Returns the search result and the found page in *Page
 *  - first tree's page for Below
 *  - next page for Between
 *  - last tree's page (or NULL if the tree is empty) for Above
 * Handling of the start and end values of the page - are they Inside or not - is governed
 *  by Flags
 */
static enum enum_search_result page_of_ranges_find_page(struct tree_table *Tree,
						      u64 Value,
						      u32 Flags,
						      struct page_of_ranges **Page);
/*
 * Constructor routine which fills the page with initial data (Data must be
 *  struct page_of_ranges_template *) to ensure the proper comparison order in the tree from the
 *  very beginning
 * Must match PUT_TREE_CONSTRUCT_ROUTINE
 */
static void range_construct(void *Tree, void *Item, void *Data);
/*
 * Compares 2 pages on their Start values
 * Must match PUT_TREE_COMPARE_ROUTINE
 */
static s32 range_compare(void *Tree, void *First, void *Second);
/*
 * Inserts the new range of values to the page
 * Returns true usually, and false only if the page has no more room -
 *  the caller should split the page in this case
 * Updates the table's and the page's counts on success
 */
static bool page_of_ranges_insert(struct range_page_table *Table,
				 struct page_of_ranges *Page,
				 URPVALUE Start, URPVALUE End);
/*
 * Inserts the new range of values to the page
 * Returns true usually, and false only if the page has no more room -
 *  the caller should split the page in this case
 * On success, sets *ValuesCountDelta to the number of values added to the page
 */
static bool page_of_ranges_insert_internal(struct page_of_ranges *Page,
					 URPVALUE Start, URPVALUE End,
					 URPVALUE *ValuesCountDelta);
/*
 * Deletes the values from the page of ranges
 * Returns true usually, and false only in the rare case if deletion will need to add
 *  a hole and a second range to the page, and the page has no more room -
 *  the caller should split the page in this case
 * Updates the table's and the page's counts on success
 * Must never be called with Start >= largest page value or with End <= first page value
 */
static bool page_of_ranges_delete(struct range_page_table *Table,
				 struct page_of_ranges *Page,
				 URPVALUE Start, URPVALUE End);
/*
 * Deletes the values from the page of ranges
 * Returns true usually, and false only in the rare case if deletion will need to add
 *  a hole and a second range to the page, and the page has no more room -
 *  the caller should split the page in this case
 * On success, sets *ValuesCountDelta to the number of values deleted from the page
 * Must never be called with Start >= largest page value or with End <= first page value
 */
static bool page_of_ranges_delete_internal(struct page_of_ranges *Page,
					 URPVALUE Start, URPVALUE End,
					 URPVALUE *ValuesCountDelta);
/*
 * Performs a binary search over Array
 * Array has Size elements
 * The caller must guarantee that Size is at least 2, and that Array is ordered
 * Returns (u32)(-1) if all elements in Array are < Value
 * Otherwise, returns the index of the first element in Array which is >= Value
 */
static u32 page_of_ranges_bsearch(URPVALUE Array[], u32 Size, URPVALUE Value);


/*  Functions              */

/*
 * Initializes a nonpaged table
 */
void range_initialize_nonpaged(struct range_page_table *Table, enum tree_kind TreeKind,
			       struct lookaside_list *lookaside, u32 PoolTag)
{
	tree_initialize_npage(&(Table->Tree), TreeKind, RANGE_PAGE_SIZE, range_compare,
			range_construct, NULL, lookaside, PoolTag);
	Table->RangesCount = 0;
	Table->ValuesCount = 0;
}

/*
 * Cleans up a table, deleting all items and cleaning up the lookaside allocator
 * Both allocator and a table can be reused then
 */
void range_cleanup(struct range_page_table *Table)
{
	/* Clean the tree and the allocator */
	tree_cleanup(&(Table->Tree));

	/*  Reset the counts */
	Table->RangesCount = 0;
	Table->ValuesCount = 0;
}

/*
 * Inserts a range to the range page table
 * Returns false on allocation failure, the table is not updated in this case
 * The caller must guarantee Start < End
 */
bool range_insert(struct range_page_table *Table, u64 Start, u64 End)
{
	enum enum_search_result Result;
	struct page_of_ranges *Page;
	struct page_of_ranges *FirstPage;
	u64 FirstPageBoundary;
	bool IsFirstPageInsert;
	struct page_of_ranges *NextToFirstPage;
	struct page_of_ranges *LastPage;
	u64 LastPageBoundary;
	bool IsLastPageInsert;
	void *NewPageList = NULL; /* initialized only to suppress a warning */

	if (Start == End)
		/*  Safeguard */
		return true;

	/*  Find the page nearest to Start
	 *  Include the page's edges
	 */
	Result = page_of_ranges_find_page(&(Table->Tree),
			Start,
			RANGE_SEARCH_WITH_START | RANGE_SEARCH_WITH_END,
			&Page);
RetryTheSwitch:
	/*
	 * Determine how Start is related to the first covered page
	 * The default (for Below case) is - not related at all
	 */
	FirstPage = NULL;
	FirstPageBoundary = Start;
	IsFirstPageInsert = false;
	switch (Result) {
	case RangeTreeInside:
		/* Start is inside Page or touches it
		 * Get the end value of Page
		 */
		FirstPageBoundary = PAGE_END_ABS(Page);
		if (End <= FirstPageBoundary) {
			/* Insert only involves Page and no other pages
			 * Try to insert directly
			 */
			if (page_of_ranges_insert(Table,
					    Page,
					    (URPVALUE)(Start - Page->Base),
					    (URPVALUE)(End - Page->Base))) {
				/*
				 * Try to merge the pages with the updated one
				 * We are not interested in whether the page was actually merged
				 * If yes - then the CRCs for the halves were surely correcly set
				 * If no - then merge was a no-op, and the tree had CRC OK before
				 *  the merge attempt
				 */
				(void)page_of_ranges_try_to_merge(Table, Page);
				/*  Fully done */
				goto Success;
			}
			/* Page is too full, split it */
			if (!page_of_ranges_split(&(Table->Tree), Page))
				/* Out of memory */
				goto Failed;

			/* Split was OK, determine how to retry properly after it
			 * The default is - we are inside Page (Result is already such)
			 */
			FirstPageBoundary = PAGE_END_ABS(Page);
			if (Start > FirstPageBoundary) {
				/* This always works after split */
				Page = (struct page_of_ranges *)tree_next(Page);
				FirstPageBoundary = PAGE_START_ABS(Page);
				if (Start < FirstPageBoundary) {
					/* Start is between old Page and current Page */
					Result = RangeTreeBetween;
				} else {
					/* Start is inside Page */
				}
			} else {
				/* Start is inside Page */
			}
			goto RetryTheSwitch;
		} else {
		}

		/*
		 * We are covering at least 2 pages - Page and something beyond it
		 * Page is covered till its very end, so, insert to Page will never want a split
		 *  (never increase Page->Count)
		 * The "head" which will be inserted to Page is only limited by the URPVALUE
		 *  bitness of "walls"
		 */
		FirstPageBoundary = range_plus_max_URPValue(Page->Base);
		if (FirstPageBoundary > End)
			FirstPageBoundary = End;

		/* [Start..FirstPageBoundary) will be inserted to FirstPage... */
		FirstPage = Page;
		IsFirstPageInsert = true;
		/*  ...and [FirstPageBoundary..End) - to next page (can be NULL) */
		Page = (struct page_of_ranges *)tree_next(Page);
		break;
	case RangeTreeBetween:
		/* We have the previous page, and must match against it */
		Page = (struct page_of_ranges *)tree_prev(Page);
	/* Fallthrough */
	case RangeTreeAbove:
		/*
		 * Page is the previous page, and we are not even touching it
		 * Also Page can be NULL if the whole tree is empty
		 * Check whether some starting part of the range can be inserted to the tail of Page
		 * This check properly handles the Page == NULL case, which is possible here
		 */
		FirstPageBoundary = range_insert_to_page_tail_check(Page, Start, End);
		if (Start < FirstPageBoundary) {
			/* [Start..FirstPageBoundary) will be inserted to FirstPage */
			FirstPage = Page;
			/* IsFirstPageInsert is already false here */
		} else {
			/* Nothing will be inserted to the previous page, just ignore its existence */
		}
		/* Take the next page after FirstPage */
		if (Page != NULL)
			Page = (struct page_of_ranges *)tree_next(Page); /* can become NULL */
		break;
	case RangeTreeBelow:
		break;
	}
	/* The first page which has something to do with [FirstPageBoundary..End) range */
	NextToFirstPage = Page;
	/*
	 * Here FirstPage/FirstPageBoundary/IsFirstPageInsert describe the fate of the
	 *  [Start..FirstPageBoundary) range
	 * Decide the fate of the [FirstPageBoundary..End) range
	 * The default is - this range does not exist (also used for Above case below)
	 */
	LastPage = NULL;
	LastPageBoundary = End;
	IsLastPageInsert = false;
	if (FirstPageBoundary < End) {
		/* Find the page nearest to End
		 * Include the page's start but not end
		 */
		Result = page_of_ranges_find_page(&(Table->Tree),
				End,
				RANGE_SEARCH_WITH_START,
				&Page);
		switch (Result) {
		case RangeTreeInside:
			/*
			 * End is inside Page or touches its start
			 * Page is covered from its very start, so, insert to Page will never want a split
			 *  (never increase Page->Count)
			 * Find the lowest possible value for Page->Base which will still be able
			 *  to hold the last "wall" (which is not touched by this insert)
			 */
			LastPageBoundary = PAGE_END_ABS(Page); /* last "wall" */
			if (LastPageBoundary <= (URPVALUE)(-1))
				LastPageBoundary = 0;
			else
				LastPageBoundary -= (URPVALUE)(-1); /* surely <= Page->Base */
			/*
			 * We are OK in setting the final LastPageBoundary as low as the current
			 *  LastPageBoundary (lowest possible base), but LastPageBoundary must never be
			 *  < FirstPageBoundary
			 */
			if (LastPageBoundary < FirstPageBoundary)
				LastPageBoundary = FirstPageBoundary;

			/*
			 * Here LastPageBoundary can be < Page->Base, and is OK to rebase there
			 * Or it can be >= Page->Base, but still within Page (even End is within page)
			 * We will insert to this page
			 */
			LastPage = Page; /* end of LastPage is > End */
			IsLastPageInsert = true;
			break;
		case RangeTreeBelow:
		/* Fallthrough */
		case RangeTreeBetween:
			/*
			 * Page is above (not touching) End
			 * Check whether some ending part of the range can be inserted
			 *  to the head of Page
			 */
			LastPageBoundary = range_insert_to_page_head_check(Page,
					FirstPageBoundary,
					End);
			if (LastPageBoundary < End) {
				/* [LastPageBoundary..End) will be inserted to LastPage */
				LastPage = Page; /* end of LastPage is > End */
				/* IsLastPageInsert is already false */
			} else {
				/* Nothing will be inserted to the next page, just ignore its existence */
			}
			break;
		case RangeTreeAbove:
			break;
		}
	} else {
	}
	/*
	 * All boundaries are set
	 * If there is a [FirstPageBoundary..LastPageBoundary) range - then it must be covered
	 *  by the newly created pages
	 */
	if (FirstPageBoundary < LastPageBoundary) {
		/* Allocate them */
		NewPageList = range_page_chain_prealloc(Table, FirstPageBoundary, LastPageBoundary);
		if (NewPageList == NULL)
			/* Allocation failed */
			goto Failed;
	} else {
	}
	/*
	 * Now we can execute the operation
	 * From now on, nothing can fail
	 * First [Start..FirstPageBoundary)...
	 */
	if (Start < FirstPageBoundary) {
		if (IsFirstPageInsert) {
			/* Insert to FirstPage, this cannot cause a split (checked above) */
			(void)page_of_ranges_insert(Table,
					FirstPage,
					(URPVALUE)(Start - FirstPage->Base),
					(URPVALUE)(FirstPageBoundary - FirstPage->Base));
		} else {
			/* Add a range to the tail of FirstPage */
			range_insert_to_page_tail_do(Table, FirstPage, Start, FirstPageBoundary);
		}
	} else {
	}
	/*
	 * Scan the pages from FirstPage on
	 * If the page is fully covered - then delete it
	 * Otherwise, delete the starting range in it
	 * Note that, if LastPage exists, then its end is > End, so LastPage will
	 *  never be deleted here
	 */
	while (NextToFirstPage != NULL) {
		/* Take the next page beforehand */
		Page = (struct page_of_ranges *)tree_next(NextToFirstPage);
		if (PAGE_END_ABS(NextToFirstPage) <= End) {
			/*
			 * Fully covered by the insertion, delete it
			 */
			Table->RangesCount -= (NextToFirstPage->Count / 2);
			Table->ValuesCount -= NextToFirstPage->ValuesCount;
			tree_delete_existing(&(Table->Tree), NextToFirstPage);
		} else if (PAGE_START_ABS(NextToFirstPage) < End) {
			/*
			 * Partially covered by the insertion, delete the range in it
			 * We are deleting from the very page start, so, we will not make a hole
			 *  and run out of space
			 */
			(void)page_of_ranges_delete(Table,
					NextToFirstPage,
					0, /* from the very start */
					(URPVALUE)(End - NextToFirstPage->Base));
			/* This is the last page */
			break;
		} else {
			/* Not covered at all
			 * This is the last page - must break out of loop
			 */
			break;
		}
		/* Retry with the next page */
		NextToFirstPage = Page;
	}
	/* ...then [FirstPageBoundary..LastPageBoundary)... */
	if (FirstPageBoundary < LastPageBoundary) {
		/* Fill and attach the preallocated pages */
		range_page_chain_attach(Table, NewPageList, FirstPageBoundary, LastPageBoundary);
	} else {
	}
	/* ...and finally [LastPageBoundary..End) */
	if (LastPageBoundary < End) {
		if (IsLastPageInsert) {
			/* Here LastPageBoundary can be < LastPage->Base, and is OK to rebase there
			 * Or it can be >= LastPage->Base, but still within LastPage
			 */
			if (LastPageBoundary < LastPage->Base) {
				/* Rebase LastPage down */
				page_of_ranges_rebase_down(LastPage, LastPageBoundary);
			} else {
			}
			/*
			 * Insert the range there
			 * This will never cause a split - either rebase above caused us to insert at 0,
			 *  or the delete above have deleted everything from this page up to End,
			 *  which is > LastPageBoundary
			 */
			(void)page_of_ranges_insert(Table,
					LastPage,
					(URPVALUE)(LastPageBoundary - LastPage->Base),
					(URPVALUE)(End - LastPage->Base));
		} else {
			/* Add a range to the head of LastPage */
			range_insert_to_page_head_do(Table, LastPage, LastPageBoundary, End);
		}
	} else {
	}
	/* Done */
Success:
	return true;
Failed:
	return false;
}

/*
 * Deletes a range from the range page table
 * Returns false on allocation failure (if hole carving was tried and failed),
 *  the table is not updated in this case
 * The caller must guarantee Start < End
 */
bool range_delete(struct range_page_table *Table, u64 Start, u64 End)
{
	struct page_of_ranges *PageToMerge;
	enum enum_search_result FirstResult;
	struct page_of_ranges *FirstPage;
	enum enum_search_result LastResult;
	struct page_of_ranges *LastPage;
	URPVALUE RPStart;
	URPVALUE RPEnd;
	struct page_of_ranges *NextPage;

	if (Start == End)
		/* Safeguard */
		return true;

RetryAfterSplit:
	PageToMerge = NULL;
	/* Find the first page touched by delete */
	FirstResult = page_of_ranges_find_page(&(Table->Tree), Start, 0, &FirstPage);
	if (FirstResult == RangeTreeAbove) {
		/* Deleting above the whole table, nothing to do */
		goto DoneTrue;
	}
	/* Find the last page touched by delete */
	LastResult = page_of_ranges_find_page(&(Table->Tree), End, 0, &LastPage);
	/* LastPage is surely non-NULL here, since the case of empty table is handled above */
	if (LastResult == RangeTreeBelow) {
		/* Deleting below the whole table, nothing to do */
		goto DoneTrue;
	}
	switch (FirstResult) {
	case RangeTreeInside:
		/*
		 * We need to delete a range from FirstPage
		 * Start is inside FirstPage, we are not deleting above or below the page
		 * Calculate the page-relative values of the delete
		 */
		RPStart = (URPVALUE)(Start - FirstPage->Base);
		if (PAGE_END_ABS(FirstPage) <= End)
			RPEnd = PAGE_END(FirstPage);
		else
			RPEnd = (URPVALUE)(End - FirstPage->Base);

		/* Delete from FirstPage */
		if (!page_of_ranges_delete(Table, FirstPage, RPStart, RPEnd)) {
			/* Out of room carving a hole
			 * Split the page
			 */
			if (!page_of_ranges_split(&(Table->Tree), FirstPage)) {
				/* Out of memory for the second page */
				return false;
			}
			goto RetryAfterSplit;
		} else {
		}
		/* This page will be the merge point */
		PageToMerge = FirstPage;
		/* LastPage is surely non-NULL here */
		if (FirstPage == LastPage) {
			/* No more affected pages */
			goto DoTheMerges;
		}

		/*
		 * Done with FirstPage
		 * LastPage is surely non-NULL here and is after FirstPage in the tree, so,
		 *  FirstPage cannot be the last one
		 */
		FirstPage = (struct page_of_ranges *)tree_next(FirstPage);
		UASSERT(FirstPage != NULL, "LastPage before FirstPage");
		break;
	case RangeTreeBelow:

	case RangeTreeBetween:
		break;
	default: /* Avoid error: enumeration value ‘*’ not handled in switch */
		break;
	}
	/* FirstPage is the first page after Start (cannot be NULL),
	 *  pages before it are no more affected
	 */
	switch (LastResult) {
	case RangeTreeInside:
		/*
		 * We need to delete a range from LastPage
		 * We will never carve a hole in it, we are deleting from start, so, ignore
		 *  the return value
		 * End is inside LastPage, Start is at least not above it, so, we are not
		 *  deleting above or below the page
		 * Calculate the page-relative values of the delete
		 * From this page start
		 */
		RPStart = PAGE_START(LastPage);
		RPEnd = (URPVALUE)(End - LastPage->Base);
		(void) page_of_ranges_delete(Table, LastPage, RPStart, RPEnd);

		/* This page will be the merge point (if there is no merge point yet) */
		if (PageToMerge == NULL)
			PageToMerge = LastPage;

		break;
	case RangeTreeAbove:
		/* Will delete pages up to the tree end */
		LastPage = NULL;
		break;
	case RangeTreeBetween:
		break;
	/* Below cannot occur here */
	default: /* Avoid error: enumeration value ‘*’ not handled in switch */
		break;
	}
	/* LastPage is no more affected (can be NULL)
	 * Delete the complete pages between FirstPage and LastPage
	 */
	if (FirstPage != NULL) { /* additional safeguard which is not harmful at all */
		while (FirstPage != LastPage) {
			/* This page will be the merge point (if there is no merge point yet) */
			if (PageToMerge == NULL)
				PageToMerge = LastPage; /* can become NULL, but this is normal */

			/* Take the next page before deletion */
			NextPage = (struct page_of_ranges *)tree_next(FirstPage);
			/* Deal with the counts */
			Table->RangesCount -= (FirstPage->Count / 2);
			Table->ValuesCount -= FirstPage->ValuesCount;
			/* Delete */
			tree_delete_existing(&(Table->Tree), FirstPage);
			FirstPage = NextPage;
		}
	}
DoTheMerges:
	/* Try to merge the pages with the updated one */
	if (PageToMerge != NULL) {
		/*
		 * We are not interested in whether the page was actually merged
		 * If yes - then the CRCs for the halves were surely correcly set
		 * If no - then merge was a no-op, and the tree had CRC OK before
		 *  the merge attempt
		 */
		(void)page_of_ranges_try_to_merge(Table, PageToMerge);
	}
	/* Done */
DoneTrue:
	return true;
}

/*
 * Deletes a range which is in the table by its reference (was found by range_hit or
 *  range_enumerate, or obtained by range_next/Prev/First/Last)
 */
void range_delete_existing(struct range_page_table *Table, struct range_ref *Range)
{
	struct page_of_ranges *Page;
	PURPVALUE Ptr;

	Page = (struct page_of_ranges *)(Range->Page);
	if (Page == NULL)
		/* NULL reference, nothing to do */
		return;
	/* Find the location within the page */
	Ptr = (((PURPVALUE)(Page + 1)) + Range->Index);

	/* Deal with the counts */
	Table->RangesCount--;
	Table->ValuesCount -= (Ptr[1] - Ptr[0]);
	Page->ValuesCount -= (Ptr[1] - Ptr[0]);
	/* Shrink Page to collapse away the range */
	if ((Range->Index + 2) < Page->Count)
		memmove(Ptr, Ptr + 2, (Page->Count - (Range->Index + 2)) * sizeof(URPVALUE));
	Page->Count -= 2;
	if (Page->Count != 0) {
		/* Try to merge the page with neighbours */
		page_of_ranges_try_to_merge(Table, Page);
	} else {
		/* Page became empty, delete it */
		tree_delete_existing(&(Table->Tree), Page);
	}
	/* Done */
}

/* Deletes all ranges from the table */
void range_delete_all(struct range_page_table *Table)
{
	/* Delete all pages */
	tree_delete_all(&(Table->Tree));

	/* Reset the counts */
	Table->RangesCount = 0;
	Table->ValuesCount = 0;
}

/*
 * Looks up whether Value is covered by some range
 * If not covered - returns false and sets *Range to reference the next table range just
 *  after Value, or to logical NULL if the table is empty or if Value is > then
 *  any table range
 * If covered - returns true and sets *Range to reference the range which contains
 *  this value
 * The returned range reference is valid till next table update or cleanup
 */
bool range_hit(struct range_page_table *Table, u64 Value,
		  struct range_ref *Range)
{
	enum enum_search_result Result;
	struct page_of_ranges *Page;
	PURPVALUE Array;
	u32 Index;
	bool IsHit;

	/* Search Value against the tree */
	Result = page_of_ranges_find_page(&(Table->Tree), Value, RANGE_SEARCH_WITH_START, &Page);
	if (Range != NULL) {
		switch (Result) {
		case RangeTreeBelow:
		/* Below all ranges in the tree (below the first page), Page is the first page
		 * Fallthru
		 */
		case RangeTreeBetween:
			/* Between the pages, below Page
			 * Return the first range in Page
			 */
			Range->Start = PAGE_START_ABS(Page);
			Range->End = (Page->Base + PAGE_START2(Page));
			Range->Page = Page;
			Range->Index = 0;
			/* No hit */
			return false;
		case RangeTreeAbove:
			/* Return logical NULL */
			Range->Start = Range->End = 0;
			Range->Page = NULL;
			/* No hit */
			return false;
		default: /* Avoid error: enumeration value ‘*’ not handled in switch */
			break;
		}
	} else {
		switch (Result) {
		case RangeTreeBelow:
		case RangeTreeBetween:
		case RangeTreeAbove:
			/* No hit in all cases */
			return false;
		default: /* Avoid error: enumeration value ‘*’ not handled in switch */
			break;
		}
	}
	/* Hit inside the page */
	Array = (PURPVALUE)(Page + 1);
	/* Find the exact position within it */
	Index = page_of_ranges_bsearch(Array, Page->Count, (URPVALUE)(Value - Page->Base));
	UASSERT(Index != (u32)(-1), "damaged tree - hit beyound the page end");
	if ((URPVALUE)(Value - Page->Base) == Array[Index]) {
		/* Exact hit to some value, move Index up so it will be the index of the first value
		 *  which is > Value
		 */
		Index++;
		UASSERT(Index < Page->Count,
				"damaged tree - hit exactly at the page end");
	}
	/* Index is the index of the first value which is > Value */
	IsHit = false;
	if ((Index & 1) != 0) {
		/* Inside some range */
		IsHit = true;
		/* Point to its start */
		Index--;
	} else {
		/* Between the ranges, Index references the next range */
	}
	if (Range != NULL) {
		/* Fill the result */
		Range->Start = (Page->Base + Array[Index]);
		Range->End = (Page->Base + Array[Index + 1]);
		Range->Page = Page;
		Range->Index = Index;
	}
	return IsHit;
}

/* Advances *Range to the next range or to logical NULL if the given range is last
 * Can return adjacent ranges
 */
void range_next(struct range_ref *Range)
{
	struct page_of_ranges *Page;

	Page = (struct page_of_ranges *)(Range->Page);
	if (Page == NULL)
		/* Already at end, nothing to do */
		return;
	/* Next range within the page */
	Range->Index += 2;
	if (Range->Index >= Page->Count) {
		/* Next page */
		Range->Page = Page = (struct page_of_ranges *)tree_next(Page);
		if (Page == NULL) {
			/* End of tree */
			Range->Start = Range->End = 0;
			return;
		}
		/* Start of next page */
		Range->Index = 0;
	}
	/* Fill the result */
	Range->Start = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index]);
	Range->End = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index + 1]);
	/* Done */
}

/* Advances *Range to the previous range or to logical NULL if the given range is first
 * Can return adjacent ranges
 */
void range_prev(struct range_ref *Range)
{
	struct page_of_ranges *Page;

	Page = (struct page_of_ranges *)(Range->Page);
	if (Page == NULL)
		/* Already at start, nothing to do */
		return;
	if (Range->Index == 0) {
		/* Previous page */
		Range->Page = Page = (struct page_of_ranges *)tree_prev(Page);
		if (Page == NULL) {
			/* Start of tree */
			Range->Start = Range->End = 0;
			return;
		}
		/* End of previous page */
		Range->Index = Page->Count;
	}
	/* Previous range within the page */
	Range->Index -= 2;
	/* Fill the result */
	Range->Start = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index]);
	Range->End = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index + 1]);
	/* Done */
}

/*
 * Sets *Range to reference the first range in the table
 *  or to logical NULL if the table is empty
 * Can return adjacent ranges
 */
void range_first(struct range_page_table *Table, struct range_ref *Range)
{
	struct page_of_ranges *Page;

	/* First page */
	Range->Page = Page = (struct page_of_ranges *)tree_first(&(Table->Tree));
	if (Page == NULL) {
		/* Tree is empty */
		Range->Start = Range->End = 0;
		return;
	}
	/* Return the first range */
	Range->Index = 0;
	Range->Start = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index]);
	Range->End = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index + 1]);
	/* Done */
}

/*
 * Sets *Range to reference the last range in the table
 *  or to logical NULL if the table is empty
 * Can return adjacent ranges
 */
void range_last(struct range_page_table *Table, struct range_ref *Range)
{
	struct page_of_ranges *Page;

	/* Last page */
	Range->Page = Page = (struct page_of_ranges *)tree_last(&(Table->Tree));
	if (Page == NULL) {
		/* Tree is empty */
		Range->Start = Range->End = 0;
		return;
	}
	/* Return the last range */
	Range->Index = (Page->Count - 2);
	Range->Start = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index]);
	Range->End = (Page->Base + ((PURPVALUE)(Page + 1))[Range->Index + 1]);
	/* Done */
}

/*
 * Table enumerator
 * Returns ranges in the comparison order
 * Does not do any reorders to the table
 * Returns (via *Range) the next range if found or logical NULL if no more items
 * Returns true if the next range was found, false otherwise
 * Range->Page should be set to (void *)(s64)(-1) before calling this for the first time
 *  to initialize the enumeration context (Range)
 * Can return adjacent ranges
 */
bool range_enumerate(struct range_page_table *Table, struct range_ref *Range)
{
	if (Range->Page == (void *)(s64)(-1))
		/* Take the first range */
		range_first(Table, Range);
	else
		/* Take the next range */
		range_next(Range);
	/* Return whether Range is logical NULL or not so */
	return (bool)((Range->Page != NULL) ? true : false);
}

/*
 * Same as range_enumerate but merges the adjacent ranges
 * Context must be initialized by range_enumerate_contexct_init before the first call,
 *  and reused for next calls
 * Return value is in RetVal
 */
bool range_enumerate_merge_adjacent(struct range_page_table *Table,
				     struct range_page_table_enum_context *Context,
				     struct range_page_table_enum_retval *RetVal)
{
	if (Context->Range.Page == (void *)(s64)(-1)) {
		/* Initialize the accumulator on the very first call */
		Context->End = 0;
		/* Take the first range */
		range_first(Table, &(Context->Range));
	}
	/*
	 * For first call - accumulator is empty, ->Range is the first range
	 *  or logical NULL if the table is empty
	 * For next call - accumulator is empty, ->Range is the next range
	 *  or logical NULL if the table has ended
	 */
	if (Context->Range.Page == NULL)
		/* End of table */
		return false;
	/* Accumulate the first/next table range */
	Context->Start = Context->Range.Start;
	Context->End = Context->Range.End;
	/* Merging loop */
	for (;; ) {
		/* We have accumulated a range here
		 * Take the next table range
		 */
		range_next(&(Context->Range));
		if (Context->Range.Page == NULL)
			/* Table ended, we must return the last accumulated range */
			break;
		if (Context->End != Context->Range.Start)
			/* Non consecutive, we must return the last accumulated range
			 * ->Range will be processed on the next call here
			 */
			break;
		/* Consecutive, consume ->Range to prolong the accumulated range */
		Context->End = Context->Range.End;
		/* Next table range */
	}
	/* Return accumulated range */
	RetVal->Start = Context->Start;
	RetVal->End = Context->End;
	/* Empty the accumulator */
	Context->End = 0;
	return true;
}

/*
 * Looks up whether the range is completely contained in the table
 * Returns true if contained
 */
bool range_check_for_dense_range(struct range_page_table *Table,
				 u64 Start, u64 End)
{
	struct range_ref Range;
	u64 PrevEnd;

	/* Find the range containing Start in the table */
	if (!range_hit(Table, Start, &Range))
		/* Start is not in a range */
		return false;
	/* Adjacent range loop */
	for (;; ) {
		/* Check whether End is within Range */
		if (End <= Range.End)
			return true;
		/* End > Range.End
		 * Take the next range
		 */
		PrevEnd = Range.End;
		range_next(&Range);
		if (Range.Page == NULL)
			/* No next range */
			break;
		if (Range.Start != PrevEnd)
			/* Not adjacent */
			break;
		/* Adjacent, do one more check for End */
	}
	/* [Start..End) is not fully covered */
	return false;
}

/*
 * Looks up whether the range is completely not contained in the table
 * Returns true if not contained
 */
bool range_check_for_hole_range(struct range_page_table *Table,
				u64 Start, u64 End)
{
	struct range_ref Range;

	/* Find the range containing Start in the table */
	if (range_hit(Table, Start, &Range))
		/* Start is in a range */
		return false;
	if (Range.Page == NULL)
		/* Table is empty, or Start is > any range
		 * Hole
		 */
		return true;
	/* Range is the first range just after Start
	 * Check that End also hits before it
	 */
	return (bool)((End <= Range.Start) ? true : false);
}

/*
 * Checks whether [Start..End) intersects with any in-table range
 * Returns true if yes, and fills the intersection (least of them if there are several)
 *  to *Range, the table location fields in *Range are not defined
 * Returns false if no, the state of *Range is not defined in this case
 */
bool range_next_range_in(struct range_page_table *Table,
			  u64 Start, u64 End,
			  struct range_ref *Range)
{
	u64 ResultStart;
	u64 ResultEnd;

	/* Check for zero length range */
	if (Start == End)
		/* No such range */
		return false;
	/* Find the range containing Start - or just after it - in the table */
	if (!range_hit(Table, Start, Range)) {
		if (Range->Page == NULL) {
			/*
			 * Start is > any range or the table is empty
			 * No intersections
			 * Internal consistency check that the requested range is really not contained
			 */
			UASSERT(range_check_for_hole_range(Table, Start, End),
					"not found anything in non-hole range - internal error");
			return false;
		}
		/* Range is the next range after Start
		 * Check whether End will hit it
		 */
		if (End <= Range->Start) {
			/* No intersection
			 * Internal consistency check that the requested range is really not contained
			 */
			UASSERT(range_check_for_hole_range(Table, Start, End),
					"not found anything in non-hole range - internal error");
			return false;
		}
		/* We intersect with Range, and its Start is correct */
		ResultStart = Range->Start;
		/* Range is the first range touched */
	} else
		/* Start is inside some range
		 * Range is the first range touched
		 */
		ResultStart = Start;
	/* ResultStart is already set correctly, and Range is the first range touched by us
	 * Adjacent range loop
	 */
	for (;; ) {
		ResultEnd = End;
		if (Range->End >= End)
			break;
		/* Range is fully covered by us */
		ResultEnd = Range->End;
		/* Take the next range */
		range_next(Range);
		if (Range->Page == NULL)
			/* No next range */
			break;
		if (Range->Start != ResultEnd)
			/* Not adjacent */
			break;
		/* Adjacent range, retry with it */
	}
	/* Fill the return value */
	Range->Start = ResultStart;
	Range->End = ResultEnd;
	/* Done */
	UASSERT(Range->Start < Range->End, "found empty range");
	/* Internal consistency check that the range we are returning is really contained */
	UASSERT(range_check_for_dense_range(Table, Range->Start, Range->End),
			"found nondense range - internal error");
	return true;
}

/*
 * Checks whether [Start..End) intersects with any hole between in-table ranges
 * Returns true if yes, and fills the intersection (least of them if there are several)
 *  to *Range, the table location fields in *Range are not defined
 * Returns false if no, the state of *Range is not defined in this case
 */
bool range_next_range_out(struct range_page_table *Table,
			   u64 Start, u64 End,
			   struct range_ref *Range)
{
	/* Check for zero length range */
	if (Start == End)
		/* No such range */
		return false;
	/* Find the range containing Start - or just after it - in the table */
	if (range_hit(Table, Start, Range)) {
		/* Start hits some range
		 * Adjacent range loop
		 */
		for (;; ) {
			/* Check whether End also hits the range */
			if (End <= Range->End) {
				/* Inside a range
				 * Internal consistency check that the requested range
				 *  is really all contained
				 */
				UASSERT(range_check_for_dense_range(Table, Start, End),
						"not found anything in nondense range - internal error");
				return false;
			}
			/* End is beyound the range
			 * Move Start to be beyound too
			 */
			Start = Range->End;
			/* Take the next range (bounds End) */
			range_next(Range);
			if ((Range->Page == NULL) || (Range->Start != Start))
				/* The range is not adjacent or no such range */
				break;
			/* Adjacent range, re-check End against it */
		}
	}
	if (Range->Page != NULL) {
		/* Range bounds End */
		if (End > Range->Start)
			/* Truncate End by the next range */
			End = Range->Start;
	}
	/*else
	 *  - Start beyound the whole table, or no range bounds End
	 */
	Range->Start = Start;
	Range->End = End;
	UASSERT(Range->Start < Range->End, "found empty range");
	/* Internal consistency check that the range we are returning is really not contained */
	UASSERT(range_check_for_hole_range(Table, Range->Start, Range->End),
			"found non-hole range - internal error");
	return true;
}


/*
 * Functions internal for this file
 *
 * Checks whether it is possible to insert a part of the [Start..End) range to the head
 *  of Page
 * Returns End if not possible, otherwise, returns the minimum start value (>= Start)
 *  of the range to be inserted
 * The caller must guarantee that End < any value in the page
 * Does not update anything
 */
static u64 range_insert_to_page_head_check(struct page_of_ranges *Page,
					  u64 Start, u64 End)
{
	u64 Min;

	/* Find the lowest possible value for Page->Base which will still be able
	 *  to hold the last "wall"
	 */
	Min = PAGE_END_ABS(Page); /* last "wall" */
	if (Min <= (URPVALUE)(-1))
		Min = 0;
	else
		Min -= (URPVALUE)(-1); /* surely <= Page->Base */

	if (End <= Min) {
		/* Even End cannot fit */
		return End;
	}
	/* The [Min..End) or [Start..End) range can be added now
	 * Check for page room
	 */
	if ((Page->Count + 2) > PAGE_RANGE_MAX) {
		/* No room in the page */
		return End;
	}
	/* Calculate the start of the range we are inserting */
	if (Min < Start)
		Min = Start;

	/* Done */
	return Min;
}

/*
 * Actually inserts the range to the head of the page
 * The table is fully set up on return
 * Must only be called after range_insert_to_page_head_check, with the same End value
 * Start must be the return value of range_insert_to_page_head_check,
 *  and must never be == End
 */
static void range_insert_to_page_head_do(struct range_page_table *Table,
					struct page_of_ranges *Page,
					u64 Start, u64 End)
{
	PURPVALUE Ptr;

	if (Start < Page->Base) {
		/* Rebase the page down */
		page_of_ranges_rebase_down(Page, Start);
	} else {
	}
	Ptr = (PURPVALUE)(Page + 1);
	/* Shift the "walls" up to free the space for 1 more range */
	memmove(Ptr + 2, Ptr, Page->Count * sizeof(URPVALUE));
	/* Add one more range to the head of Page */
	Ptr[0] = (URPVALUE)(Start - Page->Base);
	Ptr[1] = (URPVALUE)(End - Page->Base);
	/* Update the rest of Page */
	Page->Count += 2;
	Page->ValuesCount += (URPVALUE)(End - Start);
	/* Update the table's counters */
	++Table->RangesCount;
	Table->ValuesCount += (URPVALUE)(End - Start);
	/* Done */
}

/*
 * Checks whether it is possible to insert a part of the [Start..End) range to the tail
 *  of Page
 * Returns Start if not possible, otherwise, returns the maximum end value (<= End)
 *  of the range to be inserted
 * Page can be NULL, this is properly handled
 * The caller must guarantee that Start > any value in the page
 * Does not update anything
 */
static u64 range_insert_to_page_tail_check(struct page_of_ranges *Page,
					  u64 Start, u64 End)
{
	u64 Max;

	if (Page == NULL) {
		/* No Page */
		return Start;
	}
	/* Calculate the maximum value which can fit to Page */
	Max = range_plus_max_URPValue(Page->Base);
	if (Max <= Start) {
		/* Even Start cannot fit */
		return Start;
	}
	/* The [Start..Max) or [Start..End) range can be added now
	 * Check for page room
	 */
	if ((Page->Count + 2) > PAGE_RANGE_MAX) {
		/* No room in the page */
		return Start;
	}
	/* Calculate the end of the range we are inserting */
	if (Max > End)
		Max = End;

	/* Done */
	return Max;
}

/*
 * Actually inserts the range to the tail of the page
 * The table is fully set up on return
 * Must only be called after range_insert_to_page_tail_check, with the same Start value
 * End must be the return value of range_insert_to_page_tail_check,
 *  and must never be == Start
 */
static void range_insert_to_page_tail_do(struct range_page_table *Table,
					struct page_of_ranges *Page,
					u64 Start, u64 End)
{
	/* Add one more range to the tail of Page */
	((PURPVALUE)(Page + 1))[Page->Count] = (URPVALUE)(Start - Page->Base);
	((PURPVALUE)(Page + 1))[Page->Count + 1] = (URPVALUE)(End - Page->Base);
	/* Update the rest of Page */
	Page->Count += 2;
	Page->ValuesCount += (URPVALUE)(End - Start);
	/* Update the table's counters */
	++Table->RangesCount;
	Table->ValuesCount += (URPVALUE)(End - Start);
	/* Done */
}

/*
 * Creates the chain of the preallocated pages long enough to cover [Start..End)
 * Returns NULL on allocation error or the pointer to the first page on success
 * The pages are single-linked via the first word in them, interpreted as next page pointer
 */
static void *range_page_chain_prealloc(struct range_page_table *Table,
					u64 Start,
					u64 End)
{
	void *ListHead;
	URPVALUE Length;
	void *Page;

	ListHead = NULL;
	/* Retry loop for the pages */
	while (Start < End) {
		/* Check whether the current range length is limited by URPVALUE */
		if (range_diff_fits_URPValue(Start, End)) {
			Length = (URPVALUE)(End - Start);
		} else {
			/* Does not fit, use maximum URPVALUE */
			Length = (URPVALUE)(-1);
		}
		/* Allocate the next page */
		Page = tree_prealloc(&(Table->Tree));
		if (Page == NULL) {
			/* Allocation failed, free everything and quit */
			range_page_chain_prefree(Table, ListHead);
			return NULL;
		}
		/* Link the page to the head of the list */
		*(void **)Page = ListHead;
		ListHead = Page;
		/* Consume the space in the Start/End range */
		Start += Length;
		/* Try the next page */
	}
	/* Chain is done */
	return ListHead;
}

/*
 * Undo for range_page_chain_prealloc
 * ListHead is allowed to be NULL, no-op in this case
 */
static void range_page_chain_prefree(struct range_page_table *Table, void *ListHead)
{
	void *Page;

	/* Loop while the list is not empty */
	while (ListHead != NULL) {
		/* Detach the first page */
		Page = ListHead;
		ListHead = *(void **)Page;
		/* Free the detached page */
		tree_prefree(&(Table->Tree), Page);
		/* Next page */
	}
	/* Done */
}

/*
 * Fills the pages allocated by range_page_chain_prealloc to cover [Start..End)
 *  and attaches them to the tree
 * [Start..End) must be the same as for range_page_chain_prealloc
 */
static void range_page_chain_attach(struct range_page_table *Table,
				     void *ListHead,
				     u64 Start,
				     u64 End)
{
	URPVALUE Length;
	void *Page;
	struct page_of_ranges_template Template;
	struct page_of_ranges *AttachPage;

	/* Retry loop for the pages */
	while (Start < End) {
		/* Check whether the current range length is limited by URPVALUE */
		if (range_diff_fits_URPValue(Start, End)) {
			Length = (URPVALUE)(End - Start);
		} else {
			/* Does not fit, use maximum URPVALUE */
			Length = (URPVALUE)(-1);
		}
		/* Take the next page off the head of the list */
		Page = ListHead;
		ListHead = *(void **)Page;
		/* Fill the template for this page
		 * 1 range, starts at Start
		 */
		Template.Page.Base = Start;
		Template.Value0 = 0;
		/* Attach the page to the tree */
		AttachPage = (struct page_of_ranges *)tree_attach(&(Table->Tree), Page, &Template, NULL);
		/* Set the End value and the item count */
		PAGE_SET_END(AttachPage, Length);
		/* Deal with the counts */
		AttachPage->ValuesCount = Length;
		Table->RangesCount++;
		Table->ValuesCount += Length;
		/* Done
		 * Consume the space in the Start/End range
		 */
		Start += Length;
		/* Try the next page */
	}
}

/*
 * Updates the page so it still describes the same ranges, but ->Base it set to be
 *  NewBase
 * On call, NewBase must be < ->Base
 * The caller must check that the last "wall" in the page will still fit URPVALUE
 *  after rebasing, this function has no such checks
 * Does not update the page's CRC
 */
static void page_of_ranges_rebase_down(struct page_of_ranges *Page, u64 NewBase)
{
	URPVALUE Bias;
	u32 i;

	/* Calculate the rebasing bias */
	Bias = (URPVALUE)(Page->Base - NewBase);
	/* Scan all "walls" and rebase each of them */
	for (i = 0; i < Page->Count; i++)
		((PURPVALUE)(Page + 1))[i] += Bias;
	/* Update the base */
	Page->Base = NewBase;
	/* Done */
}

/*
 * Splits Page it two, moving half of its items to the newly allocated page
 *  which is inserted next to Page in the tree
 * Page is still valid after this
 * Returns false on allocation error, no updates are done in this case
 */
static bool page_of_ranges_split(struct tree_table *Tree, struct page_of_ranges *Page)
{
	u32 Index;
	URPVALUE Bias;
	struct page_of_ranges_template Template;
	struct page_of_ranges *SecondPage;
	URPVALUE MovedValuesCount;
	u32 i;
	URPVALUE Start;
	URPVALUE End;

	/* Count the pairs */
	Index = (Page->Count / 2);
	/* Ensure the counter is even */
	Index = ((Index / 2) * 2);
	UASSERT(Index != 0, "first page will be empty");
	/*
	 * Fill the template with the first item of the page we are inserting
	 * We will make the new page's Base as high as possible to simplify later probable
	 *  merges with the next pages
	 */
	Bias = ((PURPVALUE)(Page + 1))[Index];
	Template.Page.Base = (Page->Base + Bias);
	/* All ranges in the new page must be biased down by Bias
	 * Bias down the first "wall"
	 */
	Template.Value0 = 0;
	/* Insert the second page */
	SecondPage = (struct page_of_ranges *)tree_insert(Tree, &Template, NULL);
	if (SecondPage == NULL)
		/* Allocation error */
		return false;
	/* Set the second page's count */
	SecondPage->Count = (Page->Count - Index);
	/* Patch the first page's count */
	Page->Count = Index;
	/* Copy the tail items from the first page to the second page, counting the values
	 *  in progress and biasing them down
	 */
	MovedValuesCount = 0;
	for (i = 0; i < SecondPage->Count; i += 2) {
		/* Get the values */
		Start = ((PURPVALUE)(Page + 1))[Index + i];
		End = ((PURPVALUE)(Page + 1))[Index + i + 1];
		/* Update the count */
		MovedValuesCount += (End - Start);
		/* Set the values */
		((PURPVALUE)(SecondPage + 1))[i] = (Start - Bias);
		((PURPVALUE)(SecondPage + 1))[i + 1] = (End - Bias);
	}
	/* Set the value counts */
	Page->ValuesCount -= MovedValuesCount;
	SecondPage->ValuesCount = MovedValuesCount;
	/* Done */
	return true;
}

/*
 * Tries to merge Page with neighbours
 * Does nothing and returns false if there is no need to merge
 * Returns true otherwise
 */
static bool page_of_ranges_try_to_merge(struct range_page_table *Table,
				     struct page_of_ranges *Page)
{
	bool WereUpdatesDone;
	struct page_of_ranges *NextPage;

	WereUpdatesDone = false;
	/* Try to merge forward */
	for (;; ) {
		/* Get the next page */
		NextPage = (struct page_of_ranges *)tree_next(Page);
		if ((NextPage == NULL) ||
		    !page_of_ranges_is_base_ok_for_merge(Page, NextPage) ||
		    ((Page->Count + NextPage->Count) > PAGE_RANGE_MERGE_THRESHOLD))
			/* Next page does not exist or not OK to merge
			 * Merge forward pass is done
			 */
			break;
		/* Next page exists and is OK to merge
		 * Merge with next page
		 */
		page_of_ranges_merge(Table, Page, NextPage);
		WereUpdatesDone = true;
		/* Get the new NextPage and try once more */
	}
	/* Try to merge backward */
	for (;; ) {
		/* Get the previous page */
		NextPage = Page;
		Page = (struct page_of_ranges *)tree_prev(NextPage);
		if ((Page == NULL) ||
		    !page_of_ranges_is_base_ok_for_merge(Page, NextPage) ||
		    ((Page->Count + NextPage->Count) > PAGE_RANGE_MERGE_THRESHOLD))
			/* Previous page does not exist or not OK to merge
			 * Merge backward pass is done
			 */
			break;
		/* Previous page exists and is OK to merge
		 * Merge with previous page
		 */
		page_of_ranges_merge(Table, Page, NextPage);
		WereUpdatesDone = true;
		/* Get the new Page/NextPage pair and try once more */
	}
	/* Done */
	return WereUpdatesDone;
}

/*
 * Checks whether the Base value in Page is good enough for a merge, i.e.
 *  all ranges in NextPage can be represented with this base
 */
static bool page_of_ranges_is_base_ok_for_merge(struct page_of_ranges *Page,
					   struct page_of_ranges *NextPage)
{
	u64 Value;

	/* The maximum value in NextPage is the last "wall" */
	Value = PAGE_END_ABS(NextPage);
	if (Value < Page->Base) {
		/* Internal inconsistency in the table */
		UASSERT(Value >= Page->Base, "Base values out of order");
		return false;
	}
	/* Get the offset of this value if it will be represented in Page */
	Value -= Page->Base;
	/* We can merge if (and only if) this offset fits URPVALUE */
	if ((Value & ~(u64)(URPVALUE)(-1)) == 0)
		return true;

	return false;
}

/*
 * Merges NextPage to Page, deletes NextPage from Table after this
 * NextPage must be the next page after Page in Table
 */
static void page_of_ranges_merge(struct range_page_table *Table, struct page_of_ranges *Page,
				struct page_of_ranges *NextPage)
{
	PURPVALUE Src;
	PURPVALUE Dst;
	u32 i;

	UASSERT((Page->Count + NextPage->Count) <= PAGE_RANGE_MAX,
			"merge result will be too long");
	Src = (PURPVALUE)(NextPage + 1);
	/* No sanity checks are done, the pages are assumed to be OK for a merge */
	if (PAGE_END_ABS(Page) == PAGE_START_ABS(NextPage)) {
		/* Last range in Page touches the first range in NextPage
		 * Collapse the NextPage's first range away to avoid a zero-sized hole
		 *  between the 2 adjacent ranges in the merge result
		 */
		PAGE_END(Page) += (PAGE_START2(NextPage) - PAGE_START(NextPage));
		/* Will skip this next range */
		Src += 2;
		NextPage->Count -= 2;
		/* One less range in the table */
		Table->RangesCount--;
		/* ValuesCount is not touched here, anyway it will be added to Page's one below */
	}
	/* Copy the ranges from NextPage to the tail of Page */
	Dst = ((PURPVALUE)(Page + 1) + Page->Count);
	for (i = 0; i < NextPage->Count; i++)
		/* Recalculate for the new base value */
		*Dst++ = (URPVALUE)((NextPage->Base + *Src++) - Page->Base);
	/* Increment the number of ranges in Page */
	Page->Count += NextPage->Count;
	/* Transfer the values count */
	Page->ValuesCount += NextPage->ValuesCount;
	/* Delete NextPage from Table */
	tree_delete_existing(&(Table->Tree), NextPage);
	/* Done */
}

/*
 * Searches the tree of pages for given value
 * Flags is a bitmask of RANGE _SEARCH_xxx above
 * Returns the search result and the found page in *Page
 *  - first tree's page for Below
 *  - next page for Between
 *  - last tree's page (or NULL if the tree is empty) for Above
 * Handling of the start and end values of the page - are they Inside or not - is governed
 *  by Flags
 */
static enum enum_search_result page_of_ranges_find_page(struct tree_table *Tree,
						      u64 Value,
						      u32 Flags,
						      struct page_of_ranges **Page)
{
	struct page_of_ranges_template Template;
	s32 ComparisonResult;
	struct page_of_ranges *PrevPage;
	struct page_of_ranges *NextPage;
	u64 EndValue;

	/* Fill the template
	 * Junior bits to Value0, everything more senior which does not fit URPVALUE - to
	 *  Page.Base
	 */
	Template.Page.Base = (Value & ~(u64)(URPVALUE)(-1));
	Template.Value0 = (URPVALUE)(Value - Template.Page.Base);
	/* Lookup the tree */
	*Page = (struct page_of_ranges *)tree_lookup_nearest(Tree, &Template, &ComparisonResult);
	if (*Page == NULL) {
		/* Tree is empty */
		return RangeTreeAbove;
	}
	if (ComparisonResult == 0) {
		/* Exact hit at page start */
		if (Flags & RANGE_SEARCH_WITH_START)
			return RangeTreeInside;

		/* Exact hit at page start counted before the page */
		if (tree_prev(*Page) != NULL)
			return RangeTreeBetween;

		/* Before the first page */
		return RangeTreeBelow;
	}
	if (ComparisonResult > 0) {
		/* *Page is > Value and is the nearest
		 * Previous page either does not exist or the nearest one which is < Value
		 */
		PrevPage = (struct page_of_ranges *)tree_prev(*Page);
		if (PrevPage == NULL) {
			/* *Page is > Value, is the nearest, and previous page does not exist
			 * We are below the first page
			 */
			return RangeTreeBelow;
		}
		/* PrevPage is the nearest page which is < Value */
		NextPage = *Page;
		*Page = PrevPage;
	} else {
		/* *Page is the nearest page which is < Value
		 * Next page either does not exist or the nearest one which is > Value
		 */
		NextPage = (struct page_of_ranges *)tree_next(*Page);
	}
	/* *Page starts below Value
	 * Check whether Value is inside *Page
	 */
	EndValue = PAGE_END_ABS(*Page);
	if (Value < EndValue) {
		/* Inside *Page */
		return RangeTreeInside;
	}
	if ((Flags & RANGE_SEARCH_WITH_END) && (Value == EndValue)) {
		/* At the end of *Page (inside) */
		return RangeTreeInside;
	}
	/* Value is > the whole *Page */
	if (NextPage == NULL) {
		/* Next page does not exist */
		return RangeTreeAbove;
	}
	/* Next page does exist, we are in between *Page and NextPage */
	*Page = NextPage;
	return RangeTreeBetween;
}

/* Constructor routine which fills the page with initial data (Data must be
 *  struct page_of_ranges_template *) to ensure the proper comparison order in the tree from the
 *  very beginning
 * Must match PUT_TREE_CONSTRUCT_ROUTINE
 */
static void range_construct(void *Tree, void *Item, void *Data)
{
	struct page_of_ranges *Page;
	struct page_of_ranges_template *Template;
	URPVALUE Start;

	Page = (struct page_of_ranges *)Item;
	Template = (struct page_of_ranges_template *)Data;
	/* Get the Base value from the template */
	Page->Base = Template->Page.Base;
	/* Get the value from the template */
	Start = Template->Value0;
	/* Set the page start value */
	PAGE_START(Page) = Start;
	/* This sets ->Count */
	PAGE_SET_END(Page, Start);
	/* Done */
	(void)Tree;
}

/* Compares 2 pages on their Start values
 * Must match PUT_TREE_COMPARE_ROUTINE
 */
static s32 range_compare(void *Tree, void *First, void *Second)
{
	struct page_of_ranges *Page1;
	struct page_of_ranges *Page2;
	u64 Start1;
	u64 Start2;

	Page1 = (struct page_of_ranges *)First;
	Page2 = (struct page_of_ranges *)Second;
	/* Compare the first values in the pages */
	Start1 = PAGE_START_ABS(Page1);
	Start2 = PAGE_START_ABS(Page2);
	if (Start1 > Start2)
		/* Below */
		return 1;
	else if (Start1 < Start2)
		/* Above */
		return -1;
	/* Same */
	return 0;
	(void)Tree;
}

/* Inserts the new range of values to the page
 * Returns true usually, and false only if the page has no more room -
 *  the caller should split the page in this case
 * Updates the table's and the page's counts on success
 */
static bool page_of_ranges_insert(struct range_page_table *Table,
				 struct page_of_ranges *Page,
				 URPVALUE Start, URPVALUE End)
{
	u32 OldRangeCount;
	URPVALUE ValuesCountDelta;

	/* Take the current page's range count */
	OldRangeCount = (Page->Count / 2);
	/* Try to insert */
	if (!page_of_ranges_insert_internal(Page, Start, End, &ValuesCountDelta))
		/* Failed */
		return false;
	/* Update the table's range count */
	Table->RangesCount += (Page->Count / 2);
	Table->RangesCount -= OldRangeCount;
	if (ValuesCountDelta != 0) {
		/* The page was really affected by insert
		 * Update the values counts
		 */
		Page->ValuesCount += ValuesCountDelta;
		Table->ValuesCount += ValuesCountDelta;
	}
	/* Done */
	return true;
}

/* Inserts the new range of values to the page
 * Returns true usually, and false only if the page has no more room -
 *  the caller should split the page in this case
 * On success, sets *ValuesCountDelta to the number of values added to the page
 */
static bool page_of_ranges_insert_internal(struct page_of_ranges *Page,
					 URPVALUE Start, URPVALUE End,
					 URPVALUE *ValuesCountDelta)
{
	PURPVALUE Array;
	u32 StartIdx;
	u32 EndIdx;
	u32 i;

	UASSERT(End > Start, "called with zero or negative length");
	/* We never add the tailmost range to the page, see all callers of this function */
	*ValuesCountDelta = 0;
	/* Array of "walls" */
	Array = (PURPVALUE)(Page + 1);
	/* Do the following:
	 *  - find the first range probably touched by insertion, i.e. the last range surely not
	 *  touched, the first range index is in StartIdx
	 *  - find the last range probably touched by insertion, i.e. the first range surely not
	 *  touched, this first-not-touched range index is in EndIdx
	 *  - if Start is within the range, or touches the range's end - pull it down to the
	 *  range's start, so Start is either strictly outside the range (within the hole) or
	 *  at range's start
	 *  - if End is within the range, or touches the range's start - pull it up to the
	 *  range's end, so End is either strictly outside the range (within the hole) or
	 *  at range's end
	 *  - pulling of Start and End will not change the semantics
	 * Find Start in the array
	 */
	StartIdx = page_of_ranges_bsearch(Array, Page->Count, Start);
	if (StartIdx == (u32)(-1)) {
		/* Special case - adding the tailmost range to the page */
		if ((Page->Count + 2) > PAGE_RANGE_MAX) {
			/* Out of room */
			return false;
		}
		/* Now add a range */
		Array[Page->Count] = Start;
		Array[Page->Count + 1] = End;
		Page->Count += 2;
		*ValuesCountDelta += (End - Start);
		/* Done */
		return true;
	}
	/* StartIdx is valid */
	if (StartIdx & 1) {
		/* Odd index - next "wall" after Start is the end of the range
		 * At range's end, or within the range
		 * Pull down to range's start
		 */
		StartIdx--;
		Start = Array[StartIdx];
	} else {
		/* Even index - next "wall" after Start is the start of the range
		 * Do not touch anything
		 */
	}
	if (Start == Array[StartIdx]) {
		/* Start is at range's start and not in the hole
		 * Do a quick check whether End is also in the same range
		 */
		if (End <= Array[StartIdx + 1]) {
			/* End is also in the same range or at it's end
			 * We are inserting to already existing range
			 * No-op
			 */
			return true;
		}
	}
	/* Find End in the array */
	EndIdx = page_of_ranges_bsearch(Array, Page->Count, End);
	if (EndIdx == (u32)(-1)) {
		/* Special case - the range ends beyound anything in this page
		 * We have the first touched range at StartIdx, the result will have 1 range instead
		 *  of the whole tail from StartIdx up
		 * Add the new range to the values count
		 */
		*ValuesCountDelta += (End - Start);
		/* Subtract all to-be-deleted ranges from the values count */
		for (i = StartIdx; i < Page->Count; i += 2)
			*ValuesCountDelta -= (Array[i + 1] - Array[i]);

		/* Form the new range */
		Array[StartIdx] = Start;
		Array[StartIdx + 1] = End;
		Page->Count = (StartIdx + 2);
		/* Done */
		return true;
	}
	/* EndIdx is valid */
	if (EndIdx & 1) {
		/* Odd index - next "wall" after End is the end of the range
		 * At range's end, or within the range
		 * Pull to range's end
		 */
		End = Array[EndIdx];
		EndIdx++;
	} else if (End == Array[EndIdx]) {
		/* Exactly at range's start
		 * Pull to range's end
		 */
		End = Array[++EndIdx];
		EndIdx++;
	} else {
		/* Within the hole
		 * Do not touch anything
		 */
	}
	/* Now Start and End are normalized as described above, and we can do the final operation */
	if (StartIdx == EndIdx) {
		/* No ranges are touched by the insert, we are inserting to the hole without touches
		 *  with the neighbouring ranges
		 * We must stretch the array to make the space for 2 more "walls"
		 */
		if ((Page->Count + 2) > PAGE_RANGE_MAX) {
			/* Out of room */
			return false;
		}
		/* Deal with the values count */
		*ValuesCountDelta += (End - Start);
		/* Move up */
		memmove(&Array[StartIdx + 2], &Array[StartIdx],
				(Page->Count - StartIdx) * sizeof(URPVALUE));
		Page->Count += 2;
		Array[StartIdx] = Start;
		Array[StartIdx + 1] = End;
		/* Done */
		return true;
	}
	/* At least 1 - or maybe more - ranges are touched by the insert
	 * Add the new range to the values count
	 */
	*ValuesCountDelta += (End - Start);
	/* Exclude ranges [StartIdx..EndIdx) (to be replaced or deleted) from the values count */
	for (i = StartIdx; i < EndIdx; i += 2) {
		/* Subtract from the count */
		*ValuesCountDelta -= (Array[i + 1] - Array[i]);
	}
	/* Set the first touched range to [Start..End) - this is possible, Start and End are
	 *  surely not within any ranges - either in the holes or at range start/end
	 */
	Array[StartIdx] = Start;
	Array[StartIdx + 1] = End;
	if ((StartIdx + 2) < EndIdx) {
		/* Remove the covered ranges from the page */
		if (EndIdx < Page->Count) {
			/* Move down */
			memmove(&Array[StartIdx + 2], &Array[EndIdx],
					(Page->Count - EndIdx) * sizeof(URPVALUE));
		}
		Page->Count -= (EndIdx - (StartIdx + 2));
	}
	/* Done */
	return true;
}

/* Deletes the values from the page of ranges
 * Returns true usually, and false only in the rare case if deletion will need to add
 *  a hole and a second range to the page, and the page has no more room -
 *  the caller should split the page in this case
 * Updates the table's and the page's counts on success
 * Must never be called with Start >= largest page value or with End <= first page value
 */
static bool page_of_ranges_delete(struct range_page_table *Table,
				 struct page_of_ranges *Page,
				 URPVALUE Start, URPVALUE End)
{
	u32 OldRangeCount;
	URPVALUE ValuesCountDelta;

	/* Take the current page's range count */
	OldRangeCount = (Page->Count / 2);
	/* Try to delete */
	if (!page_of_ranges_delete_internal(Page, Start, End, &ValuesCountDelta))
		/* Failed */
		return false;
	/* Update the table's range count */
	Table->RangesCount += (Page->Count / 2);
	Table->RangesCount -= OldRangeCount;
	if (ValuesCountDelta != 0) {
		/* The page was really affected by delete
		 * Update the values counts
		 */
		Page->ValuesCount -= ValuesCountDelta;
		Table->ValuesCount -= ValuesCountDelta;
	}
	/* Done */
	return true;
}

/* Deletes the values from the page of ranges
 * Returns true usually, and false only in the rare case if deletion will need to add
 *  a hole and a second range to the page, and the page has no more room -
 *  the caller should split the page in this case
 * On success, sets *ValuesCountDelta to the number of values deleted from the page
 * Must never be called with Start >= largest page value or with End <= first page value
 */
static bool page_of_ranges_delete_internal(struct page_of_ranges *Page,
					 URPVALUE Start, URPVALUE End,
					 URPVALUE *ValuesCountDelta)
{
	PURPVALUE Array;
	u32 StartIdx;
	u32 EndIdx;
	u32 i;

	UASSERT(End > Start, "called with zero or negative length");
	/* We never delete beoynd the page, see all callers of this function */
	*ValuesCountDelta = 0;
	/* Array of "walls" */
	Array = (PURPVALUE)(Page + 1);
	/* Do the following:
	 *  - find the first range probably touched by deletion, i.e. the last range surely not
	 *  touched, the first range index is in StartIdx
	 *  - find the last range probably touched by deletion, i.e. the first range surely not
	 *  touched, this first-not-touched range index is in EndIdx
	 *  - if Start is within the hole, or touches the range's end - pull it up to the
	 *  range's start, so Start is either strictly inside the range or at range's start
	 *  - if End is within the hole, or touches the range's start - pull it down to the
	 *  range's end, so End is either strictly inside the range or at range's end
	 *  - pulling of Start and End will not change the semantics
	 * Find Start in the array
	 */
	StartIdx = page_of_ranges_bsearch(Array, Page->Count, Start);
	/* We never delete after the whole page (Start >= any range end in the page) */
	if (StartIdx == (u32)(-1)) {
		UASSERT(StartIdx != (u32)(-1), "deleting after anything in a page");
		/* Additional safeguard - at least the code will be correct if this will be hit
		 * No-op
		 */
		return true;
	}
	/* StartIdx is valid */
	if ((StartIdx & 1) == 0) {
		/* Within the hole, or touching the range start
		 * Pull to range start
		 */
		Start = Array[StartIdx];
	} else if (Start == Array[StartIdx]) {
		/* Touching the range's end
		 * Pull to next range start
		 */
		StartIdx++;
		if (StartIdx == Page->Count) {
			UASSERT(StartIdx < Page->Count, "deleting just after the last range's end");
			/* Additional safeguard - at least the code will be correct if this will be hit
			 * No-op
			 */
			return true;
		}
		Start = Array[StartIdx];
	} else {
		/* Within the range */
		StartIdx--;
	}
	/* After pull-ups above, Start could become >= End, which means - we are a no-op
	 * Check it now
	 */
	if (Start >= End) {
		/* No ranges are touched by deletion, deleting a hole
		 * No-op
		 */
		return true;
	}
	/* Here StartIdx is the first affected range, and Start is either its start
	 *  or inside it
	 * Find End in the array
	 */
	EndIdx = page_of_ranges_bsearch(Array, Page->Count, End);
	if (EndIdx == (u32)(-1)) {
		/* Special case - deleting from Start to page end
		 * Set End to be the last range's end
		 */
		EndIdx = (Page->Count - 1);
		End = Array[EndIdx];
	}
	/* EndIdx is valid */
	if ((EndIdx & 1) == 0) {
		/* Within the hole, or touching the range start */
		if (EndIdx == 0) {
			UASSERT(EndIdx > 0, "deleting before anything in a page");
			/* Additional safeguard - at least the code will be correct if this will be hit
			 * No-op
			 */
			return true;
		}
		/* Pull to end of previous range */
		End = Array[EndIdx - 1];
	} else {
		/* Within the range, or touching the range's end */
		EndIdx++;
	}
	/* Here:
	 *  - StartIdx is the first affected range, and Start is either its start or inside it
	 *  - EndIdx is the first non-affected range, and End is either the end of the range
	 *  previous to EndIdx or inside this previous range
	 * Note that the pull-down of End above could only occur to the end of the previous
	 *  range without crossing any other ranges, and, since Start was already in a range,
	 *  this pull-down could not cross Start - no need to check for Start < End
	 * Also note that, before the logic affecting End, it was proven that Start is < End,
	 *  so, if we will also apply the rule on StartIdx/Start just above - we can see that
	 *  StartIdx < EndIdx
	 * We can do the final operation
	 */
	if (((StartIdx + 2) == EndIdx) &&
	    /* Only 1 range is touched and... */
	    (Start > Array[StartIdx]) &&
	    /* head will remain and... */
	    (End < Array[StartIdx + 1])
	    /* tail will remain too */
	    ) {
		/* A very special case, carving a hole within a range */
		if ((Page->Count + 2) > PAGE_RANGE_MAX) {
			/* Out of room */
			return false;
		}
		/* Deal with the counter */
		*ValuesCountDelta += (End - Start);
		/* Slide the next ranges up to open the hole for 1 more range (second half) */
		if (EndIdx < Page->Count) {
			/* Move up */
			memmove(&Array[EndIdx + 2], &Array[EndIdx],
					(Page->Count - EndIdx) * sizeof(URPVALUE));
		}
		/* Fill the 2 ranges */
		Array[EndIdx + 1] = Array[StartIdx + 1];
		Array[StartIdx + 1] = Start;
		Array[EndIdx] = End;
		Page->Count += 2;
		/* Done */
		return true;
	}
	/* Check how the first touched range is affected */
	if (Start > Array[StartIdx]) {
		/* Affected partially, trim its tail and skip it
		 * Deal with the counter
		 */
		*ValuesCountDelta += (Array[StartIdx + 1] - Start);
		Array[StartIdx + 1] = Start;
		StartIdx += 2;
		if (StartIdx == EndIdx) {
			/* Nothing else is touched */
			return true;
		}
	}
	/* Here either:
	 *  - Start is the start of StartIdx range
	 *  or
	 *  - Start is the end of pre-StartIdx range, which is actually logically the same
	 * In any case, the deletion touches StartIdx range from its very start
	 * Check how the last touched range is affected
	 */
	if (End < Array[EndIdx - 1]) {
		/* Affected partially, trim its head and skip it */
		EndIdx -= 2;
		/* Deal with the counter */
		*ValuesCountDelta += (End - Array[EndIdx]);
		Array[EndIdx] = End;
		if (StartIdx == EndIdx) {
			/* Nothing else is touched */
			return true;
		}
	}
	/* Here either:
	 *  - End is the end of EndIdx-1 range
	 *  or
	 *  - End is the start of EndIdx range, which is actually logically the same
	 * In any case, the deletion touches EndIdx-1 range till its very end
	 * StartIdx is the first, and EndIdx-1 is the last range
	 *  which is subject for complete deletion
	 * Some ranges are subject for complete deletion - exclude them from the values count
	 */
	for (i = StartIdx; i < EndIdx; i += 2) {
		/* Add to the count */
		*ValuesCountDelta += (Array[i + 1] - Array[i]);
	}
	/* Shrink the page down, collapsing a hole of [StartIdx..EndIdx) */
	if (EndIdx < Page->Count) {
		/* Move down */
		memmove(&Array[StartIdx], &Array[EndIdx],
				(Page->Count - EndIdx) * sizeof(URPVALUE));
	}
	Page->Count -= (EndIdx - StartIdx);
	/* Done */
	return true;
}

/* Performs a binary search over Array
 * Array has Size elements
 * The caller must guarantee that Size is at least 2, and that Array is ordered
 * Returns (u32)(-1) if all elements in Array are < Value
 * Otherwise, returns the index of the first element in Array which is >= Value
 */
static u32 page_of_ranges_bsearch(URPVALUE Array[], u32 Size, URPVALUE Value)
{
	u32 Hi;
	u32 Lo;
	u32 Mid;

	/* Compare with the last element */
	Hi = (Size - 1);
	if (Array[Hi] == Value) {
		return Hi;
	} else if (Array[Hi] < Value) {
		/* All elements are < Value */
		return (u32)(-1);
	}
	/* Value < Array[Hi]
	 * Compare with the first element
	 */
	Lo = 0;
	if (Array[Lo] >= Value)
		return Lo;

	/* Array[Lo] < Value < Array[Hi] */
	while ((Lo + 1) < Hi) {
		/* Array[Lo] < Value < Array[Hi] */
		Mid = ((Lo + Hi) / 2);
		/* Lo < Mid < Hi */
		if (Array[Mid] == Value) {
			return Mid;
		} else if (Array[Mid] < Value) {
			Lo = Mid;
			/* Array[Lo] < Value < Array[Hi] */
		} else {
			/* Value < Array[Mid] */
			Hi = Mid;
			/* Array[Lo] < Value < Array[Hi] */
		}
	}
	/* Lo + 1 == Hi
	 * Array[Lo] < Value < Array[Hi]
	 * Sanity check
	 */
	UASSERT(Value <= Array[Hi], "Value is greater than largest item in array");
	return Hi;
}
