/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#ifndef STCVSM_SNAPSHOT_H
#define STCVSM_SNAPSHOT_H

struct cow_storage;

#define SNAPSHOT_SNAPPED	0
#define SNAPSHOT_INIT		1
#define SNAPSHOT_ERROR		2
#define SNAPSHOT_SHUTDOWN	3
struct snapshot_context {
	/* Refcount */
	atomic_t sc_refs;

	/* Original Device info */
	struct block_device *sc_bdev;
	sector_t sc_sector_start; /* Partition offset */
	sector_t sc_sector_end; /* Partition size */
	struct super_block *sc_frozen_sb;

	/* Pointer back to bdfilter */
	struct block_device_filter *bdfilter;

	/* snapshot block device */
	struct request_queue *sc_snap_queue;
	struct gendisk *sc_snap_gendisk;
	unsigned long sc_flags;
	int sc_snap_minor;

	/* Snapshot Set */
	u8 set_guid[16];
	struct list_head set_peer_list;

	/* Bitmap management*/
	unsigned int block_size;
	unsigned int sectors_per_block;
	unsigned long bits_in_map;
	unsigned long *freespc_bitmap;

	/* COW storage */
	struct cow_storage *cstor;
	unsigned int cow_enospc;

	/* VDiff */
	struct vdiff_context *vdiff;
};

/** Get an existing snapshot, based on a bio.  Increments the reference count.
 * @param bdfilter  The filter to search
 * @param bio  The bio, to match against snapshots
 *
 * @todo clarify boundary issues
 */
struct snapshot_context *get_snapshot_bio(struct block_device_filter *bdfilter,
		struct bio *bio);

/** Get an existing snapshot of the bdev.  Increments the reference count.
 * @param bdfilter  The filter to search
 * @param bdev  The snapshotted device
 */
struct snapshot_context *get_snapshot_bdev(struct block_device_filter *bdfilter,
		struct block_device *bdev);

struct block_device *get_snapshot_filter_bdev(struct block_device *bdev);

int orig_write_filter(struct request_queue *q, struct bio *bio,
		struct block_device_filter *bdfilter, struct snapshot_context *snap);

/** Allocates a new snapshot.  Allocates the snapshot structure and virtual block device, and
 * hooks the request queue of the target device.  Does not yet activate the snapshot.
 *
 * @param bdfilter  The filter of bdev
 * @param bdev  The device to snapshot
 * @param cow_dev  The dev_t of the copy-on-write device, or 0 if none
 */
struct snapshot_context *init_snapshot(struct block_device_filter *bdfilter,
		struct block_device *bdev, dev_t cow_dev);

/** Atomically activates all snapshots in the set.  That is, all filesystems are frozen.
 * Then, each snapshot copies vdiff, is marked active, and all filesystems are thawed.
 *
 * @param snap_list  List of snapshot(s) to activate
 */
int activate_snapshots(struct list_head *snap_list);

/** Releases a snapshot reference.  The reference was acquired from get_snapshot_bio or
 * get_snapshot_bdev.  These references keep the snapshot alive, but for lifetime and
 * ultimate cleanup, see stcvsm_shutdown_snapshot.
 *
 * @param snap  The snapshot to release
 */
void put_snapshot(struct snapshot_context *snap);

/** Shuts down a snapshot and removes it from the filter.  New references to it can no longer be
 * acquired.  Any in-progress reads will get null data.  The snapshot will be destroyed when the
 * final reference is put.
 *
 * @param snap  The snapshot to shut down
 */
void stcvsm_shutdown_snapshot(struct snapshot_context *snap);

const char *snapshot_state_to_str(unsigned long state);

#endif
