/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/semaphore.h>

#include "debug.h"
#include "filter.h"
#include "snapshot.h"
#include "snapshot_set.h"

/* Element of set_list_head below */
struct snap_set {
	u8 guid[16];
	struct list_head list; /* member of set_list_head */
	struct list_head snap_list_head; /* head of snapshot_context list */
	spinlock_t set_lock;
	unsigned long snapped;
};
LIST_HEAD(set_list_head);
static struct semaphore set_list_lock = __SEMAPHORE_INITIALIZER(set_list_lock, 1);

#define SET_GET_CREATE		1
#define SET_GET_NOOP		0
#define SET_GET_REMOVE		-1

static struct snap_set *get_sc_list(u8 guid[16], int behavior)
{
	struct snap_set *set;
	struct snap_set *retval = NULL;
	bool found = false;

	stcvsm_info(LOG_SETS, "%d\n", behavior);
	down(&set_list_lock);
	list_for_each_entry(set, &set_list_head, list) {
		if (memcmp(set->guid, guid, 16) == 0) {
			retval = set;
			found = true;
			break;
		}
	}
	if (found && behavior == SET_GET_REMOVE)
		list_del(&set->list);
	if (!found && behavior == SET_GET_CREATE) {
		set = kmalloc(sizeof(*set), GFP_KERNEL);
		if (set) {
			memcpy(set->guid, guid, 16);
			list_add(&set->list, &set_list_head);
			INIT_LIST_HEAD(&set->snap_list_head);
			spin_lock_init(&set->set_lock);
			set->snapped = 0;
			retval = set;
		} else {
			retval = ERR_PTR(-ENOMEM);
		}
	}
	up(&set_list_lock);
	return retval;
}

static void free_set(struct snap_set *set)
{
	kfree(set);
}

int set_add(u8 guid[16], dev_t bdev_num, dev_t cow_num)
{
	int err = 0;
	struct block_device *bdev;
	struct block_device_filter *bdfilter;
	struct snapshot_context *snap = NULL;

	printk_uuid(notice, "", guid);
	bdev = stcvsm_open_blkdev(bdev_num, FMODE_READ, NULL); /* NOTE: Scope Guard: stcvsm_close_blkdev(bdev) on exit*/
	if (IS_ERR(bdev))
		return PTR_ERR(bdev);
	bdfilter = get_filter(bdev, true); /* NOTE: Scope Guard: put_filter(bdfilter) on exit */
	if (IS_ERR(bdfilter)) {
		stcvsm_close_blkdev(bdev, FMODE_READ);
		stcvsm_warning(LOG_SETS, "Failed to create block device filter (%ld)\n", PTR_ERR(bdfilter));
		return PTR_ERR(bdfilter);
	}
	stcvsm_debug(LOG_SETS, "Creating snapshot\n");
	snap = init_snapshot(bdfilter, bdev, cow_num); /* Adds refcount to bdfilter */
	if (IS_ERR(snap)) {
		err = PTR_ERR(snap);
		stcvsm_err(LOG_SETS, "Failed to create snapshot (%d)\n", err);
	} else {
		struct snap_set *set = get_sc_list(guid, SET_GET_CREATE);

		stcvsm_info(LOG_SETS, "Snapshot initialized: %s\n", snap->sc_snap_gendisk->disk_name);
		if (!set) {
			err = -ENOENT;
		} else if (IS_ERR(set)) {
			err = PTR_ERR(set);
		} else {
			spin_lock(&set->set_lock);
			if (!set->snapped) {
				list_add(&snap->set_peer_list, &set->snap_list_head);
				memcpy(snap->set_guid, guid, 16);
			} else {
				err = -EBUSY;
				stcvsm_err(LOG_SETS, "Set found, but already snapped\n");
			}
			spin_unlock(&set->set_lock);
		}
		if (err) {
			stcvsm_err(LOG_SETS, "Failed to add snapshot (%d)\n", err);
			stcvsm_shutdown_snapshot(snap);
		}
	}
	put_filter(bdfilter);
	stcvsm_close_blkdev(bdev, FMODE_READ);
	stcvsm_info(LOG_SETS, "Done creating snapshot\n");
	return err;
}

int set_snap(u8 guid[16])
{
	int err = 0;
	struct snap_set *set = get_sc_list(guid, SET_GET_NOOP);

	printk_uuid(notice, "", guid);
	if (!set) /* Did you ever call set_add? */
		return -ENODEV;
	spin_lock(&set->set_lock);
	if (set->snapped == 0) {
		set->snapped = 1;
	} else {
		err = -EBUSY;
		stcvsm_err(LOG_SETS, "Set found, but already snapped\n");
	}
	spin_unlock(&set->set_lock);
	if (!err) {
		err = activate_snapshots(&set->snap_list_head);
		if (err)
			set_unsnap(guid);
	}
	return err;
}

int set_unsnap(u8 guid[16])
{
	int err = 0;
	struct snap_set *set = get_sc_list(guid, SET_GET_REMOVE);

	printk_uuid(notice, "", guid);
	if (!set) {
		err = -ENOENT;
	} else if (IS_ERR(set)) {
		err = PTR_ERR(set);
	} else {
		struct snapshot_context *snap;
		struct snapshot_context *next;
		/* "Unsnap the set" is used even for teardown cases, when the
		 * set was not successfully snapped.  Tear down whatever was
		 * found, regardless of the state.
		 */
		list_for_each_entry_safe(snap, next, &set->snap_list_head, set_peer_list) {
			list_del(&snap->set_peer_list);
			stcvsm_shutdown_snapshot(snap);
		}
		free_set(set);
	}
	if (err)
		stcvsm_err(LOG_SETS, "Failed to deactivate snapshot set (%d)\n", err);
	return err;
}
