/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

/*
 * tree table implementation
 * 64bit ready
 */


#include "stcvsm_os.h"

#include "uassert.h"
#include "tree.h"

/* Macros */

/* Colors for red-black tree */
#undef RB_BLACK  /* Linux may have conflicting rbtree */
#undef RB_RED
#define RB_BLACK                        0
#define RB_RED                          1

/* Item pointer from the splay links pointer */
#define ITEM_STRUCT(L)                  container_of(L, struct tree_table_item, links)
/*
 *Item pointer from the splay links pointer
 * Handles NULL correctly
 */
#define ITEM_STRUCT_OR_NULL(L)          ((L != NULL) ? \
					 container_of(L, struct tree_table_item, links) : NULL)
/* Data area pointer from the splay links pointer */
#define ITEM_DATA_AREA(L)               ((void *) ((container_of(L, struct tree_table_item, links)) + 1))
/*
 * Swaps 2 pointers
 * Tmp is the temporary name
 */
#define SWAP_POINTERS(Ptr1, Ptr2)       do { Tmp = (Ptr1); (Ptr1) = (Ptr2); (Ptr2) = Tmp; } while (0)
/*
 * Child pointer by offset
 * This is a Lvalue
 */
#define ITEM_CHILD(Item, Offset)        (*(struct splay_links **)((unsigned char *)Item + Offset))

/* Functions internal for this file */

/*
 * Tree scan comparison-based function
 * Returns an item structure or its parent or NULL if tree is empty
 * Returns child pointer offset via parameter - (u32)(-1) if the exact hit occured
 */
static struct splay_links *tree_scan(struct tree_table *tree, void *Data, u32 *ChildPointerOffset);
/*
 * Internal engine for deletions called by tree_delete and tree_delete_existing
 * Only detaches the item from the tree if DestroyTheItem is false
 */
static void tree_delete_internal(struct tree_table *tree, struct splay_links *Item, bool DestroyTheItem);

#ifdef ENABLE_RED_BLACK
/*
 * Insert handler for a red-black tree
 * Must match tree_handler_fn
 */
static void tree_insert_red_black(struct tree_table_item *Inserted, struct splay_links **root);
/*
 * Delete handler for a red-black tree
 * Must match tree_delete_handler_fn
 */
static void tree_delete_red_black(struct tree_table_item *ParentOfDeleted, struct tree_table_item *Deleted,
			     struct tree_table_item *ChildOfDeleted,
			     u32 OurOffset, u32 BrotherOffset, struct splay_links **root);
#endif

/*
 * Insert handler for an AVL tree
 * Must match tree_handler_fn
 */
static void tree_insert_avl(struct tree_table_item *Inserted, struct splay_links **root);
/*
 * Delete handler for an AVL tree
 * Must match tree_delete_handler_fn
 */
static void tree_delete_avl(struct tree_table_item *ParentOfDeleted, struct tree_table_item *Deleted,
			struct tree_table_item *ChildOfDeleted,
			u32 OurOffset, u32 BrotherOffset, struct splay_links **root);

/*
 * Deleting enumerator
 * Detaches the next leaf item from the tree
 * *Item must be NULL on first call and must be preserved for next calls
 * Returns the next item to delete (detached from the tree) or NULL if no more items
 * *root is updated if necessary
 */
static struct splay_links *tree_delete_next_tree_item(struct splay_links **Item, struct splay_links **root);
/*
 * Swaps 2 tree items which can be immediate parent and child or root, updating all links
 * root must point to a variable which will contain the new tree root if the root
 *  is changed by this
 */
static void tree_swap_tree_items(struct splay_links *Item1, struct splay_links *Item2,
			    struct splay_links **root);
/*
 * Turns a tree, pulling Child up one position
 * root must point to a variable which will contain the new tree root if the root
 *  is changed by this rotation
 */
static void tree_turn_tree(struct splay_links *Child, struct splay_links **root);


/* Initialization API Functions */

/*
 * Initializes a generic tree table with caller-specified pointers and allocator semantics
 */
void tree_initialize_generic(struct tree_table *tree, enum tree_kind Kind, u32 data_area_size,
			     tree_compare_fn compare_func,
			     tree_allocate_fn allocate_func,
			     tree_construct_fn constructor_func,
			     tree_destroy_fn destroy_func,
			     tree_free_fn free_func,
			     tree_cleanup_fn cleanup_func,
			     void *allocator)
{
	/* Fill the tree */
	tree->allocator = allocator;
	tree->allocate_func = allocate_func;
	tree->constructor_func = constructor_func;
	tree->destroy_func = destroy_func;
	tree->free_func = free_func;
	tree->cleanup_func = cleanup_func;
	/* Kind-specific handlers */
	switch (Kind) {
#ifdef ENABLE_RED_BLACK
	case enum_tree_red_black_kind:
		tree->insert_handler_func = tree_insert_red_black;
		tree->delete_handler_func = tree_delete_red_black;
		tree->lookup_handler_func = NULL;
		break;
#endif
	case enum_tree_avl_kind:
		tree->insert_handler_func = tree_insert_avl;
		tree->delete_handler_func = tree_delete_avl;
		tree->lookup_handler_func = NULL;
		break;
	default:
		UASSERTD(0, "tree_initialize_generic: invalid Kind %d\n", Kind);
		break;
	}
	tree->compare_func = compare_func;
	tree->data_area_size = data_area_size;
	/* The tree is empty */
	tree->item_count = 0;
	tree->root = NULL;
	/* Done */
}

/*
 * Initializes a tree table with a nonpaged lookaside allocator provided by the caller
 * lookaside is initialized too
 */
void tree_initialize_npage(struct tree_table *tree, enum tree_kind Kind, u32 data_area_size,
			    tree_compare_fn compare_func,
			    tree_construct_fn constructor_func,
			    tree_destroy_fn destroy_func,
			    struct lookaside_list *lookaside,
			    u32 PoolTag)
{
	/* Initialize the lookaside */
	lookaside_initialize(lookaside, NULL, NULL, 0,
			TreeFullItemSize(data_area_size), PoolTag, 0);
	/* Now initialize the table with lookaside functions */
	tree_initialize_generic(tree, Kind, data_area_size, compare_func,
			(tree_allocate_fn)lookaside_allocate,
			constructor_func, destroy_func,
			(tree_free_fn)lookaside_free,
			(tree_cleanup_fn)lookaside_delete,
			lookaside);
}

/*
 * Same as tree_initialize_npage but using already initialized lookaside
 */
void tree_initialize_npage_with_lookaside(struct tree_table *tree, enum tree_kind Kind,
					  u32 data_area_size,
					  tree_compare_fn compare_func,
					  tree_construct_fn constructor_func,
					  tree_destroy_fn destroy_func,
					  struct lookaside_list *lookaside)
{
	/* Initialize the table with lookaside functions */
	tree_initialize_generic(tree, Kind, data_area_size, compare_func,
			(tree_allocate_fn)lookaside_allocate,
			constructor_func, destroy_func,
			(tree_free_fn)lookaside_free,
			(tree_cleanup_fn)lookaside_delete,
			lookaside);
}


/* Operational API Functions */

/*
 * Copies the tree table from one place to another
 * Src is empty after the copy, and shares the common allocator with Dst
 *  (must only be cleaned up once)
 */
void tree_copy(struct tree_table *Dst, struct tree_table *Src)
{
	/* Copy the whole structure, this is fine */
	*Dst = *Src;
	/* Empty the source */
	Src->item_count = 0;
	Src->root = NULL;
	/* Done */
}

/*
 * Cleans up a tree table, deleting all items and cleaning up an allocator
 * Both allocator and a tree table can be reused then
 */
void tree_cleanup(struct tree_table *tree)
{
	struct splay_links *Item;
	struct splay_links *RestartKey;

	if ((tree->destroy_func != NULL) || (tree->free_func != NULL)) {
		/* We will need to disassemble the tree and delete all items one by one */
		RestartKey = NULL;
		for (;;) {
			Item = tree_delete_next_tree_item(&RestartKey, &(tree->root));
			if (Item == NULL)
				/* No more items */
				break;
			/* Delete this item */
			if (tree->destroy_func != NULL)
				(tree->destroy_func)(tree, ITEM_DATA_AREA(Item));
			/* Free it */
			if (tree->free_func != NULL)
				(tree->free_func)(tree->allocator, ITEM_STRUCT(Item));
		}
	}
	/* No more items to clean up
	 * Shutdown the allocator if necessary
	 */
	if (tree->cleanup_func != NULL)
		(tree->cleanup_func)(tree->allocator);
}

/*
 * Deletes all items from the tree table
 * The table is still initialized and empty after this
 */
void tree_delete_all(struct tree_table *tree)
{
	void *Item;

	/* Delete all items */
	while ((Item = tree_first(tree)) != NULL)
		tree_delete_existing(tree, Item);
	/* Done */
}

/*
 * Inserts an item described by Data to the tree
 * Duplicate is ignored
 * Returns the item inserted or duplicate found or NULL on allocation failure
 * *NewElement is set to true if the item was inserted or to false if duplicate was found
 */
void *tree_insert(struct tree_table *tree, void *Data, bool *NewElement)
{
	struct splay_links *Item;
	u32 ChildPointerOffset = 0; /* Avoid uninitialized error */
	struct tree_table_item *NewItem;

	/* Do the lookup */
	Item = tree_scan(tree, Data, &ChildPointerOffset);
	if ((Item != NULL) && (ChildPointerOffset == (u32)(-1))) {
		/*
		 * A hit
		 * Duplicate
		 */
		if (NewElement != NULL)
			*NewElement = false;
		/* Return the data area */
		return ITEM_DATA_AREA(Item);
	}
	/*
	 * Item is a parent here or NULL if the tree is empty
	 * Allocate a new item
	 */
	NewItem = (struct tree_table_item *)((tree->allocate_func)(tree->allocator));
	if (NewItem == NULL)
		/* Failed */
		return NULL;
	/* Construct if there is a constructor, else just copy the data */
	if (tree->constructor_func != NULL)
		(tree->constructor_func)(tree, NewItem + 1, Data);
	else
		memcpy(NewItem + 1, Data, tree->data_area_size);
	/* This initializes as a root */
	initialize_splay_links(&(NewItem->links));
	/* Attach it to the tree */
	if (Item == NULL)
		/* First item */
		tree->root = &(NewItem->links);
	else {
		/* Attach to the found parent */
		ITEM_CHILD(Item, ChildPointerOffset) = &(NewItem->links);
		NewItem->links.parent = Item;
	}
	/* Now do the kind-specific postprocessing of just attached item
	 * This initializes Value field if necessary and possibly rebalances the tree
	 */
	(tree->insert_handler_func)(NewItem, &(tree->root));
	/* Done */
	if (NewElement != NULL)
		*NewElement = true;
	tree->item_count++;
	/* Return the data area */
	return NewItem + 1;
}

/*
 * Allocates the new item, but does not attach it to the tree
 * The item must later be either attached to the tree by tree_attach,
 *  or deleted by tree_prefree
 * Returns NULL on allocation error
 * The item is not filled at all
 */
void *tree_prealloc(struct tree_table *tree)
{
	struct tree_table_item *NewItem;

	/* Allocate a new item */
	NewItem = (struct tree_table_item *)((tree->allocate_func)(tree->allocator));
	if (NewItem == NULL)
		/* Failed */
		return NULL;
	/* Return the data area */
	return NewItem + 1;
}

/* Undo for tree_prealloc, in a case the item will never be attached to the tree */
void tree_prefree(struct tree_table *tree, void *Item)
{
	if (tree->free_func != NULL)
		(tree->free_func)(tree->allocator, ((struct tree_table_item *)Item) - 1);
}

/*
 * Attaches the item allocated by tree_prealloc to the tree
 * Data is the item's construction data
 * Same as tree_insert but uses preallocated item and thus cannot fail
 * Returns the item inserted or duplicate found
 * If duplicate was found - then returns with *NewElement set to false
 *  and calls tree_prefree(Item) internally
 * Otherwise, really attaches the item to the tree and returns with *NewElement set to true
 */
void *tree_attach(struct tree_table *tree,
		   void *PreallocatedItem,
		   void *Data,
		   bool *NewElement)
{
	struct splay_links *Item;
	u32 ChildPointerOffset = 0; /* Avoid uninitialized error */
	struct tree_table_item *NewItem;

	/* Do the lookup */
	Item = tree_scan(tree, Data, &ChildPointerOffset);
	if ((Item != NULL) && (ChildPointerOffset == (u32)(-1))) {
		/* A hit
		 * Duplicate
		 */
		if (NewElement != NULL)
			*NewElement = false;
		/* No need in a preallocated item */
		tree_prefree(tree, PreallocatedItem);
		/* Return the data area */
		return ITEM_DATA_AREA(Item);
	}
	/*
	 * Item is a parent here or NULL if the tree is empty
	 * Use the preallocated item
	 */
	NewItem = (((struct tree_table_item *)PreallocatedItem) - 1);
	/* Construct if there is a constructor, else just copy the data */
	if (tree->constructor_func != NULL)
		(tree->constructor_func)(tree, NewItem + 1, Data);
	else
		memcpy(NewItem + 1, Data, tree->data_area_size);
	/* This initializes as a root */
	initialize_splay_links(&(NewItem->links));
	/* Attach it to the tree */
	if (Item == NULL)
		/* First item */
		tree->root = &(NewItem->links);
	else {
		/* Attach to the found parent */
		ITEM_CHILD(Item, ChildPointerOffset) = &(NewItem->links);
		NewItem->links.parent = Item;
	}
	/*
	 * Now do the kind-specific postprocessing of just attached item
	 * This initializes Value field if necessary and possibly rebalances the tree
	 */
	(tree->insert_handler_func)(NewItem, &(tree->root));
	/* Done */
	if (NewElement != NULL)
		*NewElement = true;
	tree->item_count++;
	/* Return the data area */
	return NewItem + 1;
}

/*
 * Deletes an item described by Data
 * No-op if no such item
 * Returns true if the item was really deleted, false if it was not found
 */
bool tree_delete(struct tree_table *tree, void *Data)
{
	struct splay_links *Item;
	u32 ChildPointerOffset;

	/* Do the lookup */
	Item = tree_scan(tree, Data, &ChildPointerOffset);
	if ((Item == NULL) || (ChildPointerOffset != (u32)(-1)))
		/* No hit or tree is empty */
		return false;
	/* Found */
	tree_delete_internal(tree, Item, true);
	return true;
}

/*
 * Deletes an item without the tree scan before deletion
 * This function requires Item to be the in-tree item returned by tree_insert, tree_lookup
 *  or tree_enumerate
 */
void tree_delete_existing(struct tree_table *tree, void *Item)
{
	/* Call the engine */
	tree_delete_internal(tree, &((((struct tree_table_item *)Item) - 1)->links), true);
}

/*
 * Detaches the item from the tree but does not delete it
 * Useful if the destroy routine cannot be called under the lock which guards the tree
 *  (for instance, if the tree objects reference one another)
 * This function requires Item to be the in-tree item returned by tree_insert, tree_lookup
 *  or tree_enumerate
 */
void tree_detach(struct tree_table *tree, void *Item)
{
	/* Call the engine */
	tree_delete_internal(tree, &((((struct tree_table_item *)Item) - 1)->links), false);
}

/*
 * Looks up for an item described by Data
 * Returns an item if found or NULL if not found
 */
void *tree_lookup(struct tree_table *tree, void *Data)
{
	struct splay_links *Item;
	u32 ChildPointerOffset;

	/* Do the lookup */
	Item = tree_scan(tree, Data, &ChildPointerOffset);
	if ((Item == NULL) || (ChildPointerOffset != (u32)(-1)))
		/* No hit or tree is empty */
		return NULL;
	/* A hit
	 * Call the lookup handler if any
	 */
	if (tree->lookup_handler_func != NULL)
		(tree->lookup_handler_func)(ITEM_STRUCT(Item), &(tree->root));
	/* Return the data area */
	return ITEM_DATA_AREA(Item);
}

/*
 * Looks up for an item described by Data or the nearest to it in the comparison order
 * Always returns some item except the case if the table is empty
 * *ComparisonResult is set to the comparison result between Data and returned item
 *  < 0 - ReturnedItem < Data
 *  == 0 - exact hit as in tree_lookup
 *  > 0 - ReturnedItem > Data
 */
void *tree_lookup_nearest(struct tree_table *tree, void *Data, s32 *ComparisonResult)
{
	struct splay_links *Item;
	u32 ChildPointerOffset;

	/* Do the lookup */
	Item = tree_scan(tree, Data, &ChildPointerOffset);
	if (Item == NULL)
		/* Tree is empty */
		return NULL;
	/*
	 * If exact match, Item is exact match
	 * If no exact match and ChildPointerOffset == offsetof(struct splay_links, left_child)
	 *  - then we should add Data as a left child of Item, Item is the nearest and is > Data
	 * If no exact match and ChildPointerOffset == offsetof(struct splay_links, right_child)
	 *  - then we should add Data as a right child of Item, Item is the nearest and is < Data
	 * Set ComparisionResult
	 */
	if (ChildPointerOffset == offsetof(struct splay_links, left_child))
		/* Item > Data */
		*ComparisonResult = 1;
	else if (ChildPointerOffset == offsetof(struct splay_links, right_child))
		/* Item < Data */
		*ComparisonResult = -1;
	else
		/* ChildPointerOffset == (u32)(-1) - exact hit */
		*ComparisonResult = 0;
	/* Call the lookup handler if any */
	if (tree->lookup_handler_func != NULL)
		(tree->lookup_handler_func)(ITEM_STRUCT(Item), &(tree->root));
	/* Return Item's data area anyway */
	return ITEM_DATA_AREA(Item);
}

/*
 * Returns the next item in the comparison order or NULL if the given existing tree item
 *  is last
 */
void *tree_next(void *ExistingTreeItem)
{
	struct splay_links *Item;

	/* Get the splay links pointer */
	Item = &((((struct tree_table_item *)ExistingTreeItem) - 1)->links);
	/* Next in the tree */
	Item = tree_real_successor(Item);
	if (Item == NULL)
		/* Last item */
		return NULL;
	/* Return the data area */
	return ITEM_DATA_AREA(Item);
}

/*
 * Returns the previous item in the comparison order or NULL if the given existing tree item
 *  is first
 */
void *tree_prev(void *ExistingTreeItem)
{
	struct splay_links *Item;

	/* Get the splay links pointer */
	Item = &((((struct tree_table_item *)ExistingTreeItem) - 1)->links);
	/* Previous in the tree */
	Item = tree_real_predecessor(Item);
	if (Item == NULL)
		/* First item */
		return NULL;
	/* Return the data area */
	return ITEM_DATA_AREA(Item);
}

/*
 * Returns the first tree item in the comparison order or NULL if the tree is empty
 */
void *tree_first(struct tree_table *tree)
{
	struct splay_links *Item;

	if (tree->root == NULL)
		/* Tree is empty */
		return NULL;
	/* Go left till the next link will be NULL */
	for (Item = tree->root; LeftChild(Item) != NULL; Item = LeftChild(Item))
		/* Empty body */
		;
	/* Return the data area */
	return ITEM_DATA_AREA(Item);
}

/*
 * Returns the last tree item in the comparison order or NULL if the tree is empty
 */
void *tree_last(struct tree_table *tree)
{
	struct splay_links *Item;

	if (tree->root == NULL)
		/* Tree is empty */
		return NULL;
	/* Go right till the next link will be NULL */
	for (Item = tree->root; RightChild(Item) != NULL; Item = RightChild(Item))
		/* Empty body */
		;
	/* Return the data area */
	return ITEM_DATA_AREA(Item);
}

/*
 * Tree enumerator
 * Returns items in the comparison order
 * Does not do any balancing reorders
 * Returns the next item if found or NULL if no more items
 * *RestartKey must be set to NULL to get the first item, and must be preserved for future
 *  calls
 */
void *tree_enumerate(struct tree_table *tree, void **RestartKey)
{
	struct splay_links *Item;

	/* Check for the tree being non-empty */
	if (tree->root == NULL)
		return NULL;
	/* Check if we want to take the first item */
	if (*RestartKey == NULL) {
		/* Take the leftmost item */
		for (Item = tree->root; LeftChild(Item) != NULL; Item = LeftChild(Item))
			/* Empty body */
			;
		/*
		 * Set it as a restart key
		 * It is nonNULL
		 */
		*RestartKey = Item;
	} else {
		/* Find the successor for the item described by RestartKey */
		Item = tree_real_successor((struct splay_links *)(*RestartKey));
		if (Item != NULL)
			/* Set it as a restart key */
			*RestartKey = Item;
	}
	if (Item == NULL)
		/* Not found */
		return NULL;
	/* Return the data area */
	return ITEM_DATA_AREA(Item);
}

/* Functions internal for this file */

/*
 * Tree scan comparison-based function
 * Returns an item structure or its parent or NULL if tree is empty
 * Returns child pointer offset via parameter - (u32)(-1) if the exact hit occured
 */
static struct splay_links *tree_scan(struct tree_table *tree, void *Data,
				   u32 *ChildPointerOffset)
{
	struct splay_links *Item;
	s32 CompareResult;

	/* Start from root */
	Item = tree->root;
	if (Item == NULL)
		/* Tree is empty */
		return NULL;
	/* Go down with comparisons */
	for (;;) {
		/* Compare */
#ifdef FAST_UT_TREE
		if (tree->compare_func == NULL)
			CompareResult = *((s32 *)(ITEM_DATA_AREA(Item))) - *((s32 UNALIGNED *)Data);
		else
		/* Call the function */
#endif
		CompareResult = (tree->compare_func)(tree, ITEM_DATA_AREA(Item), Data);
		if (CompareResult == 0) {
			/* A hit */
			*ChildPointerOffset = (u32)(-1);
			break;
		}
		/* Not found - set the child pointer offset - left or right */
		*ChildPointerOffset = (CompareResult < 0) ?
					/* Item < Data */
					offsetof(struct splay_links, right_child) :
					/* Item > Data */
					offsetof(struct splay_links, left_child);
		/* Check whether we can step on the child */
		if (ITEM_CHILD(Item, *ChildPointerOffset) == NULL)
			/*
			 * Cannot do the next step
			 * ChildPointerOffset is set
			 * Quit and return parent
			 */
			break;
		/* Do the next step */
		Item = ITEM_CHILD(Item, *ChildPointerOffset);
	}
	/*
	 * ChildPointerOffset is set
	 * Quit and return parent
	 */
	return Item;
}

/*
 * Internal engine for deletions called by tree_delete and tree_delete_existing
 * Only detaches the item from the tree if DestroyTheItem is false
 */
static void tree_delete_internal(struct tree_table *tree, struct splay_links *Item, bool DestroyTheItem)
{
	struct splay_links *Successor;
	struct splay_links *parent;
	s32 SavedValue;
	/* Initialized only to suppress a warning */
	u32 OurOffset = 0;
	u32 BrotherOffset = 0;

	/*
	 * If this item has both children - swap it with its subtree successor and delete the
	 *  new successor
	 * The successor surely has not more than 1 child (easy to delete), and the swap introduces
	 *  only 1 disorder which is never mind since one of the disordered items is going to die
	 * The tree-balancing kind-dependent engine considers the tree shape only, not the
	 *  comparison order or data, so, the swap will not confuse it, provided we have
	 *  swapped the Value field (which is connected to the tree position and not item data
	 *  or comparison order)
	 */
	if ((LeftChild(Item) != NULL) && (RightChild(Item) != NULL)) {
		/*
		 * Find the successor
		 * It will not be NULL
		 */
		Successor = tree_subtree_successor(Item);
		/*
		 * Now swap them
		 * Item will have the tree position of Successor (not more then 1 child) and its Value
		 *  which is connected to a tree position
		 * Successor will have the tree position of Item and its Value
		 * Swap Values
		 */
		SavedValue = ITEM_STRUCT(Successor)->value;
		ITEM_STRUCT(Successor)->value = ITEM_STRUCT(Item)->value;
		ITEM_STRUCT(Item)->value = SavedValue;
		/* Now swap links */
		tree_swap_tree_items(Item, Successor, &(tree->root));
	}
	/*
	 * Item has no more then 1 child here
	 * Set Successor to a child if any
	 */
	Successor = LeftChild(Item);
	if (Successor == NULL)
		Successor = RightChild(Item);
	/* Detach from the tree */
	if (IsRoot(Item)) {
		/*
		 * We are deleting the root
		 * New root - it is possibly NULL
		 */
		tree->root = Successor;
		if (Successor != NULL)
			/* A root */
			Successor->parent = Successor;
		/* Pass NULL parent to rebalancing engine */
		parent = NULL;
	} else {
		/* Not a root */
		parent = Parent(Item);
		/* Set the parenting direction offsets and update the child in Parent to Successor */
		if (IsLeftChild(Item)) {
			OurOffset = offsetof(struct splay_links, left_child);
			BrotherOffset = offsetof(struct splay_links, right_child);
			parent->left_child = Successor;
		} else {
			OurOffset = offsetof(struct splay_links, right_child);
			BrotherOffset = offsetof(struct splay_links, left_child);
			parent->right_child = Successor;
		}
		/* Update the parent in Successor if any */
		if (Successor != NULL)
			Successor->parent = parent;
	}
	/* Call the rebalancing engine */
	if (tree->delete_handler_func != NULL)
		(tree->delete_handler_func)(ITEM_STRUCT_OR_NULL(parent), ITEM_STRUCT(Item),
				ITEM_STRUCT_OR_NULL(Successor),
				OurOffset, BrotherOffset, &(tree->root));
	if (DestroyTheItem) {
		/* Destroy and delete the item */
		if (tree->destroy_func != NULL)
			(tree->destroy_func)(tree, ITEM_DATA_AREA(Item));
		/* Free it */
		if (tree->free_func != NULL)
			(tree->free_func)(tree->allocator, ITEM_STRUCT(Item));
	}
	/* Done */
	tree->item_count--;
}

#ifdef ENABLE_RED_BLACK
/*
 * Insert handler for a red-black tree
 * Must match tree_handler_fn
 */
static void tree_insert_red_black(struct tree_table_item *Inserted, struct splay_links **root)
{
	struct splay_links *Item;
	struct splay_links *parent;
	struct splay_links *Uncle;
	struct splay_links *Tmp;

	/*
	 * Red-black tree rules:
	 *
	 *  1) root is black
	 *  2) parent and a child cannot be both red
	 *  3) the number of black items on any path down from any item to any leaf is constant
	 *      for an item called "black height"
	 *
	 * It is easily proven that the tree with this conditions is not more than 2 times
	 *  debalanced:
	 *
	 * A.
	 * 1 item tree is either black (BH = 1) or red (BH = 0), thus (1 << BH) - 1 is 0 or 1
	 * Anyway its number of items = 1 >= (1 << BH) - 1
	 * B.
	 * If we have 2 trees both of the same BH, and build a new tree of them by adding a root,
	 *  then the NewTreeBH = BH + 1 if the root is black, and NewTreeBH = BH is it is red.
	 *  Number of items in the new tree is >= (1 << BH) - 1 + (1 << BH) - 1 + 1 or
	 *  (1 << (BH + 1)) - 1 which is >= (1 << NewTreeBH) - 1
	 *  So, number of items for a new tree is again >= (1 << BH) - 1
	 * C.
	 * By induction from A and B - ANY red-black tree has a number of items >= (1 << BH) - 1
	 * D.
	 * We cannot have 2 adjacent red items, so, BH is >= H/2, where H is a path length
	 * Having H > BH * 2 would require to have 2 adjacent red items.
	 * E.
	 * So, number of items N is >= (1 << H/2) - 1, or, by simple maths, H <= 2 * lg(N + 1)
	 * F.
	 * For a balanced tree, H ~ lg(N + 1) + 1, so, red-black tree cannot be more than 2 times
	 *  debalanced. QED.
	 *
	 * Red-black tree properties:
	 *
	 * 1.1) Each black item which is not a root has a brother
	 *  Otherwise, there would be a BH debalance at the parent item
	 *  So, any brother-less item is red
	 * 1.2) Each red item either has no children, or has 2 black ones
	 *  Followed from 1.1 and Rule 2
	 *
	 * Now we will decide the new item color and enforce the red-black rules over the whole
	 *  tree, possibly by turns
	 */
	if (IsRoot(&(Inserted->links))) {
		/* Color root as black and quit */
		Inserted->value = RB_BLACK;
		return;
	}
	/*
	 * Adding a non-root item
	 * We will color it red, this will guarantee that no black heights have changed
	 */
	Inserted->value = RB_RED;
	Item = &(Inserted->links);
	for (;;) {
		/*
		 * Item is red (and thus not a root), all black heights are OK
		 * However, we could violate the "no 2 red parent-child items" condition
		 */
		parent = Parent(Item);
		if (ITEM_STRUCT(parent)->value == RB_BLACK) {
			/*
			 * parent is black
			 * The tree is still fine
			 * Done
			 */
			return;
		}
		if (parent == NULL)
			/* Will never be executed, just to satisfy PREfast */
			return;
		/*
		 * We have Item and parent both red, which is a condition violation
		 * The Brother is either black or does not exist
		 * Since root is always black, parent is not root, then Granny exists (and is black),
		 *  and possibly even Uncle exists
		 */
		Uncle = IsLeftChild(parent) ? Parent(parent)->right_child :
			Parent(parent)->left_child;
		if ((Uncle == NULL) || (ITEM_STRUCT(Uncle)->value == RB_BLACK))
			/* No Uncle or black Uncle */
			break;
		/*
		 * Existing red Uncle
		 * Color both red parent and its red brother black, and black Granny - red
		 */
		ITEM_STRUCT(parent)->value = RB_BLACK;
		ITEM_STRUCT(Uncle)->value = RB_BLACK;
		ITEM_STRUCT(Parent(parent))->value = RB_RED;
		/*
		 * This would solve "red Item and red parent" problem and will not change
		 *  any black heights, but could introduce the problem for Granny - we colored it
		 *  red, so, it could violate the "no 2 red parent-child items" condition for Granny
		 * We must re-check this and possibly re-run the logic for Granny (which we will name
		 *  Item)
		 */
		Item = Parent(parent);
		/* First check whether Granny is the root */
		if (IsRoot(Item)) {
			/*
			 * root
			 * Just color it black - this will increase all black heights by 1 and
			 *  will deal with all condition violations
			 */
			ITEM_STRUCT(Item)->value = RB_BLACK;
			/* Done */
			return;
		}
		/*
		 * Granny is not a root, is red and possible has a red parent
		 * Check for this violation by running the same again for Granny
		 */
	}
	/*
	 * Red non-root Item, red non-root parent, black Granny, and no Uncle or black Uncle
	 * Check whether the "childing direction" for Item and parent is different
	 */
	if ((IsLeftChild(Item) && IsRightChild(parent)) ||
	    (IsRightChild(Item) && IsLeftChild(parent))) {
		/*
		 * Different
		 * Make a turn to pull Item up
		 * Since Item and parent (which will go down) are both red
		 *  this do not change anything
		 */
		tree_turn_tree(Item, root);
		/* Swap parent and Item for a parent to be really parent */
		SWAP_POINTERS(parent, Item);
	}
	/*
	 * Red non-root Item, red non-root parent, black Granny, no Uncle or black Uncle,
	 *  no Brother or black Brother, and the "childing direction" for both Item and
	 *  parent is the same
	 * Item is Granny
	 */
	Item = Parent(parent);
	/* Make a turn to pull parent up */
	tree_turn_tree(parent, root);
	/*
	 * Red parent, its children are red original Item and black Granny who is now Item
	 *  Granny's children are former Brother and former Uncle
	 * Color the parent black and the Granny red - this will get rid of both 2 adjacent
	 *  red items and of different black heights
	 */
	ITEM_STRUCT(parent)->value = RB_BLACK;
	ITEM_STRUCT(Item)->value = RB_RED;
	/*
	 * We have a good tree again with a maximum of 2 turns
	 * Done
	 */
}

/*
 * Delete handler for a red-black tree
 * Must match tree_delete_handler_fn
 */
static void tree_delete_red_black(struct tree_table_item *ParentOfDeleted, struct tree_table_item *Deleted,
			     struct tree_table_item *ChildOfDeleted,
			     u32 OurOffset, u32 BrotherOffset, struct splay_links **root)
{
	struct splay_links *Brother;
	struct splay_links *InnerNephew;
	struct splay_links *OuterNephew;
	struct splay_links *Item;

	if (ParentOfDeleted == NULL) {
		/* Deleting a root will lead to the empty tree and is OK */
		return;
	}
	/* Deleted has a maximum of 1 child, and it is ChildOfDeleted */
	if (Deleted->value == RB_RED) {
		/*
		 * Deleting a red item does not introduce any rules violation
		 * Done
		 */
		return;
	}
	/*
	 * Deleted is black
	 * If the child exists and red - then we can color the child black and quit
	 * No rules are violated by this
	 */
	if ((ChildOfDeleted != NULL) && (ChildOfDeleted->value == RB_RED)) {
		/* Color it black and quit */
		ChildOfDeleted->value = RB_BLACK;
		return;
	}
	/*
	 * Deleted is black, ChildOfDeleted is black or does not exist
	 * Deletion will conceptually make ChildOfDeleted as "double black", even if it does not
	 *  exist
	 * Initial Item is Deleted
	 */
	Item = &(Deleted->links);
	/*
	 * Item is black
	 * Also - the Brother exists due to 1.1 tree property
	 */
	for (;;) {
		/* Get the Brother */
		Brother = ITEM_CHILD(&(ParentOfDeleted->links), BrotherOffset);
		if (ITEM_STRUCT(Brother)->value == RB_RED) {
			/*
			 * Brother is red
			 * ParentOfDeleted is black
			 * Both Nephews are black
			 * Non-existent Nephews are impossible since this will violate the BH balance
			 *  on ParentOfDeleted
			 * Pull the Brother up, then color it black, and ParentOfDeleted (who is now its
			 *  child) - red
			 * We will have - parent (ParentOfDeleted) is red, Brother (former inner Nephew)
			 *  is black, Granny (former Brother) is black, and Uncle (former outer Nephew)
			 *  is black
			 * The same BH balance and no "2 red nodes" rule violation, but the
			 *  new Brother - which was the inner Nephew - black
			 */
			tree_turn_tree(Brother, root);
			ITEM_STRUCT(Brother)->value = RB_BLACK;
			ParentOfDeleted->value = RB_RED;
			/* Use the new Brother - the former inner Nephew */
			Brother = ITEM_CHILD(&(ParentOfDeleted->links), BrotherOffset);
		}
		/*
		 * Brother is black
		 * Take Nephews
		 */
		InnerNephew = ITEM_CHILD(Brother, OurOffset);
		OuterNephew = ITEM_CHILD(Brother, BrotherOffset);
		if (((InnerNephew != NULL) && (ITEM_STRUCT(InnerNephew)->value == RB_RED)) ||
		    ((OuterNephew != NULL) && (ITEM_STRUCT(OuterNephew)->value == RB_RED)))
			/* Some of the Nephews is red */
			break;
		/*
		 * Both Nephews are either nonexistent or black
		 * OK to color the Brother red - will not violate the Brother-Nephew relation
		 *  and will compensate our BH decrement
		 */
		ITEM_STRUCT(Brother)->value = RB_RED;
		if (ParentOfDeleted->value == RB_RED) {
			/*
			 * OK to color the parent black - this renders the tree back in shape by
			 *  migrating one of the "2 black charges" to the red parent
			 */
			ParentOfDeleted->value = RB_BLACK;
			/* Done */
			return;
		}
		/*
		 * ParentOfDeleted is black
		 * Move the "double blackness" from the child to the ParentOfDeleted conceptually
		 *  and repeat everything for a new "double black" node
		 * Check it to not be the root
		 */
		Item = &(ParentOfDeleted->links);
		if (IsRoot(Item)) {
			/* Just forget the "double blackness" of the root and quit */
			return;
		}
		/* Update ParentOfDeleted and try once more */
		ParentOfDeleted = ITEM_STRUCT(Parent(Item));
		/* Update the direction */
		if (IsLeftChild(Item)) {
			OurOffset = offsetof(struct splay_links, left_child);
			BrotherOffset = offsetof(struct splay_links, right_child);
		} else {
			OurOffset = offsetof(struct splay_links, right_child);
			BrotherOffset = offsetof(struct splay_links, left_child);
		}
		/* Try once more for this node */
	}
	/* Some of the Nephews is red */
	if ((OuterNephew == NULL) || (ITEM_STRUCT(OuterNephew)->value == RB_BLACK)) {
		/*
		 * OuterNephew is nonexistent or black
		 * InnerNephew is red
		 * Pull InnerNephew up, color it black, and its former parent Brother - red
		 * Safe to color the former parent red since its new chilren - OuterNephew
		 *  and InnerNephew's former outer child - are both black
		 * Black heights are preserved too
		 * So, a safe operation
		 */
		tree_turn_tree(InnerNephew, root);
		ITEM_STRUCT(InnerNephew)->value = RB_BLACK;
		ITEM_STRUCT(Brother)->value = RB_RED;
		/* OuterNephew is former Brother, now red */
		OuterNephew = Brother;
		/* The new Brother, now black */
		Brother = InnerNephew;
	}
	/*
	 * OuterNephew is red
	 * Pull the Brother up
	 */
	tree_turn_tree(Brother, root);
	/*
	 * Color the pulled up Brother the same color as was its previous parent
	 * This will never violate the "2 red" condition
	 */
	ITEM_STRUCT(Brother)->value = ParentOfDeleted->value;
	/* Color the previous parent as black */
	ParentOfDeleted->value = RB_BLACK;
	/* Color the OuterNephew as black */
	ITEM_STRUCT(OuterNephew)->value = RB_BLACK;
	/*
	 * We have got rid of the extra blackness
	 * We have a good tree again with a maximum of 3 turns
	 * Done
	 */
}
#endif

/*
 * Insert handler for an AVL tree
 * Must match tree_handler_fn
 */
static void tree_insert_avl(struct tree_table_item *Inserted, struct splay_links **root)
{
	struct splay_links *Child;
	struct splay_links *Item;
	struct splay_links *parent;
	s32 BalanceValue;

	/* The Value for just-inserted item is zero, since it has no children */
	Inserted->value = 0;
	/* Start the way up */
	Item = &(Inserted->links);
	for (;;) {
		/* Move to the parent */
		parent = Parent(Item);
		if (parent == Item) {
			/*
			 * Item was root
			 * End of the way up
			 */
			return;
		}
		/* Get the balance value */
		BalanceValue = IsLeftChild(Item) ? -1 : 1;
		if (ITEM_STRUCT(parent)->value != 0)
			/* Imbalanced parent */
			break;
		/*
		 * Balanced parent
		 * We are still going up the ideally balanced tree started at where we added an item
		 * Add the balance value and continue the way up
		 */
		ITEM_STRUCT(parent)->value = BalanceValue;
		Item = parent;
	}
	/* Imbalanced parent */
	ITEM_STRUCT(parent)->value += BalanceValue;

	/* Value will be either 0 or +-2 */
	UASSERTD(((ITEM_STRUCT(parent)->value >= -2) && (ITEM_STRUCT(parent)->value <= 2)),
		"invalid Value", ITEM_STRUCT(parent)->value);

	/* If we did not cause 2 imbalances - done */
	if (ITEM_STRUCT(parent)->value == 0)
		return;

#if 0
	Value is + -2
	Serious imbalance - we will turn the tree
	The picture is :

	P
	/ \
	I B
	/ \
	/  \
	OC IC
	/ \
	/   \
	ICOC ICIC
#endif
	/* Either old parent->value and BalanceValue were -1, or both were 1 */
	if (ITEM_STRUCT(Item)->value == BalanceValue) {
		/*
		 * Outer child - this is where an item was added - is 1 heavier then inner child
		 * Item is 2 heavier then brother - outer child is 1 heavier then brother
		 * Brother and inner child are same weight
		 * Pull our Item up and set the Value for both Item and its former parent as 0,
		 *  this will lead to a balanced tree
		 */
		tree_turn_tree(Item, root);
		ITEM_STRUCT(Item)->value = 0;
		ITEM_STRUCT(parent)->value = 0;
		/* Done */
		return;
	}
	/* Value must not be 0 */
	UASSERTP(ITEM_STRUCT(Item)->value == -BalanceValue, "serious imbalance on balanced Item", Item);
	/*
	 * Inner child - this is where an item was added - is 1 heavier then outer child
	 * Item is 2 heavier then brother - inner child is 1 heavier then brother
	 * Brother and outer child are same weight
	 * We will pull inner child up 2 times, this will cause the picture of:
	 */
#if 0
	IC
	/  \
	/    \
	/      \
	I P
	/ \        / \
	/  \      /  \
	OC ICOC ICIC B
#endif
	/* Take the inner child */
	Child = (BalanceValue == -1) ? RightChild(Item) : LeftChild(Item);
	/* Do the 2 turns to the above picture */
	tree_turn_tree(Child, root);
	tree_turn_tree(Child, root);
	if (ITEM_STRUCT(Child)->value == BalanceValue) {
		/*
		 * ICOC is 1 heavier then ICIC
		 *  OC, B and ICOC are the same weight, B is 1 heavier then ICIC
		 * Set Value of I to 0 and of P to 1
		 *  (or vice versa for mirrored picture)
		 */
		ITEM_STRUCT(Item)->value = 0;
		ITEM_STRUCT(parent)->value = -BalanceValue;
	} else if (ITEM_STRUCT(Child)->value == -BalanceValue) {
		/*
		 * ICIC is 1 heavier then ICOC
		 *  OC, B and ICIC are the same weight, OC is 1 heavier then ICOC
		 * Set Value of I to -1 and of P to 0
		 *  (or vice versa for mirrored picture)
		 */
		ITEM_STRUCT(Item)->value = BalanceValue;
		ITEM_STRUCT(parent)->value = 0;
	} else {
		/* ICIC and ICOC - and thus also OC and B - are the same weight */
		ITEM_STRUCT(Item)->value = 0;
		ITEM_STRUCT(parent)->value = 0;
	}
	/* IC is always balanced */
	ITEM_STRUCT(Child)->value = 0;
	/*
	 *  We have a good tree again with a maximum of 2 turns
	 * Done
	 */
}

/*
 * Delete handler for an AVL tree
 * Must match tree_delete_handler_fn
 */
static void tree_delete_avl(struct tree_table_item *ParentOfDeleted, struct tree_table_item *Deleted,
			struct tree_table_item *ChildOfDeleted,
			u32 OurOffset, u32 BrotherOffset, struct splay_links **root)
{
	struct splay_links *parent;
	struct splay_links *Brother;
	struct splay_links *Nephew;
	s32 BalanceValue;

	if (ParentOfDeleted == NULL) {
		/* Deleting a root will lead to the empty tree and is OK */
		return;
	}
	/* Get the initial balance value delta and the parent */
	BalanceValue = (OurOffset == offsetof(struct splay_links, left_child)) ? 1 : -1;
	parent = &(ParentOfDeleted->links);
	/* Start the way up */
	for (;;) {
		/* Update the parent's Value */
		ITEM_STRUCT(parent)->value += BalanceValue;
		if (ITEM_STRUCT(parent)->value == BalanceValue) {
			/*
			 * The parent was balanced, its weight did not changed
			 * Done
			 */
			break;
		}
		if (ITEM_STRUCT(parent)->value != 0) {
			/*
			 * 2 * BalanceValue
			 * The parent became imbalanced
			 * Check for internal inconsistency
			 */
			UASSERTD((ITEM_STRUCT(parent)->value == -2) || (ITEM_STRUCT(parent)->value == 2),
				"bad Value field", ITEM_STRUCT(parent)->value);
			/*
			 * Get the Brother
			 * It is guaranteed to be nonNULL since the tree was biased this side before
			 *  deletion
			 */
			Brother = ITEM_CHILD(parent, BrotherOffset);
			if (ITEM_STRUCT(Brother)->value == 0) {
				/*
				 * Brother is balanced
				 * Pull the Brother up
				 */
				tree_turn_tree(Brother, root);
				/*
				 *  This changed the parent's imbalance to be 1 smaller
				 *  (Nephew instead of Brother)
				 */
				ITEM_STRUCT(parent)->value = BalanceValue;
				/*
				 * Brother becomes 1 imbalanced to the other side since it is parent's parent
				 *  now
				 */
				ITEM_STRUCT(Brother)->value = -BalanceValue;
				/*
				 * Brother's weight is now the same as was parent's weight
				 * Done with the tree
				 */
				break;
			} else if (ITEM_STRUCT(Brother)->value == BalanceValue) {
				/*
				 * Brother's outer side is heavier
				 * Pull the Brother up
				 */
				tree_turn_tree(Brother, root);
				/* parent is now balanced, so is Brother */
				ITEM_STRUCT(parent)->value = 0;
				ITEM_STRUCT(Brother)->value = 0;
				/*
				 * The tree weight decremented by 1
				 * Continue the way up as if we deleted the item which is now parent
				 */
				parent = Brother;
			} else {
				/*
				 * ITEM_STRUCT(Brother)->value == -BalanceValue
				 * Brother's inner side is heavier
				 * Get inner Nephew
				 */
				Nephew = ITEM_CHILD(Brother, OurOffset);
#if 0
				The tree now is :
				P
				/ \
				L B
				/ \
				IN ON
				/ \
				INI INO

				IN is 1 heavier then ON
				B is 1 heavier then IN and 2 heavier then ON
				P is 1 heavier then B - 2 then IN and 3 then ON
				P is 3 heavier then L - L is same weight as ON
				Pull the Nephew up twice
#endif
				tree_turn_tree(Nephew, root);
				tree_turn_tree(Nephew, root);
#if 0
				The tree now is :

				IN
				/  \
				/    \
				P B
				/ \    / \
				/   \  /   \
				L INI INO ON

#endif
				if (ITEM_STRUCT(Nephew)->value == BalanceValue) {
					/*
					 * INO is heavier then INI
					 * INO, ON and L are the same weight, INI is lighter
					 * parent's our side is heavier, Brother is balanced
					 */
					ITEM_STRUCT(parent)->value = -BalanceValue;
					ITEM_STRUCT(Brother)->value = 0;
				} else if (ITEM_STRUCT(Nephew)->value == -BalanceValue) {
					/*
					 * INI is heavier then INO
					 * INI, ON and L are the same weight, INO is lighter
					 * parent is balanced, Brother's their side is heavier
					 */
					ITEM_STRUCT(parent)->value = 0;
					ITEM_STRUCT(Brother)->value = BalanceValue;
				} else {
					/*
					 * INI is same weight as INO
					 * Same weight are ON and L
					 * parent is balanced, Brother is balanced
					 */
					ITEM_STRUCT(parent)->value = 0;
					ITEM_STRUCT(Brother)->value = 0;
				}
				/* Nephew is always balanced */
				ITEM_STRUCT(Nephew)->value = 0;
				/*
				 * The tree weight decremented by 1
				 * Continue the way up as if we deleted the item which is now parent
				 */
				parent = Nephew;
			}
		}
		/*
		 * else
		 * New Value is 0
		 * The parent's weight decremented, but no severe imbalance occured
		 * Continue the way up as if we deleted the item which is now parent
		 */
		if (IsRoot(parent)) {
			/* End of the way up */
			break;
		}
		/* Take the next parent, its BalanceValue delta and childing offsets */
		if (IsLeftChild(parent)) {
			BalanceValue = 1;
			OurOffset = offsetof(struct splay_links, left_child);
			BrotherOffset = offsetof(struct splay_links, right_child);
		} else {
			BalanceValue = -1;
			OurOffset = offsetof(struct splay_links, right_child);
			BrotherOffset = offsetof(struct splay_links, left_child);
		}
		/* Now the next parent */
		parent = Parent(parent);
	}
	/* Done */
	(void)Deleted;
	(void)ChildOfDeleted;
}

/*
 * Deleting enumerator
 * Detaches the next leaf item from the tree
 * *Item must be NULL on first call and must be preserved for next calls
 * Returns the next item to delete (detached from the tree) or NULL if no more items
 * *root is updated if necessary
 */
static struct splay_links *tree_delete_next_tree_item(struct splay_links **Item, struct splay_links **root)
{
	struct splay_links *ItemToDelete;

	if (*root == NULL)
		/* Tree it empty */
		return NULL;
	ItemToDelete = *Item;
	if (ItemToDelete == NULL)
		/* First call, start scanning from the root */
		ItemToDelete = *root;
	/* Do the scan for the next leaf item, starting with ItemToDelete */
	for (;;) {
		/* First go left */
		while (LeftChild(ItemToDelete) != NULL)
			ItemToDelete = LeftChild(ItemToDelete);
		/* Here LeftChild(ItemToDelete) is NULL */
		if (RightChild(ItemToDelete) == NULL)
			/* Found a leaf */
			break;
		/* Go one step right, then repeat going left */
		ItemToDelete = RightChild(ItemToDelete);
	}
	/* Here ItemToDelete is a leaf */
	if (!IsRoot(ItemToDelete)) {
		/* *Item (restart key) must be set to its parent */
		*Item = Parent(ItemToDelete);
		/* Detach from the parent */
		if (IsLeftChild(ItemToDelete))
			(*Item)->left_child = NULL;
		else
			(*Item)->right_child = NULL;
	} else
		/*
		 * No more tree root - this is the last item
		 * *Item is ignored
		 */
		*root = NULL;
	/* Done */
	return ItemToDelete;
}

/*
 * Swaps 2 tree items which can be immediate parent and child or root, updating all links
 * root must point to a variable which will contain the new tree root if the root
 *  is changed by this
 */
static void tree_swap_tree_items(struct splay_links *Item1, struct splay_links *Item2,
			    struct splay_links **root)
{
	struct splay_links *Tmp;

	/* If Item2 is root or Item2 is Item1's parent - swap the variables for simplicity */
	if (IsRoot(Item2) || (Parent(Item1) == Item2))
		SWAP_POINTERS(Item1, Item2);
	/*
	 * Here can be:
	 *  1) Item1 is root AND a parent of Item2
	 *  2) Item1 is root AND NOT a parent of Item2
	 *  3) Item1 is NOT root AND a parent of Item2
	 *  4) Item1 is NOT root AND NOT a parent of Item2
	 * Item2 cannot be root or a parent of Item1, thus, the last case is 2 independent items
	 */
	if (Parent(Item2) == Item1) {
		if (IsRoot(Item1)) {
			/*
			 * Case 1)
			 * Update the root
			 */
			*root = Item2;
			/* Make Item2 a real root */
			Item2->parent = Item2;
			/* Item2 connection to parent is OK */
		} else {
			/*
			 * Case 3)
			 * Connect Item2 to its new parent
			 */
			Item2->parent = Parent(Item1);
			/* Patch parent's child pointer for Item2 */
			if (IsLeftChild(Item1))
				Parent(Item1)->left_child = Item2;
			else
				Parent(Item1)->right_child = Item2;
			/* Item2 connection to parent is OK */
		}
		/*
		 * Commom part for cases 1) and 3) - Item2 parent link is OK here
		 * Item1->parent will be fixed below by the children's parent fixing code
		 *  since Item1 will be a child of Item2
		 * Swap all children
		 */
		SWAP_POINTERS(Item1->left_child, Item2->left_child);
		SWAP_POINTERS(Item1->right_child, Item2->right_child);
		/*
		 * We have a parent-child relationship
		 * So, the value of either Left or RightChild of the upper item (Item1)
		 *  points to Item2
		 * So, the new value of either Left or RightChild of the lower item (Item2)
		 *  points to Item2 - to itself
		 * It must be fixed to point to the new lower item (Item1)
		 */
		if (LeftChild(Item2) == Item2)
			Item2->left_child = Item1;
		else
			Item2->right_child = Item1;
		/*
		 * All 4 child pointers (including the Item1's parent's child pointer) are OK,
		 *  Item2 parent link is OK both sides
		 * Children's parent pointers and Item1->parent will be fixed below
		 */
	} else {
		/* No parent-child relation */
		if (IsRoot(Item1)) {
			/*
			 * Case 2)
			 * Update the root
			 */
			*root = Item2;
			/* Patch parent's child pointer for Item1 */
			if (IsLeftChild(Item2))
				Parent(Item2)->left_child = Item1;
			else
				Parent(Item2)->right_child = Item1;
			/* Patch Item1->parent */
			Item1->parent = Parent(Item2);
			/*
			 * Item1 connection to parent is OK
			 * Make Item2 a real root
			 */
			Item2->parent = Item2;
			/* Item2 connection to parent is OK */
		} else {
			/*
			 * Case 4)
			 * Non-root independent items
			 * Swap parent's child pointers
			 */
			if (IsLeftChild(Item1)) {
				if (IsLeftChild(Item2))
					SWAP_POINTERS(Parent(Item1)->left_child, Parent(Item2)->left_child);
				else
					SWAP_POINTERS(Parent(Item1)->left_child, Parent(Item2)->right_child);
			} else {
				if (IsLeftChild(Item2))
					SWAP_POINTERS(Parent(Item1)->right_child, Parent(Item2)->left_child);
				else
					SWAP_POINTERS(Parent(Item1)->right_child, Parent(Item2)->right_child);
			}
			/* Swap parent pointers */
			SWAP_POINTERS(Item1->parent, Item2->parent);
			/* Both parent connections are OK */
		}
		/*
		 * Commom part for cases 2) and 4) - parent connections are OK here both ways
		 * Swap all children
		 */
		SWAP_POINTERS(Item1->left_child, Item2->left_child);
		SWAP_POINTERS(Item1->right_child, Item2->right_child);
		/*
		 * All 4 child pointers are OK
		 * parent connections are OK both ways
		 * Children's parent pointers will be fixed belowu
		 */
	}
	/*
	 * Common code for all cases
	 * Fix the parent pointers in all children
	 */
	if (LeftChild(Item1) != NULL)
		LeftChild(Item1)->parent = Item1;
	if (RightChild(Item1) != NULL)
		RightChild(Item1)->parent = Item1;
	if (LeftChild(Item2) != NULL)
		LeftChild(Item2)->parent = Item2;
	if (RightChild(Item2) != NULL)
		RightChild(Item2)->parent = Item2;
	/* Done */
}

/*
 * Turns a tree, pulling child up one position
 * Root must point to a variable which will contain the new tree root if the root
 *  is changed by this rotation
 */
static void tree_turn_tree(struct splay_links *child, struct splay_links **root)
{
	u32 DirectOffset;
	u32 ContraOffset;
	struct splay_links *parent;

	parent = Parent(child);
	if (parent == child)
		/* Do nothing for a root */
		return;
	/*
	 * Decide whether this is a left turn or right turn
	 * Left
	 */
	DirectOffset = offsetof(struct splay_links, left_child);
	ContraOffset = offsetof(struct splay_links, right_child);
	if (IsLeftChild(child)) {
		/* Right */
		DirectOffset = offsetof(struct splay_links, right_child);
		ContraOffset = offsetof(struct splay_links, left_child);
	}
#if 0

	Operation(drawing is for a right turn) :

		-was :

		|
		parent
		/ \
		/   \
		/     \
		/	\
		/	  \
		/	    \
		child c
		/ \
		a b

		- will be :

		|
		child
		/ \
		/   \
		/     \
		/	\
		/	  \
		/	    \
		a parent
		/ \
		b c

		- or exactly mirrored version for a left turn
		Left or right turn are dependent on ContraOffset and DirectOffset and thus on whether
		child is a left child or right child

#endif
	/* Make Root to point to super-parent's child pointer if parent is not a root */
	if (!IsRoot(parent)) {
		/* Update the root */
		root = IsLeftChild(parent) ?
		       &(Parent(parent)->left_child) : &(Parent(parent)->right_child);
		/* Link child to its new parent */
		child->parent = Parent(parent);
	} else
		/* child will become the tree root */
		child->parent = child;
	/* Link child to its new parent (or make it a root) */
	*root = child;
	/* Link parent to b */
	ITEM_CHILD(parent, ContraOffset) = ITEM_CHILD(child, DirectOffset);
	if (ITEM_CHILD(parent, ContraOffset) != NULL)
		ITEM_CHILD(parent, ContraOffset)->parent = parent;
	/* Link child to parent */
	ITEM_CHILD(child, DirectOffset) = parent;
	parent->parent = child;
	/* Done */
}
