/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include "debug.h"
#include "filter.h"
#include "vdiff.h"

#include "stcsnap_p.h"
#include "stcsnapdef_p.h"

#define VDIFF_COPY_RESET 1

#define VDIFF_BLOCK_SHIFT 3  /* arbitrary, to save memory when using raw bitmap */
#define VDIFF_SECTORS_PER_BLOCK (1<<VDIFF_BLOCK_SHIFT)
#define VDIFF_BLOCK_SIZE (512<<VDIFF_BLOCK_SHIFT)
static inline sector_t vd_sector_to_block(sector_t sector)
{
	return sector >> VDIFF_BLOCK_SHIFT;
}
static inline sector_t vd_sectors_to_num_blocks(sector_t nr_sects)
{
	return (nr_sects + VDIFF_SECTORS_PER_BLOCK - 1) >> VDIFF_BLOCK_SHIFT;
}
static inline sector_t vd_blocks_to_sectors(sector_t blocks)
{
	return blocks * VDIFF_SECTORS_PER_BLOCK;
}

struct vdiff_context *get_vdiff_bio(struct block_device_filter *bdfilter, struct bio *bio)
{
	int idx;
	sector_t sector_start = bio_sector(bio);
	sector_t sector_end = sector_start + bio_sectors(bio);

	for (idx = 0; idx < bdfilter->num_vdiffs; ++idx) {
		struct vdiff_context *vdiff = bdfilter->vdiffs[idx];

		if (sector_start >= vdiff->vd_sector_start &&
				sector_end <= vdiff->vd_sector_start + vdiff->vd_sectors) {
			stcvsm_atomic_inc(&vdiff->vd_refs);
			return vdiff;
		}
	}
	return NULL;
}

/** Finds an existing vdiff_context, and increment reference count.
 * @return The vdiff or NULL if not found
 */
struct vdiff_context *get_vdiff_bdev(struct block_device_filter *bdfilter,
		struct block_device *bdev)
{
	struct vdiff_context *vdiff;

	spin_lock(bdf_lock(bdfilter));
	vdiff = get_filter_vdiff(bdfilter, bdev->bd_dev);
	if (vdiff != NULL)
		stcvsm_atomic_inc(&(vdiff->vd_refs));
	spin_unlock(bdf_lock(bdfilter));
	return vdiff;
}

void free_vdiff(struct vdiff_context *vdiff)
{
	stcvsm_debug(LOG_VDIFF, "Freeing vdiff %p.\n", vdiff);
	if (vdiff->vd_bdev)
		stcvsm_close_blkdev(vdiff->vd_bdev, FMODE_READ);
#ifdef USE_RANGE_TABLE
	if (vdiff->vd_range_table) {
		if (vdiff->vd_lookaside) {
			range_cleanup(vdiff->vd_range_table);
			kfree(vdiff->vd_lookaside);
		}
		kfree(vdiff->vd_range_table);
	}
#else
	kfree(vdiff->vd_bitmap);
#endif
	kfree(vdiff);
}

static void clear_vdiff_range(struct vdiff_context *vdiff, sector_t sector_start, sector_t nr_sects)
{
	sector_t block_start = vd_sector_to_block(sector_start);
	sector_t block_end = vd_sectors_to_num_blocks(sector_start + nr_sects);

	stcvsm_debug(LOG_VDIFF, "clear: [%llu-%llu)\n", (unsigned long long)block_start,
			(unsigned long long)block_end);

	/* currently not used on live bitmaps, no locking */
#ifdef USE_RANGE_TABLE
	range_delete(vdiff->vd_range_table, block_start, block_end);
#else
	while (block_start < block_end)
		clear_bit(block_start++, vdiff->vd_bitmap);
#endif
}

void set_vdiff_range(struct vdiff_context *vdiff, sector_t sector_start, sector_t nr_sects)
{
	sector_t block_start = vd_sector_to_block(sector_start);
	sector_t block_end = vd_sectors_to_num_blocks(sector_start + nr_sects);

	stcvsm_debug(LOG_VDIFF, "track: [%llu-%llu)\n", (unsigned long long)block_start,
			(unsigned long long)block_end);

	/* lock versus copy_reset_vdiff */
	spin_lock(bdf_lock(vdiff->bdfilter));
#ifdef USE_RANGE_TABLE
	range_insert(vdiff->vd_range_table, block_start, block_end);
#else
	while (block_start < block_end)
		set_bit(block_start++, vdiff->vd_bitmap);
#endif
	spin_unlock(bdf_lock(vdiff->bdfilter));
}

sector_t count_vdiff(struct vdiff_context *vdiff)
{
#ifdef USE_RANGE_TABLE
	return vd_blocks_to_sectors(RangeValuesCount(vdiff->vd_range_table));
#else
	sector_t bits_in_map = vd_sectors_to_num_blocks(vdiff->vd_sectors);

	return vd_blocks_to_sectors(bitmap_weight(vdiff->vd_bitmap, bits_in_map));
#endif
}

static void overlay_vdiff(struct vdiff_context *dst, struct vdiff_context *src)
{
	sector_t bits_in_map = min(vd_sectors_to_num_blocks(src->vd_sectors),
			vd_sectors_to_num_blocks(dst->vd_sectors));
#ifdef USE_RANGE_TABLE
	struct range_ref range = {0};

	range_first(src->vd_range_table, &range);
	while (range.Page != NULL && range.Start < bits_in_map) {
		if (range.End > bits_in_map)
			range.End = bits_in_map;

		spin_lock(bdf_lock(src->bdfilter));
		range_insert(dst->vd_range_table, range.Start, range.End);
		spin_unlock(bdf_lock(src->bdfilter));

		range_next(&range);
	}
#else
	sector_t start = 0;

	while ((start = find_next_bit(src->vd_bitmap, bits_in_map, start)) < bits_in_map) {
		sector_t end = find_next_zero_bit(src->vd_bitmap, bits_in_map, start);

		while (start < end)
			set_bit(start++, dst->vd_bitmap);
	}
#endif
}

void rollback_vdiff(struct vdiff_context *vdiff)
{
	overlay_vdiff(vdiff->vdiff_orig, vdiff);
}

static struct vdiff_context *_alloc_vdiff(sector_t sectors)
{
	struct vdiff_context *vdiff;

	if (ERROR_INJECT(ERRINJ_ALLOC))
		vdiff = NULL;
	else
		vdiff = kzalloc(sizeof(*vdiff), GFP_KERNEL);
	if (!vdiff)
		return NULL;
	atomic_set(&vdiff->vd_refs, 1);
	vdiff->vd_sectors = sectors;

#ifdef USE_RANGE_TABLE
	vdiff->vd_range_table = kmalloc(sizeof(*vdiff->vd_range_table), GFP_KERNEL);
	if (!vdiff->vd_range_table) {
		kfree(vdiff);
		return NULL;
	}
	vdiff->vd_lookaside = kmalloc(sizeof(*vdiff->vd_lookaside), GFP_KERNEL);
	if (!vdiff->vd_lookaside) {
		kfree(vdiff->vd_range_table);
		kfree(vdiff);
		return NULL;
	}
	range_initialize_nonpaged(vdiff->vd_range_table, enum_tree_avl_kind, vdiff->vd_lookaside, 0);
#else
	vdiff->vd_bitmap = kcalloc(BITS_TO_LONGS(vd_sectors_to_num_blocks(vdiff->vd_sectors)),
					sizeof(long),
					GFP_KERNEL);

	if (!vdiff->vd_bitmap) {
		kfree(vdiff);
		return NULL;
	}
#endif
	return vdiff;
}

struct vdiff_context *copy_reset_vdiff(struct block_device_filter *bdfilter,
		struct vdiff_context *vdiff_old, dev_t snap_dev)
{
	int err;
	struct vdiff_context *vdiff;
	sector_t old_nr_sects = vdiff_old->vd_sectors;
	sector_t new_nr_sects = stcvsm_nr_sectors(vdiff_old->vd_bdev);

	vdiff = _alloc_vdiff(new_nr_sects);
	if (!vdiff)
		return ERR_PTR(-ENOMEM);
	vdiff->vd_dev = snap_dev;

	spin_lock(bdf_lock(bdfilter));
	err = attach_filter_vdiff(bdfilter, vdiff);
	spin_unlock(bdf_lock(bdfilter));
	if (err) {
		free_vdiff(vdiff);
		return ERR_PTR(err);
	}

	/* Copy all the stuff! */
	vdiff->vd_sector_start = vdiff_old->vd_sector_start;
	memcpy(vdiff->VDiffGuid, vdiff_old->VDiffGuid, 16);
	/* vdiff->VDiffFileGuid = vdiff_old->VDiffFileGuid; */
	vdiff->VDiffResetCount = vdiff_old->VDiffResetCount;
	vdiff->vdiff_orig = vdiff_old;
	stcvsm_atomic_inc(&vdiff_old->vd_refs);

	/* Swap the bitmaps around */
	{
#ifdef USE_RANGE_TABLE
		struct range_page_table *tmp_range_table = vdiff->vd_range_table;
		struct lookaside_list *tmp_lookaside = vdiff->vd_lookaside;

		vdiff->vd_range_table = vdiff_old->vd_range_table;
		vdiff->vd_lookaside = vdiff_old->vd_lookaside;

		spin_lock(bdf_lock(bdfilter));
		vdiff_old->vd_range_table = tmp_range_table;
		vdiff_old->vd_lookaside = tmp_lookaside;
		vdiff_old->vd_sectors = new_nr_sects;
		spin_unlock(bdf_lock(bdfilter));
#else
		unsigned long *tmp_bitmap, *new_bitmap;

		/* Ensure bitmap can be grown before committing to the swap */
		if (new_nr_sects > old_nr_sects) {
			new_bitmap = kmalloc_array(BITS_TO_LONGS(vd_sectors_to_num_blocks(new_nr_sects)),
						sizeof(long),
						GFP_KERNEL);
			if (!new_bitmap) {
				free_vdiff(vdiff);
				return ERR_PTR(-ENOMEM);
			}
		}

		tmp_bitmap = vdiff->vd_bitmap;
		vdiff->vd_bitmap = vdiff_old->vd_bitmap;

		spin_lock(bdf_lock(bdfilter));
		vdiff_old->vd_bitmap = tmp_bitmap;
		vdiff_old->vd_sectors = new_nr_sects;
		spin_unlock(bdf_lock(bdfilter));
#endif
		if (new_nr_sects != old_nr_sects) {
			vdiff->vd_sectors = new_nr_sects;

			if (new_nr_sects > old_nr_sects) {
				/* Volume has been extended, pad stolen bitmap with dirty bits */
#ifndef USE_RANGE_TABLE
				memcpy(new_bitmap, vdiff->vd_bitmap,
					BITS_TO_LONGS(vd_sectors_to_num_blocks(old_nr_sects)) * sizeof(long));
				kfree(vdiff->vd_bitmap);
				vdiff->vd_bitmap = new_bitmap;
#endif
				set_vdiff_range(vdiff, old_nr_sects, new_nr_sects - old_nr_sects);
			} else {
				/* Volume has been reduced, truncate stolen bitmap */
				clear_vdiff_range(vdiff, new_nr_sects, old_nr_sects - new_nr_sects);
			}
		}
	}

	return vdiff;
}

void put_vdiff(struct vdiff_context *vdiff)
{
	struct block_device_filter *bdfilter = vdiff->bdfilter;

	if (stcvsm_atomic_dec_and_lock(&vdiff->vd_refs, bdf_lock(bdfilter))) {
		detach_filter_vdiff(bdfilter, vdiff);
		spin_unlock(bdf_lock(bdfilter));
		if (vdiff->vdiff_orig)
			put_vdiff(vdiff->vdiff_orig);
		free_vdiff(vdiff);
		put_filter(bdfilter);
	}
}

struct vdiff_context *alloc_vdiff(struct block_device *bdev)
{
	struct vdiff_context *vdiff;
	struct block_device *vd_bdev;

	vdiff = _alloc_vdiff(stcvsm_nr_sectors(bdev));
	if (!vdiff)
		return ERR_PTR(-ENOMEM);
	vdiff->vd_dev = bdev->bd_dev;

	/* Hold open to prevent repartitioning, but LVM can still resize */
	vd_bdev = stcvsm_open_blkdev(bdev->bd_dev, FMODE_READ, NULL);
	if (IS_ERR(vd_bdev)) {
		int err = PTR_ERR(vd_bdev);

		free_vdiff(vdiff);
		return ERR_PTR(err);
	}
	vdiff->vd_bdev = vd_bdev;
	vdiff->vd_sector_start = get_start_sect(bdev);

	return vdiff;
}

void shutdown_vdiff(struct vdiff_context *vdiff)
{
	/* Handle the race for multiple shutdown calls */
	if (!test_and_set_bit(VDIFF_SHUTDOWN, &vdiff->vd_flags))
		put_vdiff(vdiff);
	else
		stcvsm_notice(LOG_VDIFF, "Shutdown called when already shutdown.");
}

void get_vdiff_range(struct vdiff_context *vdiff, sector_t start_sect, sector_t *nr_sects, bool *diffed)
{
	sector_t run;
	sector_t start = vd_sector_to_block(start_sect);
	sector_t bits_in_map = vd_sectors_to_num_blocks(vdiff->vd_sectors);

	spin_lock(bdf_lock(vdiff->bdfilter));

#ifdef USE_RANGE_TABLE
	if (start >= bits_in_map) {
		*diffed = false;
		run = 0;
	} else {
		struct range_ref range = {0};

		if (range_next_range_in(vdiff->vd_range_table, start, bits_in_map, &range)) {
			if (start == range.Start) {
				*diffed = true;
				run = range.End - start;
			} else {
				*diffed = false;
				run = range.Start - start;
			}
		} else {
			*diffed = false;
			run = bits_in_map - start;
		}
	}
#else
	if (start >= bits_in_map) {
		*diffed = false;
		run = 0;
	} else if (test_bit(start, vdiff->vd_bitmap)) {
		*diffed = true;
		run = find_next_zero_bit(vdiff->vd_bitmap, bits_in_map, start) - start;
	} else {
		*diffed = false;
		run = find_next_bit(vdiff->vd_bitmap, bits_in_map, start) - start;
	}
#endif
	spin_unlock(bdf_lock(vdiff->bdfilter));
	*nr_sects = vd_blocks_to_sectors(run);
	/* Adjust for non-aligned sector request */
	*nr_sects -= start_sect - vd_blocks_to_sectors(start);
	stcvsm_debug(LOG_VDIFF, "VDiff found: [%d]%llu-%llu\n", *diffed, (u64)start_sect, (u64)*nr_sects);
}

static void get_vdiff_info2(struct vdiff_context *vdiff, struct block_device *bdev, VSNAP_IOCTL_INFO *info)
{
	if (!vdiff) {
		memset(info, 0, sizeof(*info));
		info->Info.SectorSize = 512;
		info->Info.TotalSectors = stcvsm_nr_sectors(bdev);
		stcvsm_debug(LOG_CTL, "VDiff not active\n");
	} else {
		info->Info.VDiffActive = true;
		info->Info.SectorSize = 512;
		info->Info.DirtySectors = count_vdiff(vdiff);
		stcvsm_debug(LOG_CTL, "Info.DirtySectors: %llu\n", info->Info.DirtySectors);
		info->Info.TotalSectors = vdiff->vd_sectors;
		info->Info.VDiffResetCount = vdiff->VDiffResetCount;
		printk_uuid(debug, "Guid     ", vdiff->VDiffGuid);
		printk_uuid(debug, "FileGuid ", vdiff->VDiffFileGuid);
		memcpy(info->Info.VDiffGuid, vdiff->VDiffGuid, 16);
		memcpy(info->Info.VDiffFileGuid, vdiff->VDiffFileGuid, 16);
		stcvsm_debug(LOG_CTL, "VDiff active\n");
	}
}

static void _reset_vdiff(struct vdiff_context *vdiff)
{
	/* Set bitmap to all zeros */
#ifdef USE_RANGE_TABLE
	range_delete_all(vdiff->vd_range_table);
#else
	bitmap_zero(vdiff->vd_bitmap, vd_sectors_to_num_blocks(vdiff->vd_sectors));
#endif
	set_bit(VDIFF_RESET, &vdiff->vd_flags);
	vdiff->VDiffResetCount++;
}

void reset_vdiff_snapshot(struct vdiff_context *vdiff)
{
	stcvsm_notice(LOG_VDIFF, "Reset vdiff %p.\n", vdiff);
	spin_lock(bdf_lock(vdiff->bdfilter));
	_reset_vdiff(vdiff);
	if (vdiff->vdiff_orig)
		vdiff->vdiff_orig->VDiffResetCount++;
	spin_unlock(bdf_lock(vdiff->bdfilter));
}

/* deprecated */
void reset_vdiff_guid(struct vdiff_context *vdiff, u8 *chain_guid)
{
	u8 null_guid[16] = {0};
	int is_null = memcmp(chain_guid, null_guid, 16) == 0;
	stcvsm_notice(LOG_VDIFF, "Reset vdiff guid %p.\n", vdiff);
	printk_uuid(info, "chain_guid ", chain_guid);

	spin_lock(bdf_lock(vdiff->bdfilter));
	_reset_vdiff(vdiff);
	if (!is_null) {
		memcpy(vdiff->VDiffGuid, chain_guid, 16);
		memset(vdiff->VDiffFileGuid, 0, 16);
		vdiff->VDiffResetCount = 0;
	}
	spin_unlock(bdf_lock(vdiff->bdfilter));
}

void rean_vdiff(struct vdiff_context *vdiff, u8 *chain_guid, u8 *file_guid, u32 reset_count)
{
	stcvsm_notice(LOG_VDIFF, "Rean vdiff %p %u.\n", vdiff, reset_count);
	printk_uuid(info, "chain_guid ", chain_guid);
	printk_uuid(info, "file_guid ", file_guid);
	spin_lock(bdf_lock(vdiff->bdfilter));
	memcpy(vdiff->VDiffGuid, chain_guid, 16);
	memcpy(vdiff->VDiffFileGuid, file_guid, 16);
	vdiff->VDiffResetCount = reset_count;
	spin_unlock(bdf_lock(vdiff->bdfilter));
}

int stcvsm_ioctl_vdiff(dev_t dev, unsigned int cmd, void *param)
{
	int retval;
	struct block_device *bdev;
	struct block_device_filter *bdfilter;
	bool create = (cmd == IOCTL_VSNAP_VDIFF_START || cmd == VDIFF_COPY_RESET);

	bdev = stcvsm_open_blkdev(dev, FMODE_READ, NULL);
	if (IS_ERR(bdev))
		return PTR_ERR(bdev);

	bdfilter = get_filter(bdev, create);
	if (!IS_ERR(bdfilter)) {
		retval = _stcvsm_ioctl_vdiff(bdfilter, bdev, cmd, param);
		if (bdfilter)
			put_filter(bdfilter);
	} else {
		retval = PTR_ERR(bdfilter);
	}
	stcvsm_close_blkdev(bdev, FMODE_READ);
	return retval;
}

int _stcvsm_ioctl_vdiff(struct block_device_filter *bdfilter, struct block_device *bdev,
		unsigned int cmd, void *param)
{
	int retval = 0;
	struct vdiff_context *vdiff = NULL;

	if (cmd == IOCTL_VSNAP_VDIFF_STATS_2) {
		if (bdfilter)
			vdiff = get_vdiff_bdev(bdfilter, bdev);
		get_vdiff_info2(vdiff, bdev, param);
		if (vdiff)
			put_vdiff(vdiff);
		return 0;
	}

	/* All other commands require that the filter was found. */
	if (!bdfilter)
		return -ENOENT;

	if (cmd == IOCTL_VSNAP_VDIFF_START) {
		vdiff = alloc_vdiff(bdev);
		if (!IS_ERR(vdiff)) {
			VSNAP_IOCTL_GUID *buf = (VSNAP_IOCTL_GUID *)param;

#ifdef USE_RANGE_TABLE
			range_insert(vdiff->vd_range_table, 0, vd_sectors_to_num_blocks(vdiff->vd_sectors));
#else
			bitmap_fill(vdiff->vd_bitmap, vd_sectors_to_num_blocks(vdiff->vd_sectors));
#endif
			memcpy(vdiff->VDiffGuid, buf->Guid, 16);
			spin_lock(bdf_lock(bdfilter));
			retval = attach_filter_vdiff(bdfilter, vdiff);
			if (retval) {
				spin_unlock(bdf_lock(bdfilter));
				stcvsm_debug(LOG_FILTER, "Failed to start incremental tracking (%d)\n", retval);
				free_vdiff(vdiff);
			} else {
				activate_filter(bdfilter);
				spin_unlock(bdf_lock(bdfilter));
			}
			vdiff = NULL;
		}
	} else if (cmd == VDIFF_COPY_RESET) {
		vdiff = copy_reset_vdiff(bdfilter, param, bdev->bd_dev);
	} else {
		vdiff = get_vdiff_bdev(bdfilter, bdev);
		if (!vdiff) {
			retval = -ENOENT;
			stcvsm_notice(LOG_CTL, "No incremental tracking found\n");
		} else {
			switch (cmd) {
			case IOCTL_VSNAP_VDIFF_RESET:
			{
				struct _VSNAP_IOCTL_REANIMATE_BUFF *buf = (struct _VSNAP_IOCTL_REANIMATE_BUFF *)param;
				retval = 0;

				switch (buf->flag) {
				case VDIFF_RESET_SNAPSHOT:
					reset_vdiff_snapshot(vdiff);
					break;
				case VDIFF_RESET_GUID:
					reset_vdiff_guid(vdiff, buf->buf.NewGuid);
					break;
				case VDIFF_RESET_REAN:
					rean_vdiff(vdiff, buf->buf.NewGuid, buf->buf.NewFileGuid, buf->buf.NewResetCount);
					break;
				default:
					retval = -EOPNOTSUPP;
					break;
				}
				break;
			}
			case IOCTL_VSNAP_VDIFF_SET_FILE_GUID:
			{
				VSNAP_IOCTL_GUID *guid = (VSNAP_IOCTL_GUID *)param;

				memcpy(vdiff->VDiffFileGuid, guid->Guid, 16);
				if (vdiff->vdiff_orig)
					memcpy(vdiff->vdiff_orig->VDiffFileGuid, guid->Guid, 16);
				break;
			}
			case IOCTL_VSNAP_VDIFF_GET_RANGE:
			{
				VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER *buf = (VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER *)param;
				sector_t nr_sects;

				get_vdiff_range(vdiff, buf->start, &nr_sects, &buf->AreDiffed);

				/* Unfortunately, NumberOfSectors is 32 bit */
				if (nr_sects >= 0x100000000UL)
					nr_sects = 0x100000000UL - VDIFF_SECTORS_PER_BLOCK;
				buf->NumberOfSectors = (u32)nr_sects;
				break;
			}
			case IOCTL_VSNAP_VDIFF_NATIVE_ON:
			{
				VSNAP_IOCTL_NATIVE_SUPP_BUFF *buf = (VSNAP_IOCTL_NATIVE_SUPP_BUFF *)param;

				retval = stcvsm_ioctl_vdiff(MKDEV(buf->dst.major, buf->dst.minor),
						VDIFF_COPY_RESET, vdiff);
				break;
			}
			case IOCTL_VSNAP_VDIFF_NATIVE_OFF:
			case IOCTL_VSNAP_VDIFF_STOP:
				shutdown_vdiff(vdiff);
				break;
			default:
				retval = -EOPNOTSUPP;
				break;
			}
		}
		if (vdiff)
			put_vdiff(vdiff);
		return retval;
	}

	if (IS_ERR(vdiff)) {
		retval = PTR_ERR(vdiff);
		stcvsm_notice(LOG_FILTER, "Failed to create incremental tracking (%d)\n", retval);
	}
	return retval;
}
