/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either Version 3 of the License, or (at your option) any later
 * version.
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, see <https://www.gnu.org/licenses>.
 * SUPPLEMENTAL TERMS UNDER SECTION 7 OF GNU GENERAL PUBLIC LICENSE
 * The following terms (“Supplemental Terms”) supplement and modify the GNU
 * General Public License, Version 3 applicable to the Program.
 *
 * Copyright © 2014-2022 Arcserve (USA) LLC
 * SPDX-License-Identifier: GPL-2.0+
 */

#include <linux/kthread.h>
#include <linux/slab.h>
#include <stdarg.h>

#include "debug.h"
#include "workqueue.h"

/* Owned by module_init/module_exit */
struct wq* global_wq()
{
	static struct wq _wq;
	return &_wq;
}

void wq_init(struct wq *wq, const char *name, ...)
{
	va_list args;

	spin_lock_init(&wq->lock);
	INIT_LIST_HEAD(&wq->work);
	init_waitqueue_head(&wq->more_work);
	wq->thread = NULL;
	va_start(args, name);
	wq->name = kvasprintf(GFP_KERNEL, name, args);
	va_end(args);
}

void wq_deinit(struct wq *wq)
{
	kfree(wq->name);
}

static int worker_thread(void *_wq)
{
	struct wq *wq = (struct wq *)_wq;
	DEFINE_WAIT(wait);

	while (1) {
		prepare_to_wait(&wq->more_work, &wait, TASK_INTERRUPTIBLE);
		if (!kthread_should_stop() && list_empty(&wq->work))
			schedule();
		finish_wait(&wq->more_work, &wait);

		spin_lock_irq(&wq->lock);
		while (!list_empty(&wq->work)) {
			struct wq_work *work = list_entry(wq->work.next, struct wq_work, list);

			list_del_init(wq->work.next);
			spin_unlock_irq(&wq->lock);

			work->fn(work);

			spin_lock_irq(&wq->lock);
		}
		spin_unlock_irq(&wq->lock);

		if (kthread_should_stop())
			break;
		if (!wq->thread) {
			if (wq->cleanup_fn)
				wq->cleanup_fn(wq->cleanup_data);
			break;
		}
	}
	return 0;
}

int wq_start(struct wq *wq)
{
	struct task_struct *p;

	if (ERROR_INJECT(ERRINJ_ALLOC))
		p = ERR_PTR(-ENOMEM);
	else
		p = kthread_create(worker_thread, wq, "%s", wq->name);
	if (IS_ERR(p))
		return PTR_ERR(p);
	wq->thread = p;
	wake_up_process(p);
	return 0;
}

int wq_stop(struct wq *wq, void (*cb)(void *data), void *data)
{
	int r = 0;

	if (wq->thread) {
		if (wq->thread == current) {
			wq->cleanup_fn = cb;
			wq->cleanup_data = data;
			wq->thread = NULL;
			r = -EAGAIN;
		} else {
			kthread_stop(wq->thread);  /* wakes too; no need to wake_up() myself */
			wq->thread = NULL;
			if (cb)
				cb(data);
		}
	}
	return r;
}

void wq_insert_work(struct wq *wq, struct wq_work *work)
{
	unsigned long lflags;

	INIT_LIST_HEAD(&work->list);
	spin_lock_irqsave(&wq->lock, lflags);
	list_add_tail(&work->list, &wq->work);
	wake_up(&wq->more_work);
	spin_unlock_irqrestore(&wq->lock, lflags);
}
